@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2007  The PARI group.
+/* Copyright (C) 2007  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -18,6 +16,22 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /* Not so fast arithmetic with polynomials over Fp */
 
+static GEN
+get_FpX_red(GEN T, GEN *B)
+{
+  if (typ(T)!=t_VEC) { *B=NULL; return T; }
+  *B = gel(T,1); return gel(T,2);
+}
+
+GEN
+get_FpX_mod(GEN T) { return typ(T)==t_VEC? gel(T,2): T; }
+
+long
+get_FpX_var(GEN T) { return typ(T)==t_VEC? varn(gel(T,2)): varn(T); }
+
+long
+get_FpX_degree(GEN T) { return typ(T)==t_VEC? degpol(gel(T,2)): degpol(T); }
+
 /***********************************************************************/
 /**                                                                   **/
 /**                              FpX                                  **/
@@ -35,6 +49,33 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
  * compatible.
  */
 
+static ulong
+to_Flx(GEN *P, GEN *Q, GEN p)
+{
+  ulong pp = (ulong)p[2];
+  *P = ZX_to_Flx(*P, pp);
+  *Q = ZX_to_Flx(*Q, pp); return pp;
+}
+
+static ulong
+to_Flxq(GEN *P, GEN *T, GEN p)
+{
+  ulong pp = (ulong)p[2];
+  if (P) *P = ZX_to_Flx(*P, pp);
+  *T = ZXT_to_FlxT(*T, pp); return pp;
+}
+
+GEN
+Z_to_FpX(GEN a, GEN p, long v)
+{
+  pari_sp av = avma;
+  GEN z = cgetg(3, t_POL);
+  GEN x = modii(a, p);
+  if (!signe(x)) { avma =av; return pol_0(v); }
+  z[1] = evalsigne(1) | evalvarn(v);
+  gel(z,2) = x; return z;
+}
+
 /* z in Z[X], return lift(z * Mod(1,p)), normalized*/
 GEN
 FpX_red(GEN z, GEN p)
@@ -53,6 +94,20 @@ FpXV_red(GEN z, GEN p)
   return x;
 }
 
+GEN
+FpXT_red(GEN z, GEN p)
+{
+  if (typ(z) == t_POL)
+    return FpX_red(z, p);
+  else
+  {
+    long i,l = lg(z);
+    GEN x = cgetg(l, t_VEC);
+    for (i=1; i<l; i++) gel(x,i) = FpXT_red(gel(z,i), p);
+    return x;
+  }
+}
+
 GEN
 FpX_normalize(GEN z, GEN p)
 {
@@ -208,7 +263,7 @@ Fp_FpX_sub(GEN x, GEN y, GEN p)
     z = cgetg(3, t_POL);
     x = (ly == 3)? Fp_sub(x, gel(y,2), p): modii(x, p);
     if (!signe(x)) { avma = (pari_sp)(z + 3); return pol_0(varn(y)); }
-    z[1] = y[1]; gel(z,2) = x; return z;
+    z[1] = evalsigne(1)|y[1]; gel(z,2) = x; return z;
   }
   z = cgetg(ly,t_POL);
   gel(z,2) = Fp_sub(x, gel(y,2), p);
@@ -221,19 +276,42 @@ Fp_FpX_sub(GEN x, GEN y, GEN p)
 GEN
 FpX_mul(GEN x,GEN y,GEN p) { return FpX_red(ZX_mul(x, y), p); }
 
+GEN
+FpX_mulspec(GEN a, GEN b, GEN p, long na, long nb)
+{ return FpX_red(ZX_mulspec(a, b, na, nb), p); }
+
 GEN
 FpX_sqr(GEN x,GEN p) { return FpX_red(ZX_sqr(x), p); }
 
+GEN
+FpX_mulu(GEN y, ulong x,GEN p)
+{
+  GEN z;
+  long i, l;
+  if (!x) return zeropol(varn(y));
+  z = cgetg_copy(y, &l); z[1] = y[1];
+  for(i=2; i<l; i++) gel(z,i) = Fp_mulu(gel(y,i), x, p);
+  return z;
+}
+
+GEN
+FpX_Fp_mulspec(GEN y,GEN x,GEN p,long ly)
+{
+  GEN z;
+  long i;
+  if (!signe(x)) return pol_0(0);
+  z = cgetg(ly+2,t_POL); z[1] = evalsigne(1);
+  for(i=0; i<ly; i++) gel(z,i+2) = Fp_mul(gel(y,i), x, p);
+  return ZX_renormalize(z, ly+2);
+}
+
 GEN
 FpX_Fp_mul(GEN y,GEN x,GEN p)
 {
-  GEN z;
-  long i, l;
-  if (!signe(x)) return pol_0(varn(y));
-  z = cgetg_copy(y, &l); z[1] = y[1];
-  for(i=2; i<l; i++) gel(z,i) = Fp_mul(gel(y,i), x, p);
-  return ZX_renormalize(z, l);
+  GEN z = FpX_Fp_mulspec(y+2,x,p,lgpol(y));
+  setvarn(z, varn(y)); return z;
 }
+
 GEN
 FpX_Fp_mul_to_monic(GEN y,GEN x,GEN p)
 {
@@ -244,14 +322,14 @@ FpX_Fp_mul_to_monic(GEN y,GEN x,GEN p)
   gel(z,l-1) = gen_1; return z;
 }
 
-GEN
-FpX_divrem(GEN x, GEN y, GEN p, GEN *pr)
+static GEN
+FpX_divrem_basecase(GEN x, GEN y, GEN p, GEN *pr)
 {
   long vx, dx, dy, dz, i, j, sx, lr;
   pari_sp av0, av;
   GEN z,p1,rem,lead;
 
-  if (!signe(y)) pari_err(gdiver);
+  if (!signe(y)) pari_err_INV("FpX_divrem",y);
   vx = varn(x);
   dy = degpol(y);
   dx = degpol(x);
@@ -281,10 +359,8 @@ FpX_divrem(GEN x, GEN y, GEN p, GEN *pr)
   av0 = avma; dz = dx-dy;
   if (lgefint(p) == 3)
   { /* assume ab != 0 mod p */
-    ulong pp = (ulong)p[2];
-    GEN a = ZX_to_Flx(x, pp);
-    GEN b = ZX_to_Flx(y, pp);
-    z = Flx_divrem(a,b,pp, pr);
+    ulong pp = to_Flx(&x, &y, p);
+    z = Flx_divrem(x, y, pp, pr);
     avma = av0; /* HACK: assume pr last on stack, then z */
     if (!z) return NULL;
     z = leafcopy(z);
@@ -350,17 +426,13 @@ FpX_divrem(GEN x, GEN y, GEN p, GEN *pr)
 GEN
 FpX_div_by_X_x(GEN a, GEN x, GEN p, GEN *r)
 {
-  long l = lg(a), i;
-  GEN z = cgetg(l-1, t_POL), a0, z0;
+  long l = lg(a)-1, i;
+  GEN z = cgetg(l, t_POL);
   z[1] = evalsigne(1) | evalvarn(0);
-  a0 = a + l-1;
-  z0 = z + l-2; *z0 = *a0--;
-  for (i=l-3; i>1; i--) /* z[i] = (a[i+1] + x*z[i+1]) % p */
-  {
-    GEN t = addii(gel(a0--,0), Fp_mul(x, gel(z0--,0), p));
-    *z0 = (long)t;
-  }
-  if (r) *r = addii(gel(a0,0), Fp_mul(x, gel(z0,0), p));
+  gel(z, l-1) = gel(a,l);
+  for (i=l-2; i>1; i--) /* z[i] = a[i+1] + x*z[i+1] */
+    gel(z, i) = Fp_addmul(gel(a,i+1), x, gel(z,i+1), p);
+  if (r) *r = Fp_addmul(gel(a,2), x, gel(z,2), p);
   return z;
 }
 
@@ -444,10 +516,8 @@ FpX_FpXM_qmul(GEN q, GEN M, GEN p)
 static GEN
 matid2_FpXM(long v)
 {
-  GEN m = cgetg(3, t_MAT);
-  gel(m,1) = mkcol2(pol_1(v),pol_0(v));
-  gel(m,2) = mkcol2(pol_0(v),pol_1(v));
-  return m;
+  retmkmat2(mkcol2(pol_1(v),pol_0(v)),
+            mkcol2(pol_0(v),pol_1(v)));
 }
 
 static GEN
@@ -458,12 +528,12 @@ FpX_halfgcd_split(GEN x, GEN y, GEN p)
   GEN y1, r, q;
   long l = lgpol(x), n = l>>1, k;
   if (lgpol(y)<=n) return matid2_FpXM(varn(x));
-  R = FpX_halfgcd(RgX_shift(x,-n),RgX_shift(y,-n),p);
+  R = FpX_halfgcd(RgX_shift_shallow(x,-n),RgX_shift_shallow(y,-n),p);
   V = FpXM_FpX_mul2(R,x,y,p); y1 = gel(V,2);
   if (lgpol(y1)<=n) return gerepilecopy(av, R);
   q = FpX_divrem(gel(V,1), y1, p, &r);
   k = 2*n-degpol(y1);
-  S = FpX_halfgcd(RgX_shift(y1,-k), RgX_shift(r,-k),p);
+  S = FpX_halfgcd(RgX_shift_shallow(y1,-k), RgX_shift_shallow(r,-k),p);
   return gerepileupto(av, FpXM_mul2(S,FpX_FpXM_qmul(q,R,p),p));
 }
 
@@ -485,11 +555,17 @@ FpX_halfgcd(GEN x, GEN y, GEN p)
   GEN M,q,r;
   if (lgefint(p)==3)
   {
-    ulong pp=p[2];
-    M = FlxM_to_ZXM(Flx_halfgcd(ZX_to_Flx(x, pp),ZX_to_Flx(y, pp), pp));
+    ulong pp = to_Flx(&x, &y, p);
+    M = FlxM_to_ZXM(Flx_halfgcd(x, y, pp));
   }
   else
   {
+    if (!signe(x))
+    {
+      long v = varn(x);
+      retmkmat2(mkcol2(pol_0(v),pol_1(v)),
+                mkcol2(pol_1(v),pol_0(v)));
+    }
     if (degpol(y)<degpol(x)) return FpX_halfgcd_i(x,y,p);
     q = FpX_divrem(y,x,p,&r);
     M = FpX_halfgcd_i(x,r,p);
@@ -522,11 +598,10 @@ FpX_gcd(GEN x, GEN y, GEN p)
   pari_sp av = avma;
   if (lgefint(p)==3)
   {
-    ulong pp = p[2];
+    ulong pp;
     (void)new_chunk((lg(x) + lg(y)) << 2); /* scratch space */
-    x = ZX_to_Flx(x, pp);
-    y = ZX_to_Flx(y, pp);
-    x = Flx_gcd(x,y, pp);
+    pp = to_Flx(&x, &y, p);
+    x = Flx_gcd(x, y, pp);
     avma = av; return Flx_to_ZX(x);
   }
   x = FpX_red(x, p);
@@ -627,9 +702,7 @@ FpX_extgcd(GEN x, GEN y, GEN p, GEN *ptu, GEN *ptv)
   pari_sp ltop=avma;
   if (lgefint(p)==3)
   {
-    ulong pp=p[2];
-    x = ZX_to_Flx(x, pp);
-    y = ZX_to_Flx(y, pp);
+    ulong pp = to_Flx(&x, &y, p);
     d = Flx_extgcd(x,y, pp, ptu,ptv);
     d = Flx_to_ZX(d);
     if (ptu) *ptu=Flx_to_ZX(*ptu);
@@ -711,12 +784,12 @@ FpX_eval(GEN x,GEN y,GEN p)
         goto fppoleval;/*sorry break(2) no implemented*/
       }
     r = (i==j)? y: Fp_powu(y,i-j+1,p);
-    p1 = modii(addii(mulii(p1,r), gel(x,j)),p);
+    p1 = Fp_addmul(gel(x,j), p1, r, p);
+    if ((i & 7) == 0) { affii(p1, res); p1 = res; avma = av; }
   }
  fppoleval:
   modiiz(p1,p,res);
-  avma=av;
-  return res;
+  avma = av; return res;
 }
 
 /* Tz=Tx*Ty where Tx and Ty coprime
@@ -777,6 +850,17 @@ FpX_resultant(GEN a, GEN b, GEN p)
   return gerepileuptoint(av, res);
 }
 
+GEN
+FpX_disc(GEN x, GEN p)
+{
+  pari_sp av = avma;
+  GEN L, D = FpX_resultant(x, FpX_deriv(x,p), p);
+  if (!D || !signe(D)) return gen_0;
+  L = leading_term(x); if (!equali1(L)) D = Fp_div(D,L,p);
+  if (degpol(x) & 2) D = Fp_neg(D,p);
+  return gerepileuptoint(av, D);
+}
+
 static GEN _FpX_mul(void *p,GEN a,GEN b){return FpX_mul(a,b,(GEN)p);}
 GEN
 FpXV_prod(GEN V, GEN p)
@@ -795,7 +879,8 @@ FpV_roots_to_pol(GEN V, GEN p, long v)
   return gerepileupto(ltop,FpXV_prod(g,p));
 }
 
-/* invert all elements of x mod p using Montgomery's trick. Not stack-clean. */
+/* invert all elements of x mod p using Montgomery's multi-inverse trick.
+ * Not stack-clean. */
 GEN
 FpV_inv(GEN x, GEN p)
 {
@@ -833,12 +918,12 @@ FqV_inv(GEN x, GEN T, GEN p)
 
 /***********************************************************************/
 /**                                                                   **/
-/**                  Montgomery reduction                            **/
+/**                      Barrett reduction                            **/
 /**                                                                   **/
 /***********************************************************************/
 
 static GEN
-FpX_invMontgomery_basecase(GEN T, GEN p)
+FpX_invBarrett_basecase(GEN T, GEN p)
 {
   long i, l=lg(T)-1, lr = l-1, k;
   GEN r=cgetg(lr, t_POL); r[1]=T[1];
@@ -871,13 +956,7 @@ FpX_recipspec(GEN x, long l, long n)
 }
 
 static GEN
-FpX_mulspec(GEN a, GEN b, GEN p, long na, long nb)
-{
-  return FpX_red(ZX_mulspec(a, b, na, nb), p);
-}
-
-static GEN
-FpX_invMontgomery_Newton(GEN T, GEN p)
+FpX_invBarrett_Newton(GEN T, GEN p)
 {
   pari_sp av = avma;
   long nold, lx, lz, lq, l = degpol(T), i, lQ;
@@ -889,6 +968,7 @@ FpX_invMontgomery_Newton(GEN T, GEN p)
 
   /* initialize */
   gel(x,0) = Fp_inv(gel(q,0), p);
+  if (lQ>1) gel(q,1) = Fp_red(gel(q,1), p);
   if (lQ>1 && signe(gel(q,1)))
   {
     GEN u = gel(q, 1);
@@ -931,67 +1011,171 @@ FpX_invMontgomery_Newton(GEN T, GEN p)
 
 /* 1/polrecip(T)+O(x^(deg(T)-1)) */
 GEN
-FpX_invMontgomery(GEN T, GEN p)
+FpX_invBarrett(GEN T, GEN p)
 {
   pari_sp ltop = avma;
   long l = lg(T);
   GEN r;
-  if (l<5) return pol_0(T[1]);
-  if (l<=FpX_INVMONTGOMERY_LIMIT)
+  if (l<5) return pol_0(varn(T));
+  if (l<=FpX_INVBARRETT_LIMIT)
   {
     GEN c = gel(T,l-1), ci=gen_1;
     if (!equali1(c))
     {
       ci = Fp_inv(c, p);
       T = FpX_Fp_mul(T, ci, p);
-      r = FpX_invMontgomery_basecase(T, p);
+      r = FpX_invBarrett_basecase(T, p);
       r = FpX_Fp_mul(r, ci, p);
     } else
-      r = FpX_invMontgomery_basecase(T, p);
+      r = FpX_invBarrett_basecase(T, p);
   }
   else
-    r = FpX_invMontgomery_Newton(T, p);
+    r = FpX_invBarrett_Newton(T, p);
   return gerepileupto(ltop, r);
 }
 
-/* Compute x mod T where degpol(x)<=2*(degpol(T)-1) i.e. lgpol(x)<2*lgpol(T)-2
- * and mg is the Montgomery inverse of T.
- */
 GEN
-FpX_rem_Montgomery(GEN x, GEN mg, GEN T, GEN p)
+FpX_get_red(GEN T, GEN p)
 {
-  pari_sp ltop=avma;
-  GEN z;
-  long l  = lgpol(x);
+  if (typ(T)==t_POL && lg(T)>FpX_BARRETT_LIMIT)
+    retmkvec2(FpX_invBarrett(T,p),T);
+  return T;
+}
+
+/* Compute x mod T where 2 <= degpol(T) <= l+1 <= 2*(degpol(T)-1)
+ * and mg is the Barrett inverse of T. */
+static GEN
+FpX_divrem_Barrettspec(GEN x, long l, GEN mg, GEN T, GEN p, GEN *pr)
+{
+  GEN q, r;
   long lt = degpol(T); /*We discard the leading term*/
   long ld, lm, lT, lmg;
-  if (l<=lt)
-    return ZX_copy(x);
   ld = l-lt;
   lm = minss(ld, lgpol(mg));
   lT  = ZX_lgrenormalizespec(T+2,lt);
   lmg = ZX_lgrenormalizespec(mg+2,lm);
-  z = FpX_recipspec(x+2+lt,ld,ld);              /* z = rec(x)     lz<=ld*/
-  z = FpX_mulspec(z+2,mg+2,p,lgpol(z),lmg);    /* z = rec(x) * mg lz<=ld+lm*/
-  z = FpX_recipspec(z+2,minss(ld,lgpol(z)),ld);/* z = rec (rec(x) * mg) lz<=ld*/
-  z = FpX_mulspec(z+2,T+2,p,lgpol(z),lT);      /* z *= pol        lz<=ld+lt*/
-  z = FpX_subspec(x+2,z+2,p,lt,minss(lt,lgpol(z)));/* z = x - z   lz<=lt */
-  z[1] = x[1];
-  return gerepileupto(ltop, z);
+  q = FpX_recipspec(x+lt,ld,ld);              /* q = rec(x)     lq<=ld*/
+  q = FpX_mulspec(q+2,mg+2,p,lgpol(q),lmg);    /* q = rec(x) * mg lq<=ld+lm*/
+  q = FpX_recipspec(q+2,minss(ld,lgpol(q)),ld);/* q = rec (rec(x) * mg) lq<=ld*/
+  if (!pr) return q;
+  r = FpX_mulspec(q+2,T+2,p,lgpol(q),lT);      /* r = q*pol        lr<=ld+lt*/
+  r = FpX_subspec(x,r+2,p,lt,minss(lt,lgpol(r)));/* r = x - r   lr<=lt */
+  if (pr == ONLY_REM) return r;
+  *pr = r; return q;
+}
+
+static GEN
+FpX_divrem_Barrett_noGC(GEN x, GEN mg, GEN T, GEN p, GEN *pr)
+{
+  GEN q = NULL, r = FpX_red(x, p);
+  long l = lgpol(r), lt = degpol(T), lm = 2*lt-1;
+  long i;
+  if (l <= lt)
+  {
+    if (pr == ONLY_REM) return r;
+    if (pr == ONLY_DIVIDES) return signe(x)? NULL: pol_0(varn(x));
+    if (pr) *pr = r;
+    return pol_0(varn(T));
+  }
+  if (lt <= 1)
+    return FpX_divrem_basecase(r,T,p,pr);
+  if (pr != ONLY_REM && l>lm)
+  {
+    q = cgetg(l-lt+2, t_POL);
+    for (i=0;i<l-lt;i++) gel(q+2,i) = gen_0;
+  }
+  while (l>lm)
+  {
+    GEN zr, zq = FpX_divrem_Barrettspec(r+2+l-lm,lm,mg,T,p,&zr);
+    long lz = lgpol(zr);
+    if (pr != ONLY_REM)
+    {
+      long lq = lgpol(zq);
+      for(i=0; i<lq; i++) gel(q+2+l-lm,i) = gel(zq,2+i);
+    }
+    for(i=0; i<lz; i++) gel(r+2+l-lm,i) = gel(zr,2+i);
+    l = l-lm+lz;
+  }
+  if (pr != ONLY_REM)
+  {
+    if (l > lt)
+    {
+      GEN zq = FpX_divrem_Barrettspec(r+2,l,mg,T,p,&r);
+      if (!q) q = zq;
+      else
+      {
+        long lq = lgpol(zq);
+        for(i=0; i<lq; i++) gel(q+2,i) = gel(zq,2+i);
+      }
+    }
+    else
+      r = FpX_renormalize(r, l+2);
+  }
+  else
+  {
+    if (l > lt)
+      r = FpX_divrem_Barrettspec(r+2, l, mg, T, p, ONLY_REM);
+    else
+      r = FpX_renormalize(r, l+2);
+    r[1] = x[1]; return FpX_renormalize(r, lg(r));
+  }
+  if (pr) { r[1] = x[1]; r = FpX_renormalize(r, lg(r)); }
+  q[1] = x[1]; q = FpX_renormalize(q, lg(q));
+  if (pr == ONLY_DIVIDES) return signe(r)? NULL: q;
+  if (pr) *pr = r;
+  return q;
 }
 
 GEN
-FpX_rem(GEN x, GEN y, GEN p)
+FpX_divrem(GEN x, GEN T, GEN p, GEN *pr)
 {
+  GEN B, y = get_FpX_red(T, &B);
   long dy = degpol(y), dx = degpol(x), d = dx-dy;
-  if (d < 0) return FpX_red(x,p);
-  if (d+3 < FpX_REM_MONTGOMERY_LIMIT || d>dy-2)
-    return FpX_divrem(x,y,p,ONLY_REM);
-  else
+  if (pr==ONLY_REM) return FpX_rem(x, y, p);
+  if (!B && d+3 < FpX_DIVREM_BARRETT_LIMIT)
+    return FpX_divrem_basecase(x,y,p,pr);
+  else if (lgefint(p)==3)
+  {
+    pari_sp av = avma;
+    ulong pp = to_Flxq(&x, &T, p);
+    GEN z = Flx_divrem(x, T, pp, pr);
+    if (!z) return NULL;
+    if (!pr || pr == ONLY_DIVIDES)
+      return Flx_to_ZX_inplace(gerepileuptoleaf(av, z));
+    z = Flx_to_ZX(z);
+    *pr = Flx_to_ZX(*pr);
+    gerepileall(av, 2, &z, pr);
+    return z;
+  } else
   {
     pari_sp av=avma;
-    GEN mg = FpX_invMontgomery(y, p);
-    return gerepileupto(av, FpX_rem_Montgomery(x, mg, y, p));
+    GEN mg = B? B: FpX_invBarrett(y, p);
+    GEN q1 = FpX_divrem_Barrett_noGC(x,mg,y,p,pr);
+    if (!q1) {avma=av; return NULL;}
+    if (!pr || pr==ONLY_DIVIDES) return gerepilecopy(av, q1);
+    gerepileall(av,2,&q1,pr);
+    return q1;
+  }
+}
+
+GEN
+FpX_rem(GEN x, GEN T, GEN p)
+{
+  GEN B, y = get_FpX_red(T, &B);
+  long dy = degpol(y), dx = degpol(x), d = dx-dy;
+  if (d < 0) return FpX_red(x,p);
+  if (!B && d+3 < FpX_REM_BARRETT_LIMIT)
+    return FpX_divrem_basecase(x,y,p,ONLY_REM);
+  else if (lgefint(p)==3)
+  {
+    pari_sp av = avma;
+    ulong pp = to_Flxq(&x, &T, p);
+    return Flx_to_ZX_inplace(gerepileuptoleaf(av, Flx_rem(x, T, pp)));
+  } else
+  {
+    pari_sp av = avma;
+    GEN mg = B? B: FpX_invBarrett(y, p);
+    return gerepileupto(av, FpX_divrem_Barrett_noGC(x, mg, y, p, ONLY_REM));
   }
 }
 
@@ -1027,9 +1211,9 @@ FpXQ_sqr(GEN x, GEN T, GEN p)
 /* Inverse of x in Z/pZ[X]/(pol) or NULL if inverse doesn't exist
  * return lift(1 / (x mod (p,pol))) */
 GEN
-FpXQ_invsafe(GEN x, GEN T, GEN p)
+FpXQ_invsafe(GEN x, GEN y, GEN p)
 {
-  GEN V, z = FpX_extgcd(T, x, p, NULL, &V);
+  GEN V, z = FpX_extgcd(get_FpX_mod(y), x, p, NULL, &V);
   if (degpol(z)) return NULL;
   z = Fp_invsafe(gel(z,2), p);
   if (!z) return NULL;
@@ -1041,7 +1225,7 @@ FpXQ_inv(GEN x,GEN T,GEN p)
 {
   pari_sp av = avma;
   GEN U = FpXQ_invsafe(x, T, p);
-  if (!U) pari_err(gdiver);
+  if (!U) pari_err_INV("FpXQ_inv",x);
   return gerepileupto(av, U);
 }
 
@@ -1052,58 +1236,60 @@ FpXQ_div(GEN x,GEN y,GEN T,GEN p)
   return gerepileupto(av, FpXQ_mul(x,FpXQ_inv(y,T,p),T,p));
 }
 
-static GEN
-FpXQ_mul_mg(GEN x,GEN y,GEN mg,GEN T,GEN p)
-{
-  GEN z = FpX_mul(x,y,p);
-  if (lg(T) > lg(z)) return z;
-  return FpX_rem_Montgomery(z, mg, T, p);
-}
-
-/* Square of y in Z/pZ[X]/(T), as t_VECSMALL. */
-static GEN
-FpXQ_sqr_mg(GEN y,GEN mg,GEN T,GEN p)
-{
-  GEN z = FpX_sqr(y,p);
-  if (lg(T) > lg(z)) return z;
-  return FpX_rem_Montgomery(z, mg, T, p);
-}
-
-typedef struct {
-  GEN T, p, mg;
-} FpX_muldata;
+struct _FpXQ {
+  GEN T, p, mg, aut;
+};
 
 static GEN
-_sqr_montgomery(void *data, GEN x)
+_FpXQ_add(void *data, GEN x, GEN y)
 {
-  FpX_muldata *D = (FpX_muldata*)data;
-  return FpXQ_sqr_mg(x,D->mg, D->T, D->p);
+  (void) data;
+  return ZX_add(x, y);
 }
 static GEN
-_mul_montgomery(void *data, GEN x, GEN y)
+_FpXQ_cmul(void *data, GEN P, long a, GEN x)
 {
-  FpX_muldata *D = (FpX_muldata*)data;
-  return FpXQ_mul_mg(x,y,D->mg, D->T, D->p);
+  (void) data;
+  return ZX_Z_mul(x, gel(P,a+2));
 }
-
 static GEN
 _FpXQ_sqr(void *data, GEN x)
 {
-  FpX_muldata *D = (FpX_muldata*)data;
+  struct _FpXQ *D = (struct _FpXQ*)data;
   return FpXQ_sqr(x, D->T, D->p);
 }
 static GEN
 _FpXQ_mul(void *data, GEN x, GEN y)
 {
-  FpX_muldata *D = (FpX_muldata*)data;
+  struct _FpXQ *D = (struct _FpXQ*)data;
   return FpXQ_mul(x,y, D->T, D->p);
 }
+static GEN
+_FpXQ_zero(void *data)
+{
+  struct _FpXQ *D = (struct _FpXQ*)data;
+  return pol_0(get_FpX_var(D->T));
+}
+static GEN
+_FpXQ_one(void *data)
+{
+  struct _FpXQ *D = (struct _FpXQ*)data;
+  return pol_1(get_FpX_var(D->T));
+}
+static GEN
+_FpXQ_red(void *data, GEN x)
+{
+  struct _FpXQ *D = (struct _FpXQ*)data;
+  return FpX_red(x,D->p);
+}
+
+static struct bb_algebra FpXQ_algebra = { _FpXQ_red,_FpXQ_add,_FpXQ_mul,_FpXQ_sqr,_FpXQ_one,_FpXQ_zero };
 
 /* x,pol in Z[X], p in Z, n in Z, compute lift(x^n mod (p, pol)) */
 GEN
 FpXQ_pow(GEN x, GEN n, GEN T, GEN p)
 {
-  FpX_muldata D;
+  struct _FpXQ D;
   pari_sp av;
   long s = signe(n);
   GEN y;
@@ -1113,33 +1299,33 @@ FpXQ_pow(GEN x, GEN n, GEN T, GEN p)
   av = avma;
   if (!is_bigint(p))
   {
-    ulong pp = p[2];
-    T = ZX_to_Flx(T, pp);
-    x = ZX_to_Flx(x, pp);
-    y = Flx_to_ZX( Flxq_pow(x, n, T, pp) );
+    ulong pp = to_Flxq(&x, &T, p);
+    y = Flxq_pow(x, n, T, pp);
+    return Flx_to_ZX_inplace(gerepileuptoleaf(av, y));
   }
-  else
+  if (s < 0) x = FpXQ_inv(x,T,p);
+  D.p = p; D.T = FpX_get_red(T,p);
+  y = gen_pow(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
+  return gerepileupto(av, y);
+}
+
+GEN /*Assume n is very small*/
+FpXQ_powu(GEN x, ulong n, GEN T, GEN p)
+{
+  struct _FpXQ D;
+  pari_sp av;
+  GEN y;
+  if (!n) return pol_1(varn(x));
+  if (n==1) return FpXQ_red(x,T,p);
+  av = avma;
+  if (!is_bigint(p))
   {
-    long lx = lgpol(x), lT = lgpol(T);
-    D.T = T;
-    D.p = p;
-    if (s < 0) x = FpXQ_inv(x,T,p);
-    if (lT+2>FpX_POW_MONTGOMERY_LIMIT)
-    {
-      D.mg  = FpX_invMontgomery(T,p);
-      if (lx>=lT)
-      {
-        if (lx<2*lT-2) x = FpX_rem_Montgomery(x,D.mg,T,p);
-        else x = FpX_rem(x,T,p);
-      }
-      y = gen_pow(x, n, (void*)&D, &_sqr_montgomery, &_mul_montgomery);
-    }
-    else
-    {
-      if (lx>=lT) x = FpX_rem(x,T,p);
-      y = gen_pow(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
-    }
+    ulong pp = to_Flxq(&x, &T, p);
+    y = Flxq_powu(x, n, T, pp);
+    return Flx_to_ZX_inplace(gerepileuptoleaf(av, y));
   }
+  D.T = FpX_get_red(T, p); D.p = p;
+  y = gen_powu(x, n, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul);
   return gerepileupto(av, y);
 }
 
@@ -1147,53 +1333,17 @@ FpXQ_pow(GEN x, GEN n, GEN T, GEN p)
 GEN
 FpXQ_powers(GEN x, long l, GEN T, GEN p)
 {
-  GEN V=cgetg(l+2,t_VEC);
-  long i;
-  gel(V,1) = pol_1(varn(T)); if (l==0) return V;
-  gel(V,2) = ZX_copy(x);       if (l==1) return V;
-  if (lgefint(p) == 3) {
-    long pp = p[2];
-    return FlxC_to_ZXC(Flxq_powers(ZX_to_Flx(x, pp), l, ZX_to_Flx(T,pp), pp));
+  struct _FpXQ D;
+  int use_sqr;
+  if (l>2 && lgefint(p) == 3) {
+    pari_sp av = avma;
+    ulong pp = to_Flxq(&x, &T, p);
+    GEN z = FlxV_to_ZXV(Flxq_powers(x, l, T, pp));
+    return gerepileupto(av, z);
   }
-  if (lg(T)>FpX_POW_MONTGOMERY_LIMIT)
-  {
-    GEN mg = FpX_invMontgomery(T,p);
-    gel(V,3) = FpXQ_sqr_mg(x,mg,T,p);
-    if ((degpol(x)<<1) < degpol(T)) {
-      for(i = 4; i < l+2; i++)
-        gel(V,i) = FpXQ_mul_mg(gel(V,i-1),x,mg,T,p);
-    } else { /* use squarings if degree(x) is large */
-      for(i = 4; i < l+2; i++)
-        gel(V,i) = odd(i)? FpXQ_sqr_mg(gel(V, (i+1)>>1),mg,T,p)
-                       : FpXQ_mul_mg(gel(V, i-1),x,mg,T,p);
-    }
-  } else {
-    gel(V,3) = FpXQ_sqr(x,T,p);
-    if ((degpol(x)<<1) < degpol(T)) {
-      for(i = 4; i < l+2; i++)
-        gel(V,i) = FpXQ_mul(gel(V,i-1),x,T,p);
-    } else { /* use squarings if degree(x) is large */
-      for(i = 4; i < l+2; i++)
-        gel(V,i) = odd(i)? FpXQ_sqr(gel(V, (i+1)>>1),T,p)
-                       : FpXQ_mul(gel(V, i-1),x,T,p);
-    }
-  }
-  return V;
-}
-
-/* assume T irreducible mod p */
-int
-FpXQ_issquare(GEN x, GEN T, GEN p)
-{
-  pari_sp av;
-  GEN m, z;
-  long res;
-  if (lg(x) == 2 || equalui(2, p)) return 1;
-  av = avma;
-  m = diviiexact(subis(powiu(p, degpol(T)), 1), subis(p,1));
-  z = constant_term( FpXQ_pow(x, m, T, p) );
-  res = kronecker(z, p) == 1;
-  avma = av; return res;
+  use_sqr = (degpol(x)<<1)>=get_FpX_degree(T);
+  D.T = FpX_get_red(T,p); D.p = p;
+  return gen_powers(x, l, use_sqr, (void*)&D, &_FpXQ_sqr, &_FpXQ_mul,&_FpXQ_one);
 }
 
 GEN
@@ -1202,11 +1352,193 @@ FpXQ_matrix_pow(GEN y, long n, long m, GEN P, GEN l)
   return RgXV_to_RgM(FpXQ_powers(y,m-1,P,l),n);
 }
 
+GEN
+FpX_FpXQV_eval(GEN Q, GEN x, GEN T, GEN p)
+{
+  struct _FpXQ D;
+  D.T = FpX_get_red(T,p); D.p = p;
+  return gen_bkeval_powers(Q,degpol(Q),x,(void*)&D,&FpXQ_algebra,_FpXQ_cmul);
+}
+
+GEN
+FpX_FpXQ_eval(GEN Q, GEN x, GEN T, GEN p)
+{
+  struct _FpXQ D;
+  int use_sqr;
+  if (lgefint(p) == 3)
+  {
+    pari_sp av = avma;
+    ulong pp = to_Flxq(&x, &T, p);
+    GEN z = Flx_Flxq_eval(ZX_to_Flx(Q, pp), x, T, pp);
+    return Flx_to_ZX_inplace(gerepileuptoleaf(av, z));
+  }
+  use_sqr = (degpol(x)<<1) >= get_FpX_degree(T);
+  D.T = FpX_get_red(T,p); D.p = p;
+  return gen_bkeval(Q,degpol(Q),x,use_sqr,(void*)&D,&FpXQ_algebra,_FpXQ_cmul);
+}
+
+GEN
+FpXQ_autpowers(GEN aut, long f, GEN T, GEN p)
+{
+  pari_sp av = avma;
+  long n = get_FpX_degree(T);
+  long i, nautpow = brent_kung_optpow(n-1,f-2,1);
+  long v = get_FpX_var(T);
+  GEN autpow, V;
+  T = FpX_get_red(T, p);
+  autpow = FpXQ_powers(aut, nautpow,T,p);
+  V = cgetg(f + 2, t_VEC);
+  gel(V,1) = pol_x(v); if (f==0) return gerepileupto(av, V);
+  gel(V,2) = gcopy(aut);
+  for (i = 3; i <= f+1; i++)
+    gel(V,i) = FpX_FpXQV_eval(gel(V,i-1),autpow,T,p);
+  return gerepileupto(av, V);
+}
+
+static GEN
+FpXQ_autpow_sqr(void *E, GEN x)
+{
+  struct _FpXQ *D = (struct _FpXQ*)E;
+  return FpX_FpXQ_eval(x, x, D->T, D->p);
+}
+
+static GEN
+FpXQ_autpow_mul(void *E, GEN x, GEN y)
+{
+  struct _FpXQ *D = (struct _FpXQ*)E;
+  return FpX_FpXQ_eval(x, y, D->T, D->p);
+}
+
+GEN
+FpXQ_autpow(GEN x, ulong n, GEN T, GEN p)
+{
+  struct _FpXQ D;
+  D.T = FpX_get_red(T, p); D.p = p;
+  if (n==0) return pol_x(varn(x));
+  if (n==1) return ZX_copy(x);
+  return gen_powu(x,n,(void*)&D,FpXQ_autpow_sqr,FpXQ_autpow_mul);
+}
+
+static GEN
+FpXQ_autsum_mul(void *E, GEN x, GEN y)
+{
+  struct _FpXQ *D = (struct _FpXQ*)E;
+  GEN phi1 = gel(x,1), a1 = gel(x,2);
+  GEN phi2 = gel(y,1), a2 = gel(y,2);
+  ulong d = brent_kung_optpow(maxss(degpol(phi1),degpol(a1)),2,1);
+  GEN V2 = FpXQ_powers(phi2,d,D->T,D->p);
+  GEN phi3 = FpX_FpXQV_eval(phi1,V2,D->T,D->p);
+  GEN aphi = FpX_FpXQV_eval(a1,V2,D->T,D->p);
+  GEN a3 = FpXQ_mul(aphi,a2,D->T,D->p);
+  return mkvec2(phi3, a3);
+}
+static GEN
+FpXQ_autsum_sqr(void *E, GEN x)
+{ return FpXQ_autsum_mul(E, x, x); }
+
+GEN
+FpXQ_autsum(GEN x, ulong n, GEN T, GEN p)
+{
+  struct _FpXQ D;
+  D.T = FpX_get_red(T, p); D.p = p;
+  return gen_powu(x,n,(void*)&D,FpXQ_autsum_sqr,FpXQ_autsum_mul);
+}
+
+static long
+bounded_order(GEN p, GEN b, long k)
+{
+  long i;
+  GEN a=modii(p,b);
+  for(i=1;i<k;i++)
+  {
+    if (equali1(a))
+      return i;
+    a = Fp_mul(a,p,b);
+  }
+  return 0;
+}
+
+/*
+  n = (p^d-a)\b
+  b = bb*p^vb
+  p^k = 1 [bb]
+  d = m*k+r+vb
+  u = (p^k-1)/bb;
+  v = (p^(r+vb)-a)/b;
+  w = (p^(m*k)-1)/(p^k-1)
+  n = p^r*w*u+v
+  w*u = p^vb*(p^(m*k)-1)/b
+  n = p^(r+vb)*(p^(m*k)-1)/b+(p^(r+vb)-a)/b
+*/
+
+static GEN
+FpXQ_pow_Frobenius(GEN x, GEN n, GEN aut, GEN T, GEN p)
+{
+  pari_sp av=avma;
+  long d = get_FpX_degree(T);
+  GEN an = absi(n), z, q;
+  if (cmpii(an,p)<0 || cmpis(an,d)<=0)
+    return FpXQ_pow(x, n, T, p);
+  q = powiu(p, d);
+  if (dvdii(q, n))
+  {
+    long vn = logint(an,p,NULL)-1;
+    GEN autvn = vn==1 ? aut: FpXQ_autpow(aut,vn,T,p);
+    z = FpX_FpXQ_eval(x,autvn,T,p);
+  } else
+  {
+    GEN b = diviiround(q, an), a = subii(q, mulii(an,b));
+    GEN bb, u, v, autk;
+    long vb = Z_pvalrem(b,p,&bb);
+    long m, r, k = is_pm1(bb) ? 1 : bounded_order(p,bb,d);
+    if (!k || d-vb<k) return FpXQ_pow(x,n, T, p);
+    m = (d-vb)/k; r = (d-vb)%k;
+    u = diviiexact(subis(powiu(p,k),1),bb);
+    v = diviiexact(subii(powiu(p,r+vb),a),b);
+    autk = k==1 ? aut: FpXQ_autpow(aut,k,T,p);
+    if (r)
+    {
+      GEN autr = r==1 ? aut: FpXQ_autpow(aut,r,T,p);
+      z = FpX_FpXQ_eval(x,autr,T,p);
+    } else z = x;
+    if (m > 1) z = gel(FpXQ_autsum(mkvec2(autk, z), m, T, p), 2);
+    if (!is_pm1(u)) z = FpXQ_pow(z, u, T, p);
+    if (signe(v)) z = FpXQ_mul(z, FpXQ_pow(x, v, T, p), T, p);
+  }
+  return gerepileupto(av,signe(n)>0 ? z : FpXQ_inv(z,T,p));
+}
+
+/* assume T irreducible mod p */
+int
+FpXQ_issquare(GEN x, GEN T, GEN p)
+{
+  pari_sp av;
+  long res;
+  if (lg(x) == 2 || equalui(2, p)) return 1;
+  if (lg(x) == 3) return Fq_issquare(gel(x,2), T, p);
+  /* Ng = g^((q-1)/(p-1)) */
+  av = avma; res = kronecker(FpX_resultant(T,x,p), p) == 1;
+  avma = av; return res;
+}
+int
+Fp_issquare(GEN x, GEN p)
+{
+  if (equalui(2, p)) return 1;
+  return kronecker(x, p) == 1;
+}
+/* assume T irreducible mod p */
+int
+Fq_issquare(GEN x, GEN T, GEN p)
+{
+  if (typ(x) != t_INT) return FpXQ_issquare(x, T, p);
+  return (! odd(get_FpX_degree(T)) || Fp_issquare(x, p));
+}
+
 static GEN
 famat_Z_gcd(GEN M, GEN n)
 {
   pari_sp av=avma;
-  long i, j, l=lg(M[1]);
+  long i, j, l=lgcols(M);
   GEN F=cgetg(3,t_MAT);
   gel(F,1)=cgetg(l,t_COL);
   gel(F,2)=cgetg(l,t_COL);
@@ -1237,7 +1569,7 @@ Fp_FpXQ_log(GEN a, GEN g, GEN o, GEN T, GEN p)
 
   ordp = subis(p, 1); /* even */
   ord  = dlog_get_ord(o);
-  if (!ord) ord = T? subis(powiu(p, degpol(T)), 1): ordp;
+  if (!ord) ord = T? subis(powiu(p, get_FpX_degree(T)), 1): ordp;
   if (equalii(a, ordp)) /* -1 */
     return gerepileuptoint(av, shifti(ord,-1));
   ordp = gcdii(ordp,ord);
@@ -1258,37 +1590,43 @@ Fp_FpXQ_log(GEN a, GEN g, GEN o, GEN T, GEN p)
 }
 
 static GEN
-_FpXQ_pow(void *data, GEN x, GEN y)
+_FpXQ_pow(void *data, GEN x, GEN n)
 {
-  FpX_muldata *D = (FpX_muldata*)data;
-  return FpXQ_pow(x,y, D->T, D->p);
-}
-
-static ulong
-_FpXQ_hash(GEN x)
-{
-  ulong h=0;
-  long i, l=lg(x);
-  for (i=2; i<l; i++)
-    if (signe(gel(x,i))) h ^= mod2BIL(gel(x,i));
-  return h;
+  struct _FpXQ *D = (struct _FpXQ*)data;
+  return FpXQ_pow_Frobenius(x,n, D->aut, D->T, D->p);
 }
 
 static GEN
 _FpXQ_rand(void *data)
 {
   pari_sp av=avma;
-  FpX_muldata *D = (FpX_muldata*)data;
+  struct _FpXQ *D = (struct _FpXQ*)data;
   GEN z;
   do
   {
     avma=av;
-    z=random_FpX(degpol(D->T),varn(D->T),D->p);
+    z=random_FpX(get_FpX_degree(D->T),get_FpX_var(D->T),D->p);
   } while (!signe(z));
   return z;
 }
 
-static const struct bb_group FpXQ_star={_FpXQ_mul,_FpXQ_pow,_FpXQ_rand,_FpXQ_hash,cmp_RgX,gequal1};
+static GEN
+_FpXQ_easylog(void *E, GEN a, GEN g, GEN ord)
+{
+  struct _FpXQ *s=(struct _FpXQ*) E;
+  if (degpol(a)) return NULL;
+  return Fp_FpXQ_log(constant_term(a),g,ord,s->T,s->p);
+}
+
+static const struct bb_group FpXQ_star={_FpXQ_mul,_FpXQ_pow,_FpXQ_rand,hash_GEN,ZX_equal,ZX_equal1,_FpXQ_easylog};
+
+const struct bb_group *get_FpXQ_star(void **E, GEN T, GEN p)
+{
+  GEN z = new_chunk(sizeof(struct _FpXQ));
+  struct _FpXQ *e = (struct _FpXQ *) z;
+  e->T = T; e->p  = p; e->aut =  FpXQ_pow(pol_x(get_FpX_var(T)), p, T, p);
+  *E = (void*)e; return &FpXQ_star;
+}
 
 GEN
 FpXQ_order(GEN a, GEN ord, GEN T, GEN p)
@@ -1296,82 +1634,82 @@ FpXQ_order(GEN a, GEN ord, GEN T, GEN p)
   if (lgefint(p)==3)
   {
     pari_sp av=avma;
-    ulong pp=p[2];
-    GEN z = Flxq_order(ZX_to_Flx(a,pp),ord,ZX_to_Flx(T,pp),pp);
+    ulong pp = to_Flxq(&a, &T, p);
+    GEN z = Flxq_order(a, ord, T, pp);
     return gerepileuptoint(av,z);
   }
   else
   {
-    FpX_muldata s;
-    s.T=T; s.p=p;
-    return gen_eltorder(a,ord, (void*)&s,&FpXQ_star);
+    void *E;
+    const struct bb_group *S = get_FpXQ_star(&E,T,p);
+    return gen_order(a,ord,E,S);
   }
 }
 
-static GEN
-_FpXQ_easylog(void *E, GEN a, GEN g, GEN ord)
-{
-  FpX_muldata *s=(FpX_muldata*) E;
-  if (degpol(a)) return NULL;
-  return Fp_FpXQ_log(constant_term(a),g,ord,s->T,s->p);
-}
-
 GEN
 FpXQ_log(GEN a, GEN g, GEN ord, GEN T, GEN p)
 {
   if (lgefint(p)==3)
   {
     pari_sp av=avma;
-    ulong pp=p[2];
-    GEN z = Flxq_log(ZX_to_Flx(a,pp),ZX_to_Flx(g,pp),ord,ZX_to_Flx(T,pp),pp);
+    ulong pp = to_Flxq(&a, &T, p);
+    GEN z = Flxq_log(a, ZX_to_Flx(g, pp), ord, T, pp);
     return gerepileuptoint(av,z);
   }
   else
   {
-    FpX_muldata s;
-    s.T=T; s.p=p;
-    return gen_PH_log(a,g,ord, (void*)&s,&FpXQ_star,_FpXQ_easylog);
+    void *E;
+    const struct bb_group *S = get_FpXQ_star(&E,T,p);
+    GEN z = gen_PH_log(a,g,ord,E,S);
+    return z? z: cgetg(1,t_VEC);
   }
 }
 
 GEN
 FpXQ_sqrtn(GEN a, GEN n, GEN T, GEN p, GEN *zeta)
 {
+  pari_sp av = avma;
+  GEN z;
   if (!signe(a))
   {
-    long v=varn(T);
-    if (zeta)
-      *zeta=pol_1(v);
+    long v=varn(a);
+    if (signe(n) < 0) pari_err_INV("FpXQ_sqrtn",a);
+    if (zeta) *zeta=pol_1(v);
     return pol_0(v);
   }
   if (lgefint(p)==3)
   {
-    pari_sp av=avma;
-    ulong pp=p[2];
-    GEN z = Flxq_sqrtn(ZX_to_Flx(a,pp),n,ZX_to_Flx(T,pp),pp,zeta);
+    ulong pp = to_Flxq(&a, &T, p);
+    z = Flxq_sqrtn(a, n, T, pp, zeta);
     if (!z) return NULL;
+    if (!zeta) return Flx_to_ZX_inplace(gerepileuptoleaf(av, z));
     z = Flx_to_ZX(z);
-    if (zeta)
-    {
-      *zeta=Flx_to_ZX(*zeta);
-      gerepileall(av,2,&z,zeta);
-      return z;
-    }
-    else return gerepileupto(av, z);
+    *zeta=Flx_to_ZX(*zeta);
   }
   else
   {
-    FpX_muldata s;
-    s.T=T; s.p=p;
-    return gen_Shanks_sqrtn(a,n,addis(powiu(p,degpol(T)),-1),zeta,
-                            (void*)&s,&FpXQ_star);
+    void *E;
+    const struct bb_group *S = get_FpXQ_star(&E,T,p);
+    GEN o = addis(powiu(p,get_FpX_degree(T)),-1);
+    z = gen_Shanks_sqrtn(a,n,o,zeta,E,S);
+    if (!z) return NULL;
+    if (!zeta) return gerepileupto(av, z);
   }
+  gerepileall(av, 2, &z,zeta);
+  return z;
 }
 
 GEN
-FpXQ_norm(GEN x, GEN T, GEN p)
+FpXQ_sqrt(GEN a, GEN T, GEN p)
+{
+  return FpXQ_sqrtn(a, gen_2, T, p, NULL);
+}
+
+GEN
+FpXQ_norm(GEN x, GEN TB, GEN p)
 {
   pari_sp av = avma;
+  GEN T = get_FpX_mod(TB);
   GEN y = FpX_resultant(T, x, p);
   GEN L = leading_term(T);
   if (gequal1(L) || signe(x)==0) return y;
@@ -1379,21 +1717,25 @@ FpXQ_norm(GEN x, GEN T, GEN p)
 }
 
 GEN
-FpXQ_trace(GEN x, GEN T, GEN p)
+FpXQ_trace(GEN x, GEN TB, GEN p)
 {
   pari_sp av = avma;
-  GEN z = FpX_mul(x, FpX_deriv(T, p), p);
-  z = FpX_div(RgX_shift_shallow(z, 1), T, p);
-  return gerepileuptoint(av, constant_term(z));
+  GEN T = get_FpX_mod(TB);
+  GEN dT = FpX_deriv(T,p);
+  long n = degpol(dT);
+  GEN z = FpXQ_mul(x, dT, TB, p);
+  if (degpol(z)<n) { avma = av; return gen_0; }
+  return gerepileuptoint(av, Fp_div(gel(z,2+n), gel(T,3+n),p));
 }
 
 GEN
 FpXQ_charpoly(GEN x, GEN T, GEN p)
 {
   pari_sp ltop=avma;
-  long v=varn(T);
+  long v;
   GEN R;
-  T = leafcopy(T); setvarn(T, MAXVARN);
+  T = leafcopy(get_FpX_mod(T));
+  v = varn(T); setvarn(T, MAXVARN);
   x = leafcopy(x); setvarn(x, MAXVARN);
   R = FpX_FpXY_resultant(T, deg1pol_shallow(gen_1,FpX_neg(x,p),v),p);
   return gerepileupto(ltop,R);
@@ -1421,7 +1763,7 @@ FpXQ_conjvec(GEN x, GEN T, GEN p)
 {
   pari_sp av=avma;
   long i;
-  long n = degpol(T), v = varn(T);
+  long n = get_FpX_degree(T), v = varn(x);
   GEN M = FpXQ_matrix_pow(FpXQ_pow(pol_x(v),p,T,p),n,n,T,p);
   GEN z = cgetg(n+1,t_COL);
   gel(z,1) = RgX_to_RgV(x,n);
@@ -1431,73 +1773,137 @@ FpXQ_conjvec(GEN x, GEN T, GEN p)
   return gerepilecopy(av,z);
 }
 
+/* p prime, p_1 = p-1, q = p^deg T, Lp = cofactors of some prime divisors
+ * l_p of p-1, Lq = cofactors of some prime divisors l_q of q-1, return a
+ * g in Fq such that
+ * - Ng generates all l_p-Sylows of Fp^*
+ * - g generates all l_q-Sylows of Fq^* */
+static GEN
+gener_FpXQ_i(GEN T, GEN p, GEN p_1, GEN Lp, GEN Lq)
+{
+  pari_sp av;
+  long vT = varn(T), f = degpol(T), l = lg(Lq);
+  GEN F = FpXQ_pow(pol_x(vT), p, T, p); /* Frobenius */
+  int p_is_2 = is_pm1(p_1);
+  for (av = avma;; avma = av)
+  {
+    GEN t, g = random_FpX(f, vT, p);
+    long i;
+    if (degpol(g) < 1) continue;
+    if (p_is_2)
+      t = g;
+    else
+    {
+      t = FpX_resultant(T, g, p); /* Ng = g^((q-1)/(p-1)), assuming T monic */
+      if (kronecker(t, p) == 1) continue;
+      if (lg(Lp) > 1 && !is_gener_Fp(t, p, p_1, Lp)) continue;
+      t = FpXQ_pow(g, shifti(p_1,-1), T, p);
+    }
+    for (i = 1; i < l; i++)
+    {
+      GEN a = FpXQ_pow_Frobenius(t, gel(Lq,i), F, T, p);
+      if (!degpol(a) && equalii(gel(a,2), p_1)) break;
+    }
+    if (i == l) return g;
+  }
+}
+
 GEN
 gener_FpXQ(GEN T, GEN p, GEN *po)
 {
-  long i, j, vT = varn(T), f = degpol(T);
-  GEN g, L, L2, p_1, q, o;
-  pari_sp av0 = avma, av;
+  long i, j, f = get_FpX_degree(T);
+  GEN g, Lp, Lq, p_1, q_1, N, o;
+  pari_sp av = avma;
 
+  p_1 = subiu(p,1);
   if (f == 1) {
-    GEN L, fa;
-    o = subis(p, 1);
+    GEN Lp, fa;
+    o = p_1;
     fa = Z_factor(o);
-    L = gel(fa,1);
-    L = vecslice(L, 2, lg(L)-1); /* remove 2 for efficiency */
+    Lp = gel(fa,1);
+    Lp = vecslice(Lp, 2, lg(Lp)-1); /* remove 2 for efficiency */
 
     g = cgetg(3, t_POL);
-    g[1] = evalsigne(1) | evalvarn(vT);
-    gel(g,2) = pgener_Fp_local(p, L);
+    g[1] = evalsigne(1) | evalvarn(get_FpX_var(T));
+    gel(g,2) = pgener_Fp_local(p, Lp);
     if (po) *po = mkvec2(o, fa);
     return g;
   }
   if (lgefint(p) == 3)
   {
-    ulong pp = (ulong)p[2];
-    g = gener_Flxq(ZX_to_Flx(T, pp), pp, po);
+    ulong pp = to_Flxq(NULL, &T, p);
+    g = gener_Flxq(T, pp, po);
+    if (!po) return Flx_to_ZX_inplace(gerepileuptoleaf(av, g));
     g = Flx_to_ZX(g);
-    if (!po) g = gerepileupto(av0, g);
-    else
-    {
-      gel(*po,2) = Flx_to_ZX(gel(*po,2));
-      gerepileall(av0, 2, &g, po);
-    }
+    gel(*po,2) = Flx_to_ZX(gel(*po,2));
+    gerepileall(av, 2, &g, po);
     return g;
   }
-  p_1 = subis(p,1);
-  q = diviiexact(subis(powiu(p,f), 1), p_1);
-
-  L = NULL;
-  (void)Z_lvalrem(p_1, 2, &L);
-  L = gel(Z_factor(L),1);
-  for (i=lg(L)-1; i; i--) gel(L,i) = diviiexact(p_1, gel(L,i));
+  /* p now odd */
+  q_1 = subiu(powiu(p,f), 1);
+  N = diviiexact(q_1, p_1);
+  Lp = odd_prime_divisors(p_1);
+  for (i=lg(Lp)-1; i; i--) gel(Lp,i) = diviiexact(p_1, gel(Lp,i));
   o = factor_pn_1(p,f);
-  L2 = leafcopy( gel(o, 1) );
-  for (i = j = 1; i < lg(L2); i++)
+  Lq = leafcopy( gel(o, 1) );
+  for (i = j = 1; i < lg(Lq); i++)
   {
-    if (remii(p_1, gel(L2,i)) == gen_0) continue;
-    gel(L2,j++) = diviiexact(q, gel(L2,i));
+    if (remii(p_1, gel(Lq,i)) == gen_0) continue;
+    gel(Lq,j++) = diviiexact(N, gel(Lq,i));
   }
-  setlg(L2, j);
-  for (av = avma;; avma = av)
-  {
-    GEN t;
-    g = random_FpX(f, vT, p);
-    if (degpol(g) < 1) continue;
-    t = FpX_resultant(T, g, p); /* Ng = g^q, assuming T is monic */
-    if (equali1(t) || !is_gener_Fp(t, p, p_1, L)) continue;
-    t = FpXQ_pow(g, shifti(p_1,-1), T, p);
-    for (i = 1; i < j; i++)
-    {
-      GEN a = FpXQ_pow(t, gel(L2,i), T, p);
-      if (!degpol(a) && equalii(gel(a,2), p_1)) break;
-    }
-    if (i == j) break;
-  }
-  if (!po) g = gerepilecopy(av0, g);
+  setlg(Lq, j);
+  g = gener_FpXQ_i(get_FpX_mod(T), p, p_1, Lp, Lq);
+  if (!po) g = gerepilecopy(av, g);
   else {
-    *po = mkvec2(subis(powiu(p,f), 1), o);
-    gerepileall(av0, 2, &g, po);
+    *po = mkvec2(q_1, o);
+    gerepileall(av, 2, &g, po);
   }
   return g;
 }
+
+#if 0 /* generic version: slower */
+GEN
+gener_FpXQ2(GEN T, GEN p, GEN *po)
+{
+  pari_sp av = avma;
+  void *E;
+  long f = get_FpX_degree(T);
+  GEN g, o = factor_pn_1(p,f);
+  const struct bb_group *S = get_FpXQ_star(&E,T,p);
+  g = gen_gener(o,E,S);
+  if (!po) g = gerepilecopy(av, g);
+  else {
+    *po = mkvec2(powiu(p,f), o);
+    gerepileall(av, 2, &g, po);
+  }
+  return g;
+}
+#endif
+
+GEN
+gener_FpXQ_local(GEN T, GEN p, GEN L)
+{
+  GEN Lp, Lq, p_1 = subiu(p,1), q_1, N, Q;
+  long f, i, ip, iq, l = lg(L);
+  T = get_FpX_mod(T);
+  f = degpol(T);
+  q_1 = subiu(powiu(p,f), 1);
+  N = diviiexact(q_1, p_1);
+
+  Q = is_pm1(p_1)? gen_1: shifti(p_1,-1);
+  Lp = cgetg(l, t_VEC); ip = 1;
+  Lq = cgetg(l, t_VEC); iq = 1;
+  for (i=1; i < l; i++)
+  {
+    GEN a, b, ell = gel(L,i);
+    if (equaliu(ell,2)) continue;
+    a = dvmdii(Q, ell, &b);
+    if (b == gen_0)
+      gel(Lp,ip++) = a;
+    else
+      gel(Lq,iq++) = diviiexact(N,ell);
+  }
+  setlg(Lp, ip);
+  setlg(Lq, iq);
+  return gener_FpXQ_i(T, p, p_1, Lp, Lq);
+}
