@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -11,7 +9,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 #include "pari.h"
 #include "paripriv.h"
 
@@ -55,7 +53,7 @@ ZX_divides_i(GEN x, GEN y, GEN B)
     gel(z,i-dy) = p1;
   }
   av = avma;
-  for (;; i--)
+  for (; i >= 0; i--)
   {
     p1 = gel(x,i);
     /* we always enter this loop at least once */
@@ -63,7 +61,6 @@ ZX_divides_i(GEN x, GEN y, GEN B)
       p1 = subii(p1, mulii(gel(z,j),gel(y,i-j)));
     if (signe(p1)) return NULL;
     avma = av;
-    if (!i) break;
   }
   return z - 2;
 }
@@ -83,7 +80,7 @@ two_factor_bound(GEN x)
   GEN *invbin, c, r = cgetr(3), z;
 
   x += 2; invbin = (GEN*)new_chunk(n+1);
-  z = real_1(3); /* invbin[i] = 1 / binomial(n, i) */
+  z = real_1(LOWDEFAULTPREC); /* invbin[i] = 1 / binomial(n, i) */
   for (i=0,j=n; j >= i; i++,j--)
   {
     invbin[i] = invbin[j] = z;
@@ -178,13 +175,13 @@ cmbf(GEN pol, GEN famod, GEN bound, GEN p, long a, long b,
   long K = 1, cnt = 1, i,j,k, curdeg, lfamod = lg(famod)-1;
   ulong spa_b, spa_bs2, Sbound;
   GEN lc, lcpol, pa = powiu(p,a), pas2 = shifti(pa,-1);
-  uGEN trace1   = (uGEN)cgetg(lfamod+1, t_VECSMALL);
-  uGEN trace2   = (uGEN)cgetg(lfamod+1, t_VECSMALL);
+  GEN trace1   = cgetg(lfamod+1, t_VECSMALL);
+  GEN trace2   = cgetg(lfamod+1, t_VECSMALL);
   GEN ind      = cgetg(lfamod+1, t_VECSMALL);
   GEN deg      = cgetg(lfamod+1, t_VECSMALL);
   GEN degsofar = cgetg(lfamod+1, t_VECSMALL);
-  GEN listmod  = cgetg(lfamod+1, t_COL);
-  GEN fa       = cgetg(lfamod+1, t_COL);
+  GEN listmod  = cgetg(lfamod+1, t_VEC);
+  GEN fa       = cgetg(lfamod+1, t_VEC);
 
   *pmaxK = cmbf_maxK(lfamod);
   lc = absi(leading_term(pol));
@@ -304,7 +301,7 @@ nextK:
         if (j <= K && i == ind[j]) j++;
         else
         {
-          famod[k] = famod[i];
+          gel(famod,k) = gel(famod,i);
           trace1[k] = trace1[i];
           trace2[k] = trace2[i];
           deg[k] = deg[i]; k++;
@@ -367,52 +364,6 @@ factor_quad(GEN x, GEN res, long *ptcnt)
   gel(res,cnt++) = gmul(u, gsub(pol_x(v), z2)); *ptcnt = cnt;
 }
 
-/* y > 1 and B > 0 integers. Return e such that y^(e-1) <= B < y^e, i.e
- * e = 1 + floor(log_y B). Set *ptq = y^e if non-NULL */
-long
-logint(GEN B, GEN y, GEN *ptq)
-{
-  pari_sp av = avma;
-  long e,i,fl;
-  GEN q,pow2, r = y;
-
-  if (typ(B) != t_INT) B = ceil_safe(B);
-  if (expi(B) <= (expi(y) << 6)) /* e small, be naive */
-  {
-    for (e=1;; e++)
-    { /* here, r = y^e */
-      fl = cmpii(r, B);
-      if (fl > 0) goto END;
-      r = mulii(r,y);
-    }
-  }
-  /* binary splitting: compute bits of e one by one */
-  /* compute pow2[i] = y^(2^i) [i < very crude upper bound for log_2(n)] */
-  pow2 = new_chunk(bit_accuracy(lgefint(B)));
-  gel(pow2,0) = y;
-  for (i=0,q=r;; )
-  {
-    fl = cmpii(r,B); if (fl >= 0) break;
-    q = r; r = sqri(q);
-    i++; gel(pow2,i) = r;
-  }
-  if (i == 0) { e = 1; goto END; } /* y <= B */
-
-  for (i--, e=1L<<i;;)
-  { /* y^e = q < B <= r = q * y^(2^i) */
-    if (!fl) break; /* B = r */
-    /* q < B < r */
-    if (--i < 0) { if (fl > 0) e++; break; }
-    r = mulii(q, gel(pow2,i));
-    fl = cmpii(r, B);
-    if (fl <= 0) { e += (1L<<i); q = r; }
-  }
-  if (fl <= 0) { e++; r = mulii(r,y); }
-END:
-  if (ptq) *ptq = gerepileuptoint(av, icopy(r)); else avma = av;
-  return e;
-}
-
 /* recombination of modular factors: van Hoeij's algorithm */
 
 /* Q in Z[X], return Q(2^n) */
@@ -465,7 +416,7 @@ GEN
 special_pivot(GEN x)
 {
   GEN t, perm, H = ZM_hnfperm(x,NULL,&perm);
-  long i,j, l = lg(H), h = lg(H[1]);
+  long i,j, l = lg(H), h = lgcols(H);
   for (i=1; i<h; i++)
   {
     int fl = 0;
@@ -488,7 +439,7 @@ chk_factors_get(GEN lt, GEN famod, GEN c, GEN T, GEN N)
   long i = 1, j, l = lg(famod);
   GEN V = cgetg(l, t_VEC);
   for (j = 1; j < l; j++)
-    if (signe(c[j])) V[i++] = famod[j];
+    if (signe(gel(c,j))) gel(V,i++) = gel(famod,j);
   if (lt && i > 1) gel(V,1) = RgX_Rg_mul(gel(V,1), lt);
   setlg(V, i);
   return T? FpXQXV_prod(V, T, N): FpXV_prod(V,N);
@@ -505,7 +456,7 @@ chk_factors(GEN P, GEN M_L, GEN bound, GEN famod, GEN pa)
   if (DEBUGLEVEL>7) err_printf("special_pivot output:\n%Ps\n",piv);
 
   r  = lg(piv)-1;
-  list = cgetg(r+1, t_COL);
+  list = cgetg(r+1, t_VEC);
   lt = absi(leading_term(pol));
   if (is_pm1(lt)) lt = NULL;
   ltpol = lt? ZX_Z_mul(pol, lt): pol;
@@ -548,7 +499,7 @@ LLL_check_progress(GEN Bnorm, long n0, GEN m, int final, long *ti_LLL)
   for (i=1; i<=R; i++) setlg(u[i], n0+1);
   if (R <= 1)
   {
-    if (!R) pari_err(bugparier,"LLL_cmbf [no factor]");
+    if (!R) pari_err_BUG("LLL_cmbf [no factor]");
     return NULL; /* irreducible */
   }
   setlg(u, R+1); return u;
@@ -630,7 +581,7 @@ LLL_cmbf(GEN P, GEN famod, GEN p, GEN pa, GEN bound, long a, long rec)
       GEN p2 = polsym_gen(gel(famod,i), gel(TT,i), tnew, NULL, pa);
       gel(TT,i) = p2;
       p2 += 1+tmax; /* ignore traces number 0...tmax */
-      for (j=1; j<=N0; j++) p1[j] = p2[j];
+      for (j=1; j<=N0; j++) gel(p1,j) = gel(p2,j);
       if (lP)
       { /* make Newton sums integral */
         GEN lPpow = powiu(lP, tmax);
@@ -682,7 +633,7 @@ AGAIN:
     if (DEBUGLEVEL>2)
       err_printf("LLL_cmbf: (a,b) =%4ld,%4ld; r =%3ld -->%3ld, time = %ld\n",
                  a,b, lg(m)-1, CM_L? lg(CM_L)-1: 1, timer_delay(&TI));
-    if (!CM_L) { list = mkcol(P); break; }
+    if (!CM_L) { list = mkvec(P); break; }
     if (b > bmin)
     {
       CM_L = gerepilecopy(av2, CM_L);
@@ -777,7 +728,7 @@ combine_factors(GEN target, GEN famod, GEN p, long klim)
   res     = gel(L,1);
   listmod = gel(L,2); l = lg(listmod)-1;
   famod = gel(listmod,l);
-  if (maxK >= 0 && lg(famod)-1 > 2*maxK)
+  if (maxK > 0 && lg(famod)-1 > 2*maxK)
   {
     if (l!=1) A = factor_bound(gel(res,l));
     if (DEBUGLEVEL > 4) err_printf("last factor still to be checked\n");
@@ -789,42 +740,97 @@ combine_factors(GEN target, GEN famod, GEN p, long klim)
   return res;
 }
 
-/* assume pol(0) != 0, polp = pol/lc(pol) mod p.
- * Return vector of rational roots of a */
-GEN
-DDF_roots(GEN pol, GEN polp, GEN p)
+/* Assume 'a' a squarefree ZX; return 0 if no root (fl=1) / irreducible (fl=0).
+ * Otherwise return prime p such that a mod p has fewest roots / factors */
+static ulong
+pick_prime(GEN a, long fl, pari_timer *T)
 {
-  GEN lc, lcpol, z, pe, pes2, bound;
-  long i, m, e, lz, v = varn(pol);
+  pari_sp av = avma, av1;
+  const long MAXNP = 7, da = degpol(a);
+  long nmax = da+1, np;
+  ulong chosenp = 0;
+  GEN lead = gel(a,da+2);
+  forprime_t S;
+  if (equali1(lead)) lead = NULL;
+  u_forprime_init(&S, 2, ULONG_MAX);
+  av1 = avma;
+  for (np = 0; np < MAXNP; avma = av1)
+  {
+    ulong p = u_forprime_next(&S);
+    long nfacp;
+    GEN z;
+
+    if (!p) pari_err_OVERFLOW("DDF [out of small primes]");
+    if (lead && !umodiu(lead,p)) continue;
+    z = ZX_to_Flx(a, p);
+    if (!Flx_is_squarefree(z, p)) continue;
+
+    if (fl)
+    {
+      nfacp = Flx_nbroots(z, p);
+      if (!nfacp) { chosenp = 0; break; } /* no root */
+    }
+    else
+    {
+      nfacp = Flx_nbfact(z, p);
+      if (nfacp == 1) { chosenp = 0; break; } /* irreducible */
+    }
+    if (DEBUGLEVEL>4)
+      err_printf("...tried prime %3lu (%-3ld %s). Time = %ld\n",
+                  p, nfacp, fl? "roots": "factors", timer_delay(T));
+    if (nfacp < nmax)
+    {
+      nmax = nfacp; chosenp = p;
+      if (da > 100 && nmax < 5) break; /* large degree, few factors. Enough */
+    }
+    np++;
+  }
+  avma = av; return chosenp;
+}
+
+/* Assume pol squarefree mod p; return vector of rational roots of a */
+static GEN
+DDF_roots(GEN A)
+{
+  GEN p, Ap, lc, lcpol, z, pe, pes2, bound;
+  long i, m, e, lz, v = varn(A);
+  ulong pp;
   pari_sp av, lim;
   pari_timer T;
 
   if (DEBUGLEVEL>2) timer_start(&T);
-  lc = absi(leading_term(pol));
-  if (is_pm1(lc)) lc = NULL;
-  lcpol = lc? ZX_Z_mul(pol, lc): pol;
-
-  bound = root_bound(pol);
-  if (lc) bound = mulii(lc, bound);
-  e = logint(addis(shifti(bound, 1), 1), p, &pe);
+  pp = pick_prime(A, 1, &T);
+  if (!pp) return cgetg(1,t_VEC); /* no root */
+  p = utoipos(pp);
+  lc = leading_term(A);
+  if (is_pm1(lc))
+  { lc = NULL; lcpol = A; }
+  else
+  { lc = absi(lc); lcpol = ZX_Z_mul(A, lc); }
+  Ap = ZX_to_Flx(A, pp);
+  bound = root_bound(A);
+  if (lc) { Ap = Flx_normalize(Ap, pp); bound = mulii(lc, bound); }
+  e = logint(addiu(shifti(bound, 1), 1), p, &pe);
   pes2 = shifti(pe, -1);
   if (DEBUGLEVEL>2) timer_printf(&T, "Root bound");
 
   av = avma; lim = stack_lim(av,2);
-  z = FpX_roots(polp, p);
+  z = Flx_roots(Ap, pp);
   lz = lg(z)-1;
-  if (lz > (degpol(pol) >> 2))
+  if (lz > (degpol(A) >> 2))
   { /* many roots */
-    z = shallowconcat(deg1_from_roots(z, v),
-                 FpX_div(polp, FpV_roots_to_pol(z, p, v), p));
-    z = ZpX_liftfact(pol, z, NULL, p, e, pe);
+    GEN Bp = Flx_div(Ap, Flv_roots_to_pol(z, pp, v), pp);
+    z = Flv_to_ZV(z);
+    z = shallowconcat(deg1_from_roots(z, v), Flx_to_ZX(Bp));
+    z = ZpX_liftfact(A, z, NULL, p, e, pe);
   }
   else
   {
-    z = ZpX_liftroots(pol, z, p, e);
+    z = Flv_to_ZV(z);
+    z = ZpX_liftroots(A, z, p, e);
     z = deg1_from_roots(z, v);
   }
-  if (DEBUGLEVEL>2) timer_printf(&T, "Hensel lift (mod %Ps^%ld)", p,e);
+  if (DEBUGLEVEL>2) timer_printf(&T, "Hensel lift (mod %lu^%ld)", pp,e);
 
   for (m=1, i=1; i <= lz; i++)
   {
@@ -833,19 +839,19 @@ DDF_roots(GEN pol, GEN polp, GEN p)
     y = centermod_i(y, pe, pes2);
     if (! (q = ZX_divides(lcpol, y)) ) continue;
 
-    lcpol = pol = q;
+    lcpol = A = q;
     r = negi( constant_term(y) );
     if (lc) {
       r = gdiv(r,lc);
-      pol = Q_primpart(pol);
-      lc = absi( leading_term(pol) );
-      if (is_pm1(lc)) lc = NULL; else lcpol = ZX_Z_mul(pol, lc);
+      A = Q_primpart(A);
+      lc = absi( leading_term(A) );
+      if (is_pm1(lc)) lc = NULL; else lcpol = ZX_Z_mul(A, lc);
     }
     gel(z,m++) = r;
     if (low_stack(lim, stack_lim(av,2)))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"DDF_roots, m = %ld", m);
-      gerepileall(av, lc? 4:2, &z, &pol, &lc, &lcpol);
+      gerepileall(av, lc? 4:2, &z, &A, &lc, &lcpol);
 
     }
   }
@@ -853,62 +859,30 @@ DDF_roots(GEN pol, GEN polp, GEN p)
   z[0] = evaltyp(t_VEC) | evallg(m); return z;
 }
 
-/* Assume a squarefree, degree(a) > 0, a(0) != 0.
- * If fl != 0 look only for rational roots */
+/* Assume a squarefree ZX, deg(a) > 0, return rational factors.
+ * In fact, a(0) != 0 but we don't use this */
 static GEN
-DDF(GEN a, int fl)
+DDF(GEN a)
 {
-  GEN lead, prime, famod, z, ap;
-  const long da = degpol(a);
-  long chosenp, p, nfacp, np, nmax, ti = 0;
-  pari_sp av = avma, av1;
-  byteptr pt = diffptr;
-  const long MAXNP = 7;
+  GEN ap, prime, famod, z;
+  long ti = 0;
+  ulong p = 0;
+  pari_sp av = avma;
   pari_timer T, T2;
 
   if (DEBUGLEVEL>2) { timer_start(&T); timer_start(&T2); }
-  nmax = da+1;
-  chosenp = 0;
-  lead = gel(a,da+2); if (equali1(lead)) lead = NULL;
-  av1 = avma;
-  for (p = np = 0; np < MAXNP; avma = av1)
-  {
-    NEXT_PRIME_VIADIFF_CHECK(p,pt);
-    if (lead && !smodis(lead,p)) continue;
-    z = ZX_to_Flx(a, p);
-    if (!Flx_is_squarefree(z, p)) continue;
-
-    nfacp = fl? Flx_nbroots(z, p): Flx_nbfact(z, p);
-    if (DEBUGLEVEL>4)
-      err_printf("...tried prime %3ld (%-3ld %s). Time = %ld\n",
-                  p, nfacp, fl?"roots": "factors", timer_delay(&T2));
-    if (nfacp < nmax)
-    {
-      if (nfacp <= 1)
-      {
-        if (!fl) { avma = av; return mkcol(a); } /* irreducible */
-        if (!nfacp) return cgetg(1, t_VEC); /* no root */
-      }
-      nmax = nfacp; chosenp = p;
-      if (da > 100 && nmax < 5) break; /* large degree, few factors. Enough */
-    }
-    np++;
-  }
-  prime = utoipos(chosenp);
-  ap = lead? FpX_normalize(a, prime): FpX_red(a, prime);
-  if (fl) return gerepilecopy(av, DDF_roots(a, ap, prime));
-
-  famod = cgetg(nmax+1,t_COL);
-  gel(famod,1) = ap;
-  if (nmax != FpX_split_Berlekamp((GEN*)(famod+1), prime))
-    pari_err(bugparier,"DDF: wrong numbers of factors");
+  p = pick_prime(a, 0, &T2);
+  if (!p) return mkvec(a);
+  prime = utoipos(p);
+  ap = Flx_normalize(ZX_to_Flx(a, p), p);
+  famod = gel(Flx_factor(ap, p), 1);
   if (DEBUGLEVEL>2)
   {
-    if (DEBUGLEVEL>4) timer_printf(&T2, "splitting mod p = %ld", chosenp);
+    if (DEBUGLEVEL>4) timer_printf(&T2, "splitting mod p = %lu", p);
     ti = timer_delay(&T);
     err_printf("Time setup: %ld\n", ti);
   }
-  z = combine_factors(a, famod, prime, da-1);
+  z = combine_factors(a, FlxV_to_ZXV(famod), prime, degpol(a)-1);
   if (DEBUGLEVEL>2)
     err_printf("Total Time: %ld\n===========\n", ti + timer_delay(&T));
   return gerepilecopy(av, z);
@@ -922,7 +896,7 @@ ZX_DDF(GEN x)
   GEN L;
   long m;
   x = RgX_deflate_max(x, &m);
-  L = DDF(x, 0);
+  L = DDF(x);
   if (m > 1)
   {
     GEN e, v, fa = factoru(m);
@@ -938,7 +912,7 @@ ZX_DDF(GEN x)
     {
       GEN L2 = cgetg(1,t_VEC);
       for (i=1; i < lg(L); i++)
-              L2 = shallowconcat(L2, DDF(RgX_inflate(gel(L,i), v[k]), 0));
+              L2 = shallowconcat(L2, DDF(RgX_inflate(gel(L,i), v[k])));
       L = L2;
     }
   }
@@ -950,8 +924,8 @@ ZX_DDF(GEN x)
 GEN
 ZX_squff(GEN f, GEN *ex)
 {
-  GEN T, V, W, P, e;
-  long i, k, dW, n, val;
+  GEN T, V, P, e;
+  long i, k, n, val;
 
   if (signe(leading_term(f)) < 0) f = gneg_i(f);
   val = ZX_valrem(f, &f);
@@ -962,11 +936,24 @@ ZX_squff(GEN f, GEN *ex)
   T = ZX_gcd_all(f, ZX_deriv(f), &V);
   for (k=i=1;; k++)
   {
-    W = ZX_gcd_all(T,V, &T); dW = degpol(W);
+    pari_sp av = avma;
+    GEN W = ZX_gcd_all(T,V, &T);
+    long dW = degpol(W);
     /* W = prod P^e, e > k; V = prod P^e, e >= k */
-    if (dW != degpol(V)) { gel(P,i) = Q_primpart(RgX_div(V,W)); e[i] = k; i++; }
-    if (dW <= 0) break;
-    V = W;
+    if (dW == degpol(V)) /* V | T */
+    {
+      GEN U;
+      if (!dW) { avma = av; break; }
+      while ( (U = ZX_divides(T, V)) ) { k++; T = U; }
+      T = gerepileupto(av, T);
+    }
+    else
+    {
+      gel(P,i) = Q_primpart(RgX_div(V,W));
+      e[i] = k; i++;
+      if (!dW) break;
+      V = W;
+    }
   }
   if (val) { gel(P,i) = pol_x(varn(f)); e[i] = val; i++;}
   setlg(P,i);
@@ -1001,13 +988,13 @@ ZX_factor_i(GEN x)
   GEN fa,ex,y;
   long n,i,l;
 
-  if (!signe(x)) pari_err(zeropoler,"ZX_factor");
+  if (!signe(x)) return prime_fact(x);
   fa = ZX_squff(x, &ex);
   l = lg(fa); n = 0;
   for (i=1; i<l; i++)
   {
     gel(fa,i) = ZX_DDF(gel(fa,i));
-    n += lg(fa[i])-1;
+    n += lg(gel(fa,i))-1;
   }
   y = fact_from_DDF(fa,ex,n);
   return sort_factor_pol(y, cmpii);
@@ -1046,15 +1033,15 @@ nfrootsQ(GEN x)
   GEN z;
   long val;
 
-  if (typ(x)!=t_POL) pari_err(notpoler,"nfrootsQ");
-  if (!signe(x)) pari_err(zeropoler,"nfrootsQ");
+  if (typ(x)!=t_POL) pari_err_TYPE("nfrootsQ",x);
+  if (!signe(x)) pari_err_ROOTS0("nfrootsQ");
   x = Q_primpart(x);
-  if (!RgX_is_ZX(x)) pari_err(typeer,"nfrootsQ");
+  if (!RgX_is_ZX(x)) pari_err_TYPE("nfrootsQ",x);
   val = ZX_valrem(x, &x);
   (void)ZX_gcd_all(x, ZX_deriv(x), &x);
-  z = DDF(x, 1);
+  z = DDF_roots(x);
   if (val) z = shallowconcat(z, gen_0);
-  return gerepilecopy(av, z);
+  return gerepileupto(av, sort(z));
 }
 
 /************************************************************************
@@ -1091,11 +1078,11 @@ maxnorm(GEN p)
 GEN
 ZX_gcd_all(GEN A, GEN B, GEN *Anew)
 {
-  GEN R, a, b, q, qp, H, Hp, g, Ag, Bg;
+  GEN R, a, b, q, H, Hp, g, Ag, Bg;
   long m, n, valX, valA, vA = varn(A);
   ulong p;
   pari_sp ltop, av, avlim;
-  byteptr d;
+  forprime_t S;
 
   if (!signe(A)) { if (Anew) *Anew = pol_0(vA); return ZX_copy(B); }
   if (!signe(B)) { if (Anew) *Anew = pol_1(vA); return ZX_copy(A); }
@@ -1113,12 +1100,12 @@ ZX_gcd_all(GEN A, GEN B, GEN *Anew)
     Ag = ZX_Z_mul(A,g);
     Bg = ZX_Z_mul(B,g);
   }
-
+  init_modular(&S);
   av = avma; avlim = stack_lim(av, 1);
-  H = NULL; d = init_modular(&p);
-  for(;;)
+  R = NULL;/*-Wall*/
+  H = NULL;
+  while ((p = u_forprime_next(&S)))
   {
-    NEXT_PRIME_VIADIFF_CHECK(p,d);
     if (g && !umodiu(g,p)) continue;
     a = ZX_to_Flx(A, p);
     b = ZX_to_Flx(B, p); Hp = Flx_gcd(a,b, p);
@@ -1152,20 +1139,19 @@ ZX_gcd_all(GEN A, GEN B, GEN *Anew)
       gerepileall(av, 3, &H, &q, &Hp);
     }
 
-    qp = muliu(q,p);
-    if (!ZX_incremental_CRT(&H, Hp, q, qp, p)) { q = qp; continue; }
+    if (!ZX_incremental_CRT(&H, Hp, &q, p)) continue;
     /* H stable: check divisibility */
-    q = qp;
     if (!ZX_divides(Bg, H)) continue;
     R = ZX_divides(Ag, H);
-    if (!R) continue;
-    if (Anew) {
-      A = R;
-      if (valA != valX) A = RgX_shift(A, valA - valX);
-      *Anew = A;
-    }
-    return valX ? RgX_shift(H, valX): H;
+    if (R) break;
   }
+  if (!p) pari_err_OVERFLOW("ZX_gcd_all [ran out of primes]");
+  if (Anew) {
+    A = R;
+    if (valA != valX) A = RgX_shift(A, valA - valX);
+    *Anew = A;
+  }
+  return valX ? RgX_shift(H, valX): H;
 }
 GEN
 ZX_gcd(GEN A, GEN B) { return ZX_gcd_all(A,B,NULL); }
@@ -1189,3 +1175,239 @@ QX_gcd(GEN A0, GEN B0)
   if (isint1(a)) avma = av2; else D = RgX_Rg_mul(D, a);
   return gerepileupto(av, D);
 }
+
+/*****************************************************************************
+ * Variants of the Bradford-Davenport algorithm: look for cyclotomic         *
+ * factors, and decide whether a ZX is cyclotomic or a product of cyclotomic *
+ *****************************************************************************/
+/* f of degree 1, return a cyclotomic factor (Phi_1 or Phi_2) or NULL */
+static GEN
+BD_deg1(GEN f)
+{
+  GEN a = gel(f,3), b = gel(f,2); /* f = ax + b */
+  if (!absi_equal(a,b)) return NULL;
+  return polcyclo((signe(a) == signe(b))? 2: 1, varn(f));
+}
+
+/* f a squarefree ZX; not divisible by any Phi_n, n even */
+static GEN
+BD_odd(GEN f)
+{
+  while(degpol(f) > 1)
+  {
+    GEN f1 = ZX_graeffe(f); /* contain all cyclotomic divisors of f */
+    if (ZX_equal(f1, f)) return f; /* product of cyclotomics */
+    f = ZX_gcd(f, f1);
+  }
+  if (degpol(f) == 1) return BD_deg1(f);
+  return NULL; /* no cyclotomic divisor */
+}
+
+static GEN
+myconcat(GEN v, GEN x)
+{
+  if (typ(x) != t_VEC) x = mkvec(x);
+  if (!v) return x;
+  return shallowconcat(v, x);
+}
+
+/* Bradford-Davenport algorithm.
+ * f a squarefree ZX of degree > 0, return NULL or a vector of coprime
+ * cyclotomic factors of f [ possibly reducible ] */
+static GEN
+BD(GEN f)
+{
+  GEN G = NULL, Gs = NULL, Gp = NULL, Gi = NULL;
+  GEN fs2, fp, f2, f1, fe, fo, fe1, fo1;
+  RgX_even_odd(f, &fe, &fo);
+  fe1 = ZX_eval1(fe);
+  fo1 = ZX_eval1(fo);
+  if (absi_equal(fe1, fo1)) /* f(1) = 0 or f(-1) = 0 */
+  {
+    long i, v = varn(f);
+    if (!signe(fe1))
+      G = mkvec2(polcyclo(1, v), polcyclo(2, v)); /* both 0 */
+    else if (signe(fe1) == signe(fo1))
+      G = mkvec(polcyclo(2, v)); /*f(-1) = 0*/
+    else
+      G = mkvec(polcyclo(1, v)); /*f(1) = 0*/
+    for (i = lg(G)-1; i; i--) f = RgX_div(f, gel(G,i));
+  }
+  /* f no longer divisible by Phi_1 or Phi_2 */
+  if (degpol(f) <= 1) return G;
+  f1 = ZX_graeffe(f); /* has at most square factors */
+  if (ZX_equal(f1, f)) return myconcat(G,f); /* f = product of Phi_n, n odd */
+
+  fs2 = ZX_gcd_all(f1, ZX_deriv(f1), &f2); /* fs2 squarefree */
+  if (degpol(fs2))
+  { /* fs contains all Phi_n | f, 4 | n; and only those */
+    /* In that case, Graeffe(Phi_n) = Phi_{n/2}^2, and Phi_n = Phi_{n/2}(x^2) */
+    GEN fs = RgX_inflate(fs2, 2);
+    (void)ZX_gcd_all(f, fs, &f); /* remove those Phi_n | f, 4 | n */
+    Gs = BD(fs2);
+    if (Gs)
+    {
+      long i;
+      for (i = lg(Gs)-1; i; i--) gel(Gs,i) = RgX_inflate(gel(Gs,i), 2);
+      /* prod Gs[i] is the product of all Phi_n | f, 4 | n */
+      G = myconcat(G, Gs);
+    }
+    /* f2 = f1 / fs2 */
+    f1 = RgX_div(f2, fs2); /* f1 / fs2^2 */
+  }
+  fp = ZX_gcd(f, f1); /* contains all Phi_n | f, n > 1 odd; and only those */
+  if (degpol(fp))
+  {
+    Gp = BD_odd(fp);
+    /* Gp is the product of all Phi_n | f, n odd */
+    if (Gp) G = myconcat(G, Gp);
+    f = RgX_div(f, fp);
+  }
+  if (degpol(f))
+  { /* contains all Phi_n originally dividing f, n = 2 mod 4, n > 2;
+     * and only those
+     * In that case, Graeffe(Phi_n) = Phi_{n/2}, and Phi_n = Phi_{n/2}(-x) */
+    Gi = BD_odd(ZX_unscale(f, gen_m1));
+    if (Gi)
+    { /* N.B. Phi_2 does not divide f */
+      Gi = ZX_unscale(Gi, gen_m1);
+      /* Gi is the product of all Phi_n | f, n = 2 mod 4 */
+      G = myconcat(G, Gi);
+    }
+  }
+  return G;
+}
+
+/* Let f be a non-zero QX, return the (squarefree) product of cyclotomic
+ * divisors of f */
+GEN
+polcyclofactors(GEN f)
+{
+  pari_sp av = avma;
+  if (typ(f) != t_POL) pari_err_TYPE("polcyclofactors",f);
+  (void)RgX_valrem(f, &f);
+  f = Q_primpart(f);
+  if (!RgX_is_ZX(f) || !signe(f)) pari_err_TYPE("polcyclofactors",f);
+  if (degpol(f))
+  {
+    (void)ZX_gcd_all(f, ZX_deriv(f), &f);
+    f = BD(f);
+    if (f) return gerepilecopy(av, f);
+  }
+  avma = av; return cgetg(1,t_VEC);
+}
+
+/* return t*x mod T(x), T a monic ZX. Assume deg(t) < deg(T) */
+static GEN
+ZXQ_mul_by_X(GEN t, GEN T)
+{
+  GEN lt;
+  t = RgX_shift_shallow(t, 1);
+  if (degpol(t) < degpol(T)) return t;
+  lt = leading_term(t);
+  if (is_pm1(lt)) return signe(lt) > 0 ? ZX_sub(t, T): ZX_add(t, T);
+  return ZX_sub(t, ZX_Z_mul(T, leading_term(t)));
+}
+/* f a product of Phi_n, all n odd; deg f > 1. Is it irreducible ? */
+static long
+BD_odd_iscyclo(GEN f)
+{
+  pari_sp av, lim;
+  long d, e, n, bound;
+  GEN t;
+  f = RgX_deflate_max(f, &e);
+  av = avma; lim = stack_lim(av,1);
+  /* The original f is cyclotomic (= Phi_{ne}) iff the present one is Phi_n,
+   * where all prime dividing e also divide n. If current f is Phi_n,
+   * then n is odd and squarefree */
+  d = degpol(f); /* = phi(n) */
+  /* Let e > 0, g multiplicative such that
+       g(p) = p / (p-1)^(1+e) < 1 iff p < (p-1)^(1+e)
+     For all squarefree odd n, we have g(n) < C, hence n < C phi(n)^(1+e), where
+       C = \prod_{p odd | p > (p-1)^(1+e)} g(p)
+     For e = 1/10,   we obtain p = 3, 5 and C < 1.523
+     For e = 1/100,  we obtain p = 3, 5, ..., 29 and C < 2.573
+     In fact, for n <= 10^7 odd & squarefree, we have n < 2.92 * phi(n)
+     By the above, n<10^7 covers all d <= (10^7/2.573)^(1/(1+1/100)) < 3344391.
+  */
+  if (d <= 3344391)
+    bound = (long)(2.92 * d);
+  else
+    bound = (long)(2.573 * pow(d,1.01));
+  /* IF f = Phi_n, n squarefree odd, then n <= bound */
+  t = monomial(gen_1, d-1, varn(f));
+  for (n = d; n <= bound; n++)
+  {
+    t = ZXQ_mul_by_X(t, f);
+    /* t = (X mod f(X))^d */
+    if (degpol(t) == 0) break;
+    if (low_stack(lim, stack_lim(av,1)))
+    {
+      if(DEBUGMEM>1) pari_warn(warnmem,"BD_odd_iscyclo");
+      t = gerepilecopy(av, t);
+    }
+  }
+  if (n > bound || eulerphiu(n) != (ulong)d) return 0;
+
+  if (e > 1) return (ucoprime_part(e, n) == 1)? e * n : 0;
+  return n;
+}
+
+/* Checks if f, monic squarefree ZX with |constant coeff| = 1, is a cyclotomic
+ * polynomial. Returns n if f = Phi_n, and 0 otherwise */
+static long
+BD_iscyclo(GEN f)
+{
+  pari_sp av = avma;
+  GEN f2, fn, f1;
+
+  if (degpol(f) == 1) return isint1(gel(f,2))? 2: 1;
+  f1 = ZX_graeffe(f);
+  /* f = product of Phi_n, n odd */
+  if (ZX_equal(f, f1)) { avma = av; return BD_odd_iscyclo(f); }
+
+  fn = ZX_unscale(f, gen_m1); /* f(-x) */
+  /* f = product of Phi_n, n = 2 mod 4 */
+  if (ZX_equal(f1, fn)) return 2*BD_odd_iscyclo(fn);
+
+  if (issquareall(f1, &f2))
+  {
+    GEN lt = leading_term(f2);
+    long c;
+    if (signe(lt) < 0) f2 = ZX_neg(f2);
+    c = BD_iscyclo(f2);
+    return odd(c)? 0: 2*c;
+  }
+  avma = av; return 0;
+}
+long
+poliscyclo(GEN f)
+{
+  pari_sp av = avma;
+  long d = degpol(f);
+  if (typ(f) != t_POL) pari_err_TYPE("poliscyclo", f);
+  if (d <= 0 || !RgX_is_ZX(f)) return 0;
+  if (!equali1(gel(f,d+2)) || !is_pm1(gel(f,2))) return 0;
+  if (d == 1) return signe(gel(f,2)) > 0? 2: 1;
+  if (degpol(ZX_gcd(f, ZX_deriv(f)))) { avma = av; return 0; }
+  return BD_iscyclo(f);
+}
+
+long
+poliscycloprod(GEN f)
+{
+  pari_sp av = avma;
+  long i, d = degpol(f);
+  if (typ(f) != t_POL) pari_err_TYPE("poliscycloprod",f);
+  if (!RgX_is_ZX(f)) return 0;
+  if (!equali1(leading_term(f)) || !is_pm1(constant_term(f))) return 0;
+  if (d < 2) return (d == 1);
+  if ( degpol(ZX_gcd_all(f, ZX_deriv(f), &f)) )
+  {
+    d = degpol(f);
+    if (d == 1) return 1;
+  }
+  f = BD(f); if (!f) return 0;
+  for (i = lg(f)-1; i; i--) d -= degpol(gel(f,i));
+  avma = av; return d == 0;
+}
