@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -31,7 +29,7 @@ static GEN
 get_tab(GEN nf, long *N)
 {
   GEN tab = (typ(nf) == t_MAT)? nf: gel(nf,9);
-  *N = lg(tab[1])-1; return tab;
+  *N = nbrows(tab); return tab;
 }
 
 /* x != 0, y t_INT. Return x * y (not memory clean if x = 1) */
@@ -108,7 +106,7 @@ ei_multable(GEN TAB, long i)
   GEN m, tab = get_tab(TAB, &N);
   tab += (i-1)*N;
   m = cgetg(N+1,t_MAT);
-  for (k=1; k<=N; k++) m[k] = tab[k];
+  for (k=1; k<=N; k++) gel(m,k) = gel(tab,k);
   return m;
 }
 
@@ -126,6 +124,7 @@ multable(GEN M, GEN x)
 {
   long i, N;
   GEN mul;
+  if (typ(x) == t_MAT) return x;
   M = get_tab(M, &N);
   if (typ(x) != t_COL) return scalarmat(x, N);
   mul = cgetg(N+1,t_MAT);
@@ -156,16 +155,50 @@ nftrace(GEN nf, GEN x)
   return gerepileupto(av, x);
 }
 GEN
+rnfelttrace(GEN rnf, GEN x)
+{
+  pari_sp av = avma;
+  checkrnf(rnf);
+  x = rnfeltabstorel(rnf, x);
+  x = (typ(x) == t_POLMOD)? rnfeltdown(rnf, gtrace(x))
+                          : gmulgs(x, rnf_get_degree(rnf));
+  return gerepileupto(av, x);
+}
+
+/* assume nf is a genuine nf, fa a famat */
+static GEN
+famat_norm(GEN nf, GEN fa)
+{
+  pari_sp av = avma;
+  GEN g = gel(fa,1), e = gel(fa,2), N = gen_1;
+  long i, l = lg(g);
+  for (i = 1; i < l; i++)
+    N = gmul(N, powgi(nfnorm(nf, gel(g,i)), gel(e,i)));
+  return gerepileupto(av, N);
+}
+GEN
 nfnorm(GEN nf, GEN x)
 {
   pari_sp av = avma;
   nf = checknf(nf);
+  if (typ(x) == t_MAT) return famat_norm(nf, x);
   x = nf_to_scalar_or_alg(nf, x);
   x = (typ(x) == t_POL)? RgXQ_norm(x, nf_get_pol(nf))
                        : gpowgs(x, nf_get_degree(nf));
   return gerepileupto(av, x);
 }
 
+GEN
+rnfeltnorm(GEN rnf, GEN x)
+{
+  pari_sp av = avma;
+  checkrnf(rnf);
+  x = rnfeltabstorel(rnf, x);
+  x = (typ(x) == t_POLMOD)? rnfeltdown(rnf, gnorm(x))
+                          : gpowgs(x, rnf_get_degree(rnf));
+  return gerepileupto(av, x);
+}
+
 /* sum of x and y in nf */
 GEN
 nfadd(GEN nf, GEN x, GEN y)
@@ -260,7 +293,7 @@ nfC_nf_mul(GEN nf, GEN v, GEN x)
     {
       long s = signe(x);
       if (!s) return zerocol(lg(v)-1);
-      if (is_pm1(x)) return s > 0? leafcopy(v): gneg(v);
+      if (is_pm1(x)) return s > 0? leafcopy(v): RgC_neg(v);
     }
     l = lg(v); y = cgetg(l, t_COL);
     for (i=1; i < l; i++)
@@ -289,22 +322,33 @@ nfC_nf_mul(GEN nf, GEN v, GEN x)
   }
   return dx? RgC_Rg_div(y, dx): y;
 }
+static GEN
+mulbytab(GEN M, GEN c)
+{ return typ(c) == t_COL? RgM_RgC_mul(M,c): RgC_Rg_mul(gel(M,1), c); }
 GEN
 tablemulvec(GEN M, GEN x, GEN v)
 {
   long l, i;
-  GEN y, dx = NULL;
+  GEN y;
 
-  if (RgV_isscalar(x)) return RgV_Rg_mul(v, gel(x,1));
-  x = multable(M, x);
-  l = lg(v); y = cgetg(l, t_VEC);
-  for (i=1; i < l; i++)
+  if (typ(x) == t_COL && RgV_isscalar(x))
   {
-    GEN c = gel(v,i);
-    if (typ(c)!=t_COL) c = RgC_Rg_mul(gel(x,1), c); else c = RgM_RgC_mul(x,c);
-    gel(y,i) = c;
+    x = gel(x,1);
+    return typ(v) == t_POL? RgX_Rg_mul(v,x): RgV_Rg_mul(v,x);
   }
-  return dx? gdiv(y, dx): y;
+  x = multable(M, x); /* multiplication table by x */
+  y = cgetg_copy(v, &l);
+  if (typ(v) == t_POL)
+  {
+    y[1] = v[1];
+    for (i=2; i < l; i++) gel(y,i) = mulbytab(x, gel(v,i));
+    y = normalizepol(y);
+  }
+  else
+  {
+    for (i=1; i < l; i++) gel(y,i) = mulbytab(x, gel(v,i));
+  }
+  return y;
 }
 
 /* inverse of x in nf */
@@ -526,7 +570,7 @@ nfpow(GEN nf, GEN z, GEN n)
   long s, N;
   GEN x, cx, T;
 
-  if (typ(n)!=t_INT) pari_err(talker,"not an integer exponent in nfpow");
+  if (typ(n)!=t_INT) pari_err_TYPE("nfpow",n);
   nf = checknf(nf); T = nf_get_pol(nf); N = degpol(T);
   s = signe(n); if (!s) return scalarcol_shallow(gen_1,N);
   x = nf_to_scalar_or_basis(nf, z);
@@ -587,57 +631,74 @@ pow_ei_mod_p(GEN nf, long I, GEN n, GEN p)
   long s,N;
   GEN y;
 
-  if (typ(n) != t_INT) pari_err(talker,"not an integer exponent in nfpow");
+  if (typ(n) != t_INT) pari_err_TYPE("nfpow",n);
   nf = checknf(nf); N = nf_get_degree(nf);
   s = signe(n);
-  if (s < 0) pari_err(talker,"negative power in pow_ei_mod_p");
+  if (s < 0) pari_err_IMPL("negative power in pow_ei_mod_p");
   if (!s || I == 1) return scalarcol_shallow(gen_1,N);
   D.nf = nf;
   D.p = p;
   D.I = I;
-  y = leftright_pow_fold(col_ei(N, I), n, (void*)&D, &sqr_mod, &ei_msqr_mod);
+  y = gen_pow_fold(col_ei(N, I), n, (void*)&D, &sqr_mod, &ei_msqr_mod);
   return gerepileupto(av,y);
 }
 
-/* valuation of integer x, with resp. to prime ideal P above p.
- * p.P^(-1) = b Z_K
- * [b may be given as the 'multiplication by b' matrix, possibly a t_INT]
- */
+/* valuation of integral x (ZV), with resp. to prime ideal pr */
 long
-int_elt_val(GEN nf, GEN x, GEN p, GEN b, GEN *newx)
+ZC_nfvalrem(GEN nf, GEN x, GEN pr, GEN *newx)
 {
-  long i, k, w, N;
-  GEN r, a, y, mul = zk_scalar_or_multable(nf, b);
+  long i, v, l;
+  GEN r, y, p = pr_get_p(pr), mul = zk_scalar_or_multable(nf, pr_get_tau(pr));
 
   if (typ(mul) == t_INT) return newx? ZV_pvalrem(x, p, newx):ZV_pval(x, p);
-
-  N = nf_get_degree(nf);
-  y = cgetg(N+1, t_COL); /* will hold the new x */
+  y = cgetg_copy(x, &l); /* will hold the new x */
   x = leafcopy(x);
-  for(w=0;; w++)
+  for(v=0;; v++)
   {
-    for (i=1; i<=N; i++)
-    { /* compute (x.b)_i */
-      a = mulii(gel(x,1), gcoeff(mul,i,1));
-      for (k=2; k<=N; k++) a = addii(a, mulii(gel(x,k), gcoeff(mul,i,k)));
-      /* is it divisible by p ? */
-      gel(y,i) = dvmdii(a,p,&r);
-      if (r != gen_0)
-      {
-        if (newx) *newx = x;
-        return w;
-      }
+    for (i=1; i<l; i++)
+    { /* is (x.b)[i] divisible by p ? */
+      gel(y,i) = dvmdii(ZMrow_ZC_mul(mul,x,i),p,&r);
+      if (r != gen_0) { if (newx) *newx = x; return v; }
     }
     swap(x, y);
   }
 }
+long
+ZC_nfval(GEN nf, GEN x, GEN P)
+{ return ZC_nfvalrem(nf, x, P, NULL); }
+
+/* v_P(x) != 0, x a ZV. Simpler version of ZC_nfvalrem */
+int
+ZC_prdvd(GEN nf, GEN x, GEN P)
+{
+  pari_sp av = avma;
+  long i, l;
+  GEN p = pr_get_p(P), mul = zk_scalar_or_multable(nf, pr_get_tau(P));
+  if (typ(mul) == t_INT) return ZV_Z_dvd(x, p);
+  l = lg(x);
+  for (i=1; i<l; i++)
+    if (remii(ZMrow_ZC_mul(mul,x,i), p) != gen_0) { avma = av; return 0; }
+  avma = av; return 1;
+}
+
+int
+pr_equal(GEN nf, GEN P, GEN Q)
+{
+  GEN gQ, p = pr_get_p(P);
+  long e = pr_get_e(P), f = pr_get_f(P), n;
+  if (!equalii(p, pr_get_p(Q)) || e != pr_get_e(Q) || f != pr_get_f(Q))
+    return 0;
+  gQ = pr_get_gen(Q); n = lg(gQ)-1;
+  if (2*e*f > n) return 1; /* room for only one such pr */
+  return ZV_equal(pr_get_gen(P), gQ) || ZC_prdvd(nf, gQ, P);
+}
 
 long
 nfval(GEN nf, GEN x, GEN pr)
 {
   pari_sp av = avma;
   long w, e;
-  GEN cx,p;
+  GEN cx, p;
 
   if (gequal0(x)) return LONG_MAX;
   nf = checknf(nf);
@@ -645,9 +706,9 @@ nfval(GEN nf, GEN x, GEN pr)
   p = pr_get_p(pr);
   e = pr_get_e(pr);
   x = nf_to_scalar_or_basis(nf, x);
-  if (typ(x) != t_COL) return e * Q_pval(x,p);
+  if (typ(x) != t_COL) return e*Q_pval(x,p);
   x = Q_primitive_part(x, &cx);
-  w = int_elt_val(nf,x,p, pr_get_tau(pr), NULL);
+  w = ZC_nfval(nf,x,pr);
   if (cx) w += e*Q_pval(cx,p);
   avma = av; return w;
 }
@@ -671,23 +732,24 @@ basistoalg(GEN nf, GEN x)
       return gerepilecopy(av, coltoalg(nf, x));
     }
     case t_POLMOD:
-      if (!RgX_equal_var(nf_get_pol(nf),gel(x,1)))
-        pari_err(talker,"not the same number field in basistoalg");
+      T = nf_get_pol(nf);
+      if (!RgX_equal_var(T,gel(x,1)))
+        pari_err_MODULUS("basistoalg", T,gel(x,1));
       return gcopy(x);
     case t_POL:
       T = nf_get_pol(nf);
-      if (varn(T) != varn(x)) pari_err(consister,"basistoalg");
+      if (varn(T) != varn(x)) pari_err_VAR("basistoalg",x,T);
       z = cgetg(3,t_POLMOD);
-      gel(z,1) = gcopy(T);
+      gel(z,1) = ZX_copy(T);
       gel(z,2) = RgX_rem(x, T); return z;
     case t_INT:
     case t_FRAC:
       T = nf_get_pol(nf);
       z = cgetg(3,t_POLMOD);
-      gel(z,1) = gcopy(T);
+      gel(z,1) = ZX_copy(T);
       gel(z,2) = gcopy(x); return z;
     default:
-      pari_err(typeer,"basistoalg");
+      pari_err_TYPE("basistoalg",x);
       return NULL; /* not reached */
   }
 }
@@ -708,8 +770,7 @@ nf_to_scalar_or_basis(GEN nf, GEN x)
     {
       GEN T = nf_get_pol(nf);
       long l = lg(x);
-      if (varn(x) != varn(T))
-        pari_err(talker,"incompatible variables in nf_to_scalar_or_basis");
+      if (varn(x) != varn(T)) pari_err_VAR("nf_to_scalar_or_basis", x,T);
       if (l >= lg(T)) { x = RgX_rem(x, T); l = lg(x); }
       if (l == 2) return gen_0;
       if (l == 3) return gel(x,2);
@@ -719,7 +780,7 @@ nf_to_scalar_or_basis(GEN nf, GEN x)
       if (lg(x) != lg(nf_get_zk(nf))) break;
       return QV_isscalar(x)? gel(x,1): x;
   }
-  pari_err(typeer,"nf_to_scalar_or_basis");
+  pari_err_TYPE("nf_to_scalar_or_basis",x);
   return NULL; /* not reached */
 }
 /* Let x be a polynomial with coefficients in Q or nf. Return the same
@@ -750,8 +811,7 @@ nf_to_scalar_or_alg(GEN nf, GEN x)
     {
       GEN T = nf_get_pol(nf);
       long l = lg(x);
-      if (varn(x) != varn(T))
-        pari_err(talker,"incompatible variables in nf_to_scalar_or_alg");
+      if (varn(x) != varn(T)) pari_err_VAR("nf_to_scalar_or_alg", x,T);
       if (l >= lg(T)) { x = RgX_rem(x, T); l = lg(x); }
       if (l == 2) return gen_0;
       if (l == 3) return gel(x,2);
@@ -761,32 +821,30 @@ nf_to_scalar_or_alg(GEN nf, GEN x)
       if (lg(x) != lg(nf_get_zk(nf))) break;
       return QV_isscalar(x)? gel(x,1): coltoliftalg(nf, x);
   }
-  pari_err(typeer,"nf_to_scalar_or_alg");
+  pari_err_TYPE("nf_to_scalar_or_alg",x);
   return NULL; /* not reached */
 }
 
-/* gmul(A, RgX_to_RgV(x)), A t_MAT (or t_VEC) of compatible dimensions */
+/* gmul(A, RgX_to_Rg(x)), A t_MAT (or t_VEC) of compatible dimensions */
 GEN
 mulmat_pol(GEN A, GEN x)
 {
   long i,l;
   GEN z;
   if (typ(x) != t_POL) return gmul(x,gel(A,1)); /* scalar */
-  l=lg(x)-1; if (l == 1) return typ(A)==t_VEC? gen_0: zerocol(lg(A[1])-1);
+  l=lg(x)-1; if (l == 1) return typ(A)==t_VEC? gen_0: zerocol(nbrows(A));
   x++; z = gmul(gel(x,1), gel(A,1));
   for (i=2; i<l ; i++)
     if (!gequal0(gel(x,i))) z = gadd(z, gmul(gel(x,i), gel(A,i)));
   return z;
 }
 
-/* valid for t_POL, nf a genuine nf or an rnf !
- * No garbage collecting. No check.  */
+/* x a t_POL, nf a genuine nf. No garbage collecting. No check.  */
 GEN
 poltobasis(GEN nf, GEN x)
 {
   GEN P = nf_get_pol(nf);
-  if (varn(x) != varn(P))
-    pari_err(talker, "incompatible variables in poltobasis");
+  if (varn(x) != varn(P)) pari_err_VAR( "poltobasis", x,P);
   if (degpol(x) >= degpol(P)) x = RgX_rem(x,P);
   return mulmat_pol(nf_get_invzk(nf), x);
 }
@@ -801,7 +859,7 @@ algtobasis(GEN nf, GEN x)
   {
     case t_POLMOD:
       if (!RgX_equal_var(nf_get_pol(nf),gel(x,1)))
-        pari_err(talker,"not the same number field in algtobasis");
+        pari_err_MODULUS("algtobasis", nf_get_pol(nf),gel(x,1));
       x = gel(x,2);
       switch(typ(x))
       {
@@ -818,45 +876,59 @@ algtobasis(GEN nf, GEN x)
       return gerepileupto(av,poltobasis(nf,x));
 
     case t_COL:
+      if (lg(x)-1 != nf_get_degree(nf)) pari_err_DIM("nfalgtobasis");
       return gcopy(x);
 
     case t_INT:
     case t_FRAC: return scalarcol(x, nf_get_degree(nf));
   }
-  pari_err(typeer,"algtobasis");
+  pari_err_TYPE("algtobasis",x);
   return NULL; /* not reached */
 }
 
 GEN
 rnfbasistoalg(GEN rnf,GEN x)
 {
+  const char *f = "rnfbasistoalg";
   long lx, i;
   pari_sp av = avma;
-  GEN z, nf;
+  GEN z, nf, relpol, T;
 
   checkrnf(rnf);
+  nf = rnf_get_nf(rnf);
+  T = nf_get_pol(nf);
+  relpol = QXQX_to_mod_shallow(rnf_get_pol(rnf), T);
   switch(typ(x))
   {
-    case t_VEC: case t_COL:
-      z = cgetg_copy(x, &lx); nf = gel(rnf,10);
-      for (i=1; i<lx; i++) gel(z,i) = nf_to_scalar_or_alg(nf, gel(x,i));
-      z = RgV_RgC_mul(gmael(rnf,7,1), z);
-      return gerepileupto(av, gmodulo(z,gel(rnf,1)));
-
-    case t_MAT:
+    case t_COL:
       z = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(z,i) = rnfbasistoalg(rnf,gel(x,i));
-      return z;
+      for (i=1; i<lx; i++)
+      {
+        GEN c = nf_to_scalar_or_alg(nf, gel(x,i));
+        if (typ(c) == t_POL) c = mkpolmod(c,T);
+        gel(z,i) = c;
+      }
+      z = RgV_RgC_mul(gel(rnf_get_zk(rnf),1), z);
+      return gerepileupto(av, gmodulo(z,relpol));
 
     case t_POLMOD:
-      if (!RgX_equal_var(gel(rnf,1),gel(x,1)))
-        pari_err(talker,"not the same number field in rnfbasistoalg");
-      return gcopy(x);
-
-    default: z = cgetg(3,t_POLMOD);
-      gel(z,1) = gcopy(gel(rnf,1));
-      gel(z,2) = gtopoly(x, varn(rnf[1])); return z;
+      x = polmod_nffix(f, rnf, x, 0);
+      if (typ(x) != t_POL) break;
+      retmkpolmod(RgX_copy(x), RgX_copy(relpol));
+    case t_POL:
+      if (varn(x) == varn(T))
+      {
+        if (!RgX_is_QX(x)) pari_err_TYPE(f,x);
+        x = gmodulo(x,T); break;
+      }
+      if (varn(x) == varn(relpol))
+      {
+        x = RgX_nffix(f,nf_get_pol(nf),x,0);
+        return gmodulo(x, relpol);
+      }
+      pari_err_VAR(f, x,relpol);
   }
+  retmkpolmod(scalarpol(x, varn(relpol)), RgX_copy(relpol));
 }
 
 GEN
@@ -872,9 +944,9 @@ matbasistoalg(GEN nf,GEN x)
       for (i=1; i<lx; i++) gel(z,i) = basistoalg(nf, gel(x,i));
       return z;
     case t_MAT: break;
-    default: pari_err(typeer, "matbasistoalg");
+    default: pari_err_TYPE("matbasistoalg",x);
   }
-  li = lg(x[1]);
+  li = lgcols(x);
   for (j=1; j<lx; j++)
   {
     GEN c = cgetg(li,t_COL), xj = gel(x,j);
@@ -897,9 +969,9 @@ matalgtobasis(GEN nf,GEN x)
       for (i=1; i<lx; i++) gel(z,i) = algtobasis(nf, gel(x,i));
       return z;
     case t_MAT: break;
-    default: pari_err(typeer, "matalgtobasis");
+    default: pari_err_TYPE("matalgtobasis",x);
   }
-  li = lg(x[1]);
+  li = lgcols(x);
   for (j=1; j<lx; j++)
   {
     GEN c = cgetg(li,t_COL), xj = gel(x,j);
@@ -915,7 +987,7 @@ RgM_to_nfM(GEN nf,GEN x)
   GEN z = cgetg_copy(x, &lx);
 
   if (lx == 1) return z;
-  li = lg(x[1]);
+  li = lgcols(x);
   for (j=1; j<lx; j++)
   {
     GEN c = cgetg(li,t_COL), xj = gel(x,j);
@@ -933,32 +1005,61 @@ RgC_to_nfC(GEN nf,GEN x)
   return z;
 }
 
+/* x a t_POLMOD, supposedly in rnf = K[z]/(T), K = Q[y]/(Tnf) */
+GEN
+polmod_nffix(const char *f, GEN rnf, GEN x, int lift)
+{ return polmod_nffix2(f, rnf_get_nfpol(rnf), rnf_get_pol(rnf), x,lift); }
+GEN
+polmod_nffix2(const char *f, GEN T, GEN relpol, GEN x, int lift)
+{
+  if (RgX_equal_var(gel(x,1),relpol))
+  {
+    x = gel(x,2);
+    if (typ(x) == t_POL && varn(x) == varn(relpol))
+    {
+      x = RgX_nffix(f, T, x, lift);
+      switch(lg(x))
+      {
+        case 2: return gen_0;
+        case 3: return gel(x,2);
+      }
+      return x;
+    }
+  }
+  return Rg_nffix(f, T, x, lift);
+}
 GEN
 rnfalgtobasis(GEN rnf,GEN x)
 {
-  long i, lx;
-  GEN z;
+  const char *f = "rnfalgtobasis";
+  pari_sp av = avma;
+  GEN T, relpol;
 
   checkrnf(rnf);
+  relpol = rnf_get_pol(rnf);
+  T = rnf_get_nfpol(rnf);
   switch(typ(x))
   {
-    case t_VEC: case t_COL: case t_MAT:
-      z = cgetg_copy(x, &lx);
-      for (i=1; i<lx; i++) gel(z,i) = rnfalgtobasis(rnf,gel(x,i));
-      return z;
+    case t_COL:
+      if (lg(x)-1 != rnf_get_degree(rnf)) pari_err_DIM(f);
+      x = RgV_nffix(f, T, x, 0);
+      return gerepilecopy(av, x);
 
     case t_POLMOD:
-      if (!RgX_equal_var(gel(rnf,1),gel(x,1)))
-        pari_err(talker,"not the same number field in rnfalgtobasis");
-      x = gel(x,2);
-      if (typ(x) != t_POL) { GEN A = gel(rnf,8); return gmul(x, gel(A,1)); }
-      /* fall through */
-    case t_POL: {
-      pari_sp av = avma;
-      return gerepileupto(av, poltobasis(rnf, x));
-    }
+      x = polmod_nffix(f, rnf, x, 0);
+      if (typ(x) != t_POL) break;
+      return gerepileupto(av, mulmat_pol(rnf_get_invzk(rnf), x));
+    case t_POL:
+      if (varn(x) == varn(T))
+      {
+        if (!RgX_is_QX(x)) pari_err_TYPE(f,x);
+        x = mkpolmod(x,T); break;
+      }
+      x = RgX_nffix(f, T, x, 0);
+      if (degpol(x) >= degpol(relpol)) x = RgX_rem(x,relpol);
+      return gerepileupto(av, mulmat_pol(rnf_get_invzk(rnf), x));
   }
-  return scalarcol(x, rnf_get_degree(rnf));
+  return gerepileupto(av, scalarcol(x, rnf_get_degree(rnf)));
 }
 
 /* Given a and b in nf, gives an algebraic integer y in nf such that a-b.y
@@ -1008,7 +1109,7 @@ eval_sign(GEN M, GEN x, long k)
   GEN z = gel(x,1); /* times M[k,1], which is 1 */
   for (i = 2; i < l; i++)
     z = mpadd(z, mpmul(gcoeff(M,k,i), gel(x,i)));
-  if (lg(z) < DEFAULTPREC) pari_err(precer,"nfsign_arch");
+  if (realprec(z) < DEFAULTPREC) pari_err_PREC("nfsign_arch");
   return signe(z);
 }
 
@@ -1022,12 +1123,12 @@ vec01_to_indices(GEN v)
   {
    case t_VECSMALL: return v;
    case t_VEC: break;
-   default: pari_err(typeer,"vec01_to_indices");
+   default: pari_err_TYPE("vec01_to_indices",v);
   }
   l = lg(v);
   p = new_chunk(l) + l;
   for (k=1, i=l-1; i; i--)
-    if (signe(v[i])) { *--p = i; k++; }
+    if (signe(gel(v,i))) { *--p = i; k++; }
   *--p = evallg(k) | evaltyp(t_VECSMALL);
   avma = (pari_sp)p; return p;
 }
@@ -1053,7 +1154,7 @@ nfsign_arch(GEN nf, GEN x, GEN arch)
   if (typ(x) == t_MAT)
   { /* factorisation */
     GEN g = gel(x,1), e = gel(x,2);
-    V = const_vecsmall(n, 0);
+    V = zero_zv(n);
     for (i=1; i<lg(g); i++)
       if (mpodd(gel(e,i)))
         Flv_add_inplace(V, nfsign_arch(nf,gel(g,i),archp), 2);
@@ -1064,7 +1165,8 @@ nfsign_arch(GEN nf, GEN x, GEN arch)
   switch(typ(x))
   {
     case t_INT:
-      s = signe(x); if (!s) pari_err(talker,"zero element in nfsign_arch");
+      s = signe(x);
+      if (!s) pari_err_DOMAIN("nfsign_arch","element","=",gen_0,x);
       avma = av; return const_vecsmall(n, (s < 0)? 1: 0);
     case t_FRAC:
       s = signe(gel(x,1));
@@ -1124,7 +1226,7 @@ nfinvmodideal(GEN nf, GEN x, GEN y)
   if (typ(x) == t_INT) return gerepileupto(av, Fp_inv(x, yZ));
 
   a = hnfmerge_get_1(idealhnf_principal(nf,x), y);
-  if (!a) pari_err(talker, "element not invertible in nfinvmodideal");
+  if (!a) pari_err_INV("nfinvmodideal", x);
   return gerepileupto(av, ZC_hnfrem(nfdiv(nf,a,x), y));
 }
 
@@ -1229,7 +1331,8 @@ zidealij(GEN x, GEN y, GEN *U)
   for (j=1; j<N; j++)
   {
     GEN c = gel(G,j);
-    gel(c,1) = addsi(1, gel(c,1)); /* 1 + g_j */
+    gel(c,1) = addiu(gel(c,1), 1); /* 1 + g_j */
+    if (ZV_isscalar(c)) gel(G,j) = gel(c,1);
   }
   if (U) *U = RgM_mul(*U, RgM_inv(x));
   return mkvec2(cyc, G);
@@ -1272,29 +1375,27 @@ detcyc(GEN cyc, long *L)
   *L = i; return i <= 2? icopy(s): gerepileuptoint(av,s);
 }
 
+/* (U,V) = 1. Return y = x mod U, = 1 mod V (uv: u + v = 1, u in U, v in V),
+ * namely u + x*v.
+ * Either (u,mv) are both t_INT, or
+ * u is a t_COL, mv is a t_MAT, multiplication table by v */
 static GEN
-makeprimetoideal_i(GEN u,GEN mv, GEN x)
+makeprimetoideal(GEN UV, GEN u,GEN mv, GEN x)
 {
-  GEN xv;
-  if (typ(mv) == t_INT)
+  GEN t;
+  if (typ(mv) == t_INT) /* u t_INT */
   {
-    if (typ(x) == t_INT) return ZC_Z_add(u, mulii(x,mv));
-    xv = ZC_Z_mul(x,mv);
+    if (typ(x) == t_INT) return addii(mulii(x,mv), u);
+    t = ZC_Z_add(ZC_Z_mul(x,mv), u);
   }
   else
   {
-    if (typ(x) == t_INT) xv = ZC_Z_mul(gel(mv,1),x); else xv = ZM_ZC_mul(mv,x);
+    t = typ(x) == t_INT? ZC_Z_mul(gel(mv,1),x): ZM_ZC_mul(mv,x);
+    t = ZC_add(t, u);
   }
-  return ZC_add(u, xv);
+  return ZC_hnfrem(t, UV);
 }
 
-/* (U,V) = 1. Return y = x mod U, = 1 mod V (uv: u + v = 1, u in U, v in V),
- * namely u + x*v.
- * u is a t_COL, mv = multiplication table by v is a t_MAT or t_INT */
-static GEN
-makeprimetoideal(GEN UV, GEN u,GEN mv, GEN x)
-{ return ZC_hnfrem(makeprimetoideal_i(u, mv, x), UV); }
-
 static GEN
 makeprimetoidealvec(GEN UV, GEN u,GEN mv, GEN gen)
 {
@@ -1317,12 +1418,13 @@ static GEN
 zprimestar(GEN nf, GEN pr, GEN ep, GEN x, GEN arch)
 {
   pari_sp av = avma;
-  long a, b, e = itos(ep), f = pr_get_f(pr);
-  GEN p = pr_get_p(pr), list, g, g0, y, u,v, prh, prb, pre;
+  long a, e = itos(ep), f = pr_get_f(pr);
+  GEN p = pr_get_p(pr), list, g, g0, y, u,v, prb, pre;
+  ulong mask;
 
   if(DEBUGLEVEL>3) err_printf("treating pr^%ld, pr = %Ps\n",e,pr);
   if (f == 1)
-    g = scalarcol_shallow(pgener_Fp(p), nf_get_degree(nf));
+    g = pgener_Fp(p);
   else
   {
     GEN T, modpr = zk_to_Fq_init(nf, &pr, &T, &p);
@@ -1330,39 +1432,46 @@ zprimestar(GEN nf, GEN pr, GEN ep, GEN x, GEN arch)
     g = poltobasis(nf, g);
   }
   /* g generates  (Z_K / pr)^* */
-  prh = idealhnf_two(nf,pr);
-  pre = (e==1)? prh: idealpow(nf,pr,ep);
-  g0 = g;
-  u = v = NULL; /* gcc -Wall */
+  prb = idealhnf_two(nf,pr);
+  pre = (e==1)? prb: idealpow(nf,pr,ep);
   if (x)
   {
-    GEN uv = idealaddtoone(nf,pre, idealdivpowprime(nf,x,pr,ep));
-    u = gel(uv,1);
-    v = zk_scalar_or_multable(nf, gel(uv,2));
-    g0 = makeprimetoideal(x,u,v,g);
+    GEN mv;
+    v = idealaddtoone_i(nf,idealdivpowprime(nf,x,pr,ep), pre);
+    mv = zk_scalar_or_multable(nf, v);
+    u = typ(mv) == t_INT? subui(1,mv): unnf_minus_x(v);
+    v = mv;
+    g0 = makeprimetoideal(x,u,mv,g);
+  }
+  else
+  {
+    u = v = NULL; /* gcc -Wall */
+    g0 = g;
   }
 
-  list = vectrunc_init(e+1);
-  y = mkvec5(mkvec(addis(powiu(p,f), -1)),
+  y = mkvec5(mkvec(subiu(powiu(p,f), 1)),
              mkvec(g),
              mkvec(g0),
              mkvec(nfsign_arch(nf,g0,arch)),
              gen_1);
+  if (e == 1) return gerepilecopy(av, mkvec(y));
+  list = vectrunc_init(e+1);
   vectrunc_append(list, y);
-  prb = prh;
-  for (a = b = 1; a < e; a = b)
+  mask = quadratic_prec_mask(e);
+  a = 1;
+  while (mask > 1)
   {
     GEN pra = prb, gen, z, s, U;
-    long i, l;
+    long i, l, b = a << 1;
 
-    b <<= 1;
+    if (mask & 1) b--;
+    mask >>= 1;
     /* compute 1 + pr^a / 1 + pr^b, 2a <= b */
     if(DEBUGLEVEL>3) err_printf("  treating a = %ld, b = %ld\n",a,b);
     prb = (b >= e)? pre: idealpows(nf,pr,b);
     z = zidealij(pra, prb, &U);
     gen = leafcopy(gel(z,2));
     s = cgetg_copy(gen, &l);
-    if(DEBUGLEVEL>3) err_printf("zidealij done\n");
     for (i = 1; i < l; i++)
     {
       if (x) gel(gen,i) = makeprimetoideal(x,u,v,gel(gen,i));
@@ -1370,6 +1479,7 @@ zprimestar(GEN nf, GEN pr, GEN ep, GEN x, GEN arch)
     }
     y = mkvec5(gel(z,1), gel(z,2), gen, s, U);
     vectrunc_append(list, y);
+    a = b;
   }
   return gerepilecopy(av, list);
 }
@@ -1391,11 +1501,11 @@ increment(GEN y, long k, long d)
 GEN
 archstar_full_rk(GEN x, GEN bas, GEN v, GEN gen)
 {
-  long i, r, lgmat, N = lg(bas)-1, nba = lg(v[1]) - 1;
+  long i, r, lgmat, N = lg(bas)-1, nba = nbrows(v);
   GEN lambda = cgetg(N+1, t_VECSMALL), mat = cgetg(nba+1,t_MAT);
 
   lgmat = lg(v); setlg(mat, lgmat+1);
-  for (i = 1; i < lgmat; i++) mat[i] = v[i];
+  for (i = 1; i < lgmat; i++) gel(mat,i) = gel(v,i);
   for (     ; i <= nba; i++)  gel(mat,i) = cgetg(nba+1, t_VECSMALL);
 
   if (x) { x = ZM_lll(x, 0.75, LLL_INPLACE); bas = RgV_RgM_mul(bas, x); }
@@ -1456,7 +1566,7 @@ nfarchstar(GEN nf, GEN x, GEN archp)
     else
     {
       GEN bas = nf_get_M(nf);
-      if (lg(bas[1]) > lg(archp)) bas = rowpermute(bas, archp);
+      if (lgcols(bas) > lg(archp)) bas = rowpermute(bas, archp);
       gen = cgetg(nba+1,t_VEC);
       gel(gen,1) = gZ;
       mat = archstar_full_rk(x, bas, mkmat(const_vecsmall(nba,1)), gen);
@@ -1467,38 +1577,40 @@ nfarchstar(GEN nf, GEN x, GEN archp)
   return mkvec3(cyc,gen,mat);
 }
 
+static GEN
+apply_U(GEN U, GEN a)
+{
+  GEN e;
+  if (typ(a) == t_INT)
+    e = RgC_Rg_mul(gel(U,1), subis(a, 1));
+  else
+  { /* t_COL */
+    GEN t = gel(a,1);
+    gel(a,1) = addsi(-1, gel(a,1)); /* a -= 1 */
+    e = RgM_RgC_mul(U, a);
+    gel(a,1) = t; /* restore */
+  }
+  return e;
+}
 /* a in Z_K (t_COL or t_INT), pr prime ideal, prk = pr^k,
  * list = zprimestar(nf, pr, k, ...)  */
 static GEN
 zlog_pk(GEN nf, GEN a, GEN y, GEN pr, GEN prk, GEN list, GEN *psigne)
 {
-  GEN L, e, cyc, gen, s, U;
   long i,j, llist = lg(list)-1;
-
   for (j = 1; j <= llist; j++)
   {
-    L = gel(list,j);
-    cyc = gel(L,1);
-    gen = gel(L,2);
-    s   = gel(L,4);
-    U   = gel(L,5);
+    GEN L = gel(list,j), e;
+    GEN cyc = gel(L,1), gen = gel(L,2), s = gel(L,4), U = gel(L,5);
     if (j == 1)
       e = mkcol( nf_log(nf, a, gel(gen,1), gel(cyc,1), pr) );
-    else if (typ(a) == t_INT)
-      e = RgC_Rg_mul(gel(U,1), subis(a, 1));
     else
-    { /* t_COL */
-      GEN t = gel(a,1);
-      gel(a,1) = addsi(-1, gel(a,1)); /* a -= 1 */
-      e = RgM_RgC_mul(U, a);
-      gel(a,1) = t; /* restore */
-    }
+      e = apply_U(U, a);
     /* here lg(e) == lg(cyc) */
     for (i = 1; i < lg(cyc); i++)
     {
       GEN t;
-      if (typ(gel(e,i)) != t_INT)
-        pari_err(talker,"%Ps not coprime to %Ps in zlog_pk", a, pr);
+      if (typ(gel(e,i)) != t_INT) pari_err_COPRIME("zlog_pk", a, pr);
       t = modii(negi(gel(e,i)), gel(cyc,i));
       gel(++y,0) = negi(t); if (!signe(t)) continue;
 
@@ -1643,7 +1755,7 @@ log_gen_pr(zlog_S *S, long index, GEN nf, long e)
     L = gel(L2,1);
     y = col_ei(S->n, yind+1);
     zlog_add_sign(y, gmael(L,4,1), S->lists);
-    return mkmat( ZM_ZC_mul(S->U, y) );
+    retmkmat( ZM_ZC_mul(S->U, y) );
   }
   else
   {
@@ -1659,7 +1771,7 @@ log_gen_pr(zlog_S *S, long index, GEN nf, long e)
     prk = idealpow(nf, pr, gel(S->e,index));
     for (i = 1; i < l; i++)
     {
-      GEN G = gel(g,i), sgn = const_vecsmall(narchp,0); /*positive at f_oo*/
+      GEN G = gel(g,i), sgn = zero_zv(narchp); /*positive at f_oo*/
       y = zerocol(S->n);
       (void)zlog_pk(nf, G, y + yind, pr, prk, L2, &sgn);
       zlog_add_sign(y, sgn, S->lists);
@@ -1701,6 +1813,22 @@ add_grp(GEN nf, GEN u1, GEN cyc, GEN gen, GEN bid)
     gel(bid,2) = mkvec2(h,cyc);
 }
 
+static int
+RgV_is_prV(GEN v)
+{
+  long l = lg(v), i;
+  for (i = 1; i < l; i++)
+    if (!get_prid(gel(v,i))) return 0;
+  return 1;
+}
+
+static int
+is_nf_factor(GEN F)
+{
+  return typ(F) == t_MAT && lg(F) == 3
+    && RgV_is_prV(gel(F,1)) && RgV_is_ZV(gel(F,2));
+}
+
 /* Compute [[ideal,arch], [h,[cyc],[gen]], idealfact, [liste], U]
    flag may include nf_GEN | nf_INIT */
 GEN
@@ -1714,22 +1842,44 @@ Idealstar(GEN nf, GEN ideal, long flag)
   R1 = nf_get_r1(nf);
   if (typ(ideal) == t_VEC && lg(ideal) == 3)
   {
-    arch = gel(ideal,2); ideal = gel(ideal,1);
-    i = typ(arch);
-    if (!is_vec_t(i) || lg(arch) != R1+1)
-      pari_err(talker,"incorrect archimedean component in Idealstar");
-    archp = vec01_to_indices(arch);
+    arch = gel(ideal,2);
+    ideal= gel(ideal,1);
+    switch(typ(arch))
+    {
+      case t_VEC:
+        if (lg(arch) != R1+1)
+          pari_err_TYPE("Idealstar [incorrect archimedean component]",arch);
+        archp = vec01_to_indices(arch);
+        break;
+      case t_VECSMALL:
+        archp = arch;
+        arch = vec01_to_indices(archp);
+        break;
+      default:
+        pari_err_TYPE("Idealstar [incorrect archimedean component]",arch);
+        return NULL;
+    }
   }
   else
   {
     arch = zerovec(R1);
     archp = cgetg(1, t_VECSMALL);
   }
-  x = idealhnf_shallow(nf, ideal);
-  if (lg(x) == 1 || typ(gcoeff(x,1,1)) != t_INT)
-    pari_err(talker,"Idealstar needs an integral non-zero ideal: %Ps",x);
+  if (is_nf_factor(ideal))
+  {
+    fa = ideal;
+    x = idealfactorback(nf, gel(fa,1), gel(fa,2), 0);
+  }
+  else
+  {
+    fa = NULL;
+    x = idealhnf_shallow(nf, ideal);
+  }
+  if (lg(x) == 1) pari_err_DOMAIN("Idealstar", "ideal","=",gen_0,x);
+  if (typ(gcoeff(x,1,1)) != t_INT)
+    pari_err_DOMAIN("Idealstar","denominator(ideal)", "!=",gen_1,x);
   sarch = nfarchstar(nf, x, archp);
-  fa = idealfactor(nf, ideal);
+  if (!fa) fa = idealfactor(nf, ideal);
   P = gel(fa,1);
   E = gel(fa,2); nbp = lg(P)-1;
   if (nbp)
@@ -1766,7 +1916,7 @@ Idealstar(GEN nf, GEN ideal, long flag)
         { /* log(g^f) mod divisor */
           GEN g = gel(G,k), f = gel(F,k), a = nfpowmodideal(nf,g,f,x);
           GEN sgn = mpodd(f)? nfsign_arch(nf, g, S.archp)
-                            : const_vecsmall(lg(S.archp)-1, 0);
+                            : zero_zv(lg(S.archp)-1);
           gel(h,++cp) = ZC_neg(zlog_ind(nf, a, &S, sgn, i));
           gcoeff(h,cp,cp) = f;
         }
@@ -1800,6 +1950,98 @@ Idealstar(GEN nf, GEN ideal, long flag)
   return gerepilecopy(av, y);
 }
 
+/* vectors of [[cyc],[g],U.X^-1] */
+static GEN
+principal_units(GEN nf, GEN pr, long e)
+{
+  pari_sp av = avma;
+  long a;
+  GEN list, y, prb, pre;
+  ulong mask;
+
+  if(DEBUGLEVEL>3) err_printf("treating pr^%ld, pr = %Ps\n",e,pr);
+  if (e == 1) return cgetg(1, t_VEC);
+  prb = idealhnf_two(nf,pr);
+  pre = idealpows(nf,pr,e);
+  list = vectrunc_init(e);
+  mask = quadratic_prec_mask(e);
+  a = 1;
+  while (mask > 1)
+  {
+    GEN pra = prb, z, U;
+    long b = a << 1;
+
+    if (mask & 1) b--;
+    mask >>= 1;
+    /* compute 1 + pr^a / 1 + pr^b, 2a <= b */
+    if(DEBUGLEVEL>3) err_printf("  treating a = %ld, b = %ld\n",a,b);
+    prb = (b >= e)? pre: idealpows(nf,pr,b);
+    z = zidealij(pra, prb, &U);
+    y = mkvec3(gel(z,1), gel(z,2), U);
+    vectrunc_append(list, y);
+    a = b;
+  }
+  return gerepilecopy(av, list);
+}
+
+static GEN
+log_prk(GEN nf, GEN a, long nbgen, GEN list, GEN prk)
+{
+  GEN y = zerocol(nbgen);
+  long i,j, iy = 1, llist = lg(list)-1;
+
+  for (j = 1; j <= llist; j++)
+  {
+    GEN L = gel(list,j);
+    GEN cyc = gel(L,1), gen = gel(L,2), U = gel(L,3);
+    GEN e = apply_U(U, a);
+    /* here lg(e) == lg(cyc) */
+    for (i = 1; i < lg(cyc); i++)
+    {
+      GEN t = modii(negi(gel(e,i)), gel(cyc,i));
+      gel(y, iy++) = negi(t); if (!signe(t)) continue;
+      if (j != llist) a = elt_mulpow_modideal(nf, a, gel(gen,i), t, prk);
+    }
+  }
+  return y;
+}
+
+/* multiplicative group (1 + pr) / (1 + pr^e) */
+GEN
+idealprincipalunits(GEN nf, GEN pr, long e)
+{
+  pari_sp av = avma;
+  long c, i, j, k, nbgen;
+  GEN cyc, u1 = NULL, pre, gen;
+  GEN g, EX, h, L2;
+  long cp = 0;
+
+  nf = checknf(nf); pre = idealpows(nf, pr, e);
+  L2 = principal_units(nf, pr, e);
+  c = lg(L2); gen = cgetg(c, t_VEC);
+  for (j = 1; j < c; j++) gel(gen, j) = gmael(L2,j,2);
+  gen = shallowconcat1(gen); nbgen = lg(gen)-1;
+
+  h = cgetg(nbgen+1,t_MAT);
+  for (j=1; j < lg(L2); j++)
+  {
+    GEN L = gel(L2,j), F = gel(L,1), G = gel(L,2);
+    for (k=1; k<lg(G); k++)
+    { /* log(g^f) mod pr^e */
+      GEN g = gel(G,k), f = gel(F,k), a = nfpowmodideal(nf,g,f,pre);
+      gel(h,++cp) = ZC_neg(log_prk(nf, a, nbgen, L2, pre));
+      gcoeff(h,cp,cp) = f;
+    }
+  }
+  /* assert(cp == nbgen) */
+  h = ZM_hnfall(h,NULL,0);
+  cyc = ZM_snf_group(h, NULL, &u1);
+  c = lg(u1); g = cgetg(c, t_VEC); EX = gel(cyc,1);
+  for (i=1; i<c; i++)
+    gel(g,i) = famat_to_nf_modideal_coprime(nf, gen, gel(u1,i), pre, EX);
+  return gerepilecopy(av, mkvec3(powiu(pr_norm(pr), e-1), cyc, g));
+}
+
 /* FIXME: obsolete */
 GEN
 zidealstarinitgen(GEN nf, GEN ideal)
@@ -1819,7 +2061,7 @@ idealstar0(GEN nf, GEN ideal,long flag)
     case 0: return Idealstar(nf,ideal, nf_GEN);
     case 1: return Idealstar(nf,ideal, nf_INIT);
     case 2: return Idealstar(nf,ideal, nf_INIT|nf_GEN);
-    default: pari_err(flagerr,"idealstar");
+    default: pari_err_FLAG("idealstar");
   }
   return NULL; /* not reached */
 }
@@ -1829,7 +2071,7 @@ check_nfelt(GEN x, GEN *den)
 {
   long l = lg(x), i;
   GEN t, d = NULL;
-  if (typ(x) != t_COL) pari_err(talker,"%Ps not a nfelt", x);
+  if (typ(x) != t_COL) pari_err_TYPE("check_nfelt", x);
   for (i=1; i<l; i++)
   {
     t = gel(x,i);
@@ -1839,7 +2081,7 @@ check_nfelt(GEN x, GEN *den)
       case t_FRAC:
         if (!d) d = gel(t,2); else d = lcmii(d, gel(t,2));
         break;
-      default: pari_err(talker,"%Ps not a nfelt", x);
+      default: pari_err_TYPE("check_nfelt", x);
     }
   }
   *den = d;
@@ -1917,29 +2159,30 @@ join_bid(GEN nf, GEN bid1, GEN bid2)
 {
   pari_sp av = avma;
   long i, nbgen, lx, lx1,lx2, l1,l2;
-  GEN I1,I2, f1,f2, G1,G2, fa1,fa2, lists1,lists2, cyc1,cyc2;
+  GEN I1,I2, G1,G2, fa1,fa2, lists1,lists2, cyc1,cyc2;
   GEN lists, fa, U, cyc, y, u1 = NULL, x, gen;
 
-  f1 = gel(bid1,1); I1 = gel(f1,1);
-  f2 = gel(bid2,1); I2 = gel(f2,1);
+  I1 = bid_get_ideal(bid1);
+  I2 = bid_get_ideal(bid2);
   if (gequal1(gcoeff(I1,1,1))) return bid2; /* frequent trivial case */
-  G1 = gel(bid1,2); G2 = gel(bid2,2);
-  fa1= gel(bid1,3); fa2= gel(bid2,3); x = idealmul(nf, I1,I2);
+  G1 = bid_get_grp(bid1);
+  G2 = bid_get_grp(bid2);
+  fa1= gel(bid1,3);
+  fa2= gel(bid2,3); x = idealmul(nf, I1,I2);
   fa = famat_mul_shallow(fa1, fa2);
-
   lists1 = gel(bid1,4); lx1 = lg(lists1);
   lists2 = gel(bid2,4); lx2 = lg(lists2);
   /* concat (lists1 - last elt [nfarchstar]) + lists2 */
   lx = lx1+lx2-2; lists = cgetg(lx,t_VEC);
-  for (i=1; i<lx1-1; i++) lists[i] = lists1[i];
-  for (   ; i<lx; i++)    lists[i] = lists2[i-lx1+2];
+  for (i=1; i<lx1-1; i++) gel(lists,i) = gel(lists1,i);
+  for (   ; i<lx; i++)    gel(lists,i) = gel(lists2,i-lx1+2);
 
-  cyc1 = gel(G1,2); l1 = lg(cyc1);
-  cyc2 = gel(G2,2); l2 = lg(cyc2);
+  cyc1 = abgrp_get_cyc(G1); l1 = lg(cyc1);
+  cyc2 = abgrp_get_cyc(G2); l2 = lg(cyc2);
   gen = (lg(G1)>3 && lg(G2)>3)? gen_1: NULL;
   nbgen = l1+l2-2;
   cyc = ZM_snf_group(diagonal_shallow(shallowconcat(cyc1,cyc2)),
-                 &U, gen? &u1: NULL);
+                     &U, gen? &u1: NULL);
   if (nbgen) {
     GEN U1 = gel(bid1,5), U2 = gel(bid2,5);
     U1 = l1 == 1? zeromat(nbgen,lg(U1)-1): ZM_mul(vecslice(U, 1, l1-1),   U1);
@@ -1951,19 +2194,23 @@ join_bid(GEN nf, GEN bid1, GEN bid2)
 
   if (gen)
   {
-    GEN mu,mv, u,v, uv = idealaddtoone(nf,gel(f1,1),gel(f2,1));
-    u = gel(uv,1);
-    v = gel(uv,2);
+    GEN mu,mv, u,v = idealaddtoone_i(nf,I2,I1);
     mv = zk_scalar_or_multable(nf, v);
     if (typ(mv) == t_INT)
-      mu = subsi(1,mv);
+    {
+      mu = u = subui(1,mv);
+      v = mv;
+    }
     else
+    {
+      u = unnf_minus_x(v);
       mu = RgM_Rg_add(ZM_neg(mv), gen_1); /* mult by u = 1-v */
-    gen = shallowconcat(makeprimetoidealvec(x,u,mv, gel(G1,3)),
-                        makeprimetoidealvec(x,v,mu, gel(G2,3)));
+    }
+    gen = shallowconcat(makeprimetoidealvec(x,u,mv, abgrp_get_gen(G1)),
+                        makeprimetoidealvec(x,v,mu, abgrp_get_gen(G2)));
   }
   y = cgetg(6,t_VEC);
-  gel(y,1) = mkvec2(x, gel(f1,2));
+  gel(y,1) = mkvec2(x, bid_get_arch(bid1));
   gel(y,3) = fa;
   gel(y,4) = lists;
   gel(y,5) = U;
@@ -2028,8 +2275,9 @@ static GEN
 zlog_unitsarch(GEN sgnU, GEN bid)
 {
   GEN lists = gel(bid,4), arch = gmael(bid,1,2);
-  return Flm_Flc_mul(gmael(lists, lg(lists)-1, 3),
-                     rowpermute(sgnU, vec01_to_indices(arch)), 2);
+  GEN listslast = gel(lists,lg(lists)-1);
+  GEN m = rowpermute(sgnU, vec01_to_indices(arch));
+  return Flm_mul(gel(listslast,3), m, 2);
 }
 
 /*  flag & nf_GEN : generators, otherwise no
@@ -2040,10 +2288,10 @@ Ideallist(GEN bnf, ulong bound, long flag)
 {
   const long do_units = flag & 2, big_id = !(flag & 4);
   const long istar_flag = (flag & nf_GEN) | nf_INIT;
-  byteptr ptdif = diffptr;
   pari_sp lim, av, av0 = avma;
   long i, j, l;
   GEN nf, z, p, fa, id, U, empty = cgetg(1,t_VEC);
+  forprime_t S;
   ideal_data ID;
   GEN (*join_z)(ideal_data*, GEN) =
     do_units? &join_unit
@@ -2070,17 +2318,17 @@ Ideallist(GEN bnf, ulong bound, long flag)
   ID.nf = nf;
 
   p = cgetipos(3);
+  u_forprime_init(&S, 2, bound);
   av = avma; lim = stack_lim(av,1);
-  maxprime_check(bound);
-  for (p[2] = 0; (ulong)p[2] <= bound; )
+  while ((p[2] = u_forprime_next(&S)))
   {
-    NEXT_PRIME_VIADIFF(p[2], ptdif);
     if (DEBUGLEVEL>1) { err_printf("%ld ",p[2]); err_flush(); }
     fa = idealprimedec(nf, p);
     for (j=1; j<lg(fa); j++)
     {
       GEN pr = gel(fa,j), z2;
-      ulong q, iQ, Q = itou_or_0(pr_norm(pr));
+      long f = pr_get_f(pr);
+      ulong q, Q = upowuu(p[2], f);
       if (!Q || Q > bound) break;
 
       z2 = leafcopy(z);
@@ -2088,6 +2336,7 @@ Ideallist(GEN bnf, ulong bound, long flag)
       ID.pr = ID.prL = pr;
       for (l=1; Q <= bound; l++, Q *= q) /* add pr^l */
       {
+        ulong iQ;
         ID.L = utoipos(l);
         if (big_id) {
           if (l > 1) ID.prL = idealpow(nf,pr,ID.L);
@@ -2117,7 +2366,7 @@ Ideallist(GEN bnf, ulong bound, long flag)
 }
 GEN
 ideallist0(GEN bnf,long bound, long flag) {
-  if (flag<0 || flag>4) pari_err(flagerr,"ideallist");
+  if (flag<0 || flag>4) pari_err_FLAG("ideallist");
   return Ideallist(bnf,bound,flag);
 }
 GEN
@@ -2129,18 +2378,15 @@ static GEN
 join_bid_arch(GEN nf, GEN bid1, GEN arch)
 {
   pari_sp av = avma;
-  long i, lx1;
-  GEN f1, G1, fa1, lists1, U;
+  GEN f1, G1, fa1, U;
   GEN lists, cyc, y, u1 = NULL, x, sarch, gen;
 
   checkbid(bid1);
   f1 = gel(bid1,1); G1 = gel(bid1,2); fa1 = gel(bid1,3);
   x = gel(f1,1);
   sarch = nfarchstar(nf, x, arch);
-  lists1 = gel(bid1,4); lx1 = lg(lists1);
-  lists = cgetg(lx1,t_VEC);
-  for (i=1; i<lx1-1; i++) lists[i] = lists1[i];
-  gel(lists,i) = sarch;
+  lists = leafcopy(gel(bid1,4));
+  gel(lists,lg(lists)-1) = sarch;
 
   gen = (lg(G1)>3)? gen_1: NULL;
   cyc = diagonal_shallow(shallowconcat(gel(G1,2), gel(sarch,1)));
@@ -2161,7 +2407,7 @@ join_arch(ideal_data *D, GEN x) {
 static GEN
 join_archunit(ideal_data *D, GEN x) {
   GEN bid = join_arch(D, gel(x,1)), U = gel(x,2);
-  U = ZM_mul(gel(bid,5), vconcat(U, zlog_unitsarch(D->sgnU, bid)));
+  U = ZM_mul(gel(bid,5), vconcat(U, zm_to_ZM(zlog_unitsarch(D->sgnU, bid))));
   return mkvec2(bid, U);
 }
 
@@ -2175,13 +2421,13 @@ ideallistarch(GEN bnf, GEN L, GEN arch)
   ideal_data ID;
   GEN (*join_z)(ideal_data*, GEN);
 
-  if (typ(L) != t_VEC) pari_err(typeer, "ideallistarch");
+  if (typ(L) != t_VEC) pari_err_TYPE("ideallistarch",L);
   if (l == 1) return cgetg(1,t_VEC);
   z = gel(L,1);
-  if (typ(z) != t_VEC) pari_err(typeer, "ideallistarch");
+  if (typ(z) != t_VEC) pari_err_TYPE("ideallistarch",z);
   z = gel(z,1); /* either a bid or [bid,U] */
   if (lg(z) == 3) { /* the latter: do units */
-    if (typ(z) != t_VEC) pari_err(typeer,"ideallistarch");
+    if (typ(z) != t_VEC) pari_err_TYPE("ideallistarch",z);
     ID.sgnU = nfsign_units(bnf, NULL, 1);
     join_z = &join_archunit;
   } else
