@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -175,18 +173,18 @@ qfi_random(struct buch_quad *B,GEN ex) { return random_form(B, ex, &QFI_comp); }
 /*                     Common subroutines                          */
 /*                                                                 */
 /*******************************************************************/
-double
-check_bach(double cbach, double B)
+long
+check_LIMC(long LIMC, long LIMCMAX)
 {
-  if (cbach >= B)
-   pari_err(talker,"sorry, couldn't deal with this field. PLEASE REPORT");
-  if (cbach <= 0.3)
-    cbach *= 2;
+  if (LIMC >= LIMCMAX) pari_err_BUG("Buchmann's algorithm");
+  if (LIMC <= LIMCMAX/40) /* cbach <= 0.3 */
+    LIMC *= 2;
+  else if (LIMCMAX < 60) /* \Delta_K <= 9 */
+    LIMC++;
   else
-    cbach += 0.2;
-  if (cbach > B) cbach = B;
-  if (DEBUGLEVEL) err_printf("\n*** Bach constant: %f\n", cbach);
-  return cbach;
+    LIMC += LIMCMAX / 60; /* cbach += 0.2 */
+  if (LIMC > LIMCMAX) LIMC = LIMCMAX;
+  return LIMC;
 }
 
 /* Is |q| <= p ? */
@@ -285,6 +283,68 @@ clearhash(long **hash)
   }
 }
 
+/* last prime stored */
+ulong
+GRH_last_prime(GRHcheck_t *S) { return (S->primes + S->nprimes-1)->p; }
+/* ensure that S->primes can hold at least nb primes */
+void
+GRH_ensure(GRHcheck_t *S, long nb)
+{
+  if (S->maxprimes <= nb)
+  {
+    do S->maxprimes *= 2; while (S->maxprimes <= nb);
+    S->primes = (GRHprime_t*)pari_realloc((void*)S->primes,
+                                          S->maxprimes*sizeof(*S->primes));
+  }
+}
+/* cache data for all primes up to the LIM */
+static void
+cache_prime_quad(GRHcheck_t *S, ulong LIM, GEN D)
+{
+  GRHprime_t *pr;
+  double nb;
+
+  if (S->limp >= LIM) return;
+  nb = primepi_upper_bound((double)LIM); /* #{p <= LIM} <= nb */
+  GRH_ensure(S, nb+1); /* room for one extra prime */
+  for (pr = S->primes + S->nprimes;;)
+  {
+    ulong p = u_forprime_next(&(S->P));
+    pr->p = p;
+    pr->logp = log((double)p);
+    pr->dec = (GEN)kroiu(D,p);
+    S->nprimes++;
+    pr++;
+    /* store up to nextprime(LIM) included */
+    if (p >= LIM) { S->limp = p; break; }
+  }
+}
+
+static GEN
+compute_invresquad(GRHcheck_t *S)
+{
+  pari_sp av = avma;
+  GEN invres = real_1(DEFAULTPREC);
+  GRHprime_t *pr = S->primes;
+  long i = S->nprimes, LIMC = GRH_last_prime(S)+diffptr[i]-1; /* nextprime(p+1)-1*/
+  double limp = log(LIMC) / 2;
+  for (; i > 0; pr++, i--)
+  {
+    long s = (long)pr->dec;
+    if (s)
+    {
+      ulong p = pr->p;
+      if (s>0 || pr->logp <= limp)
+        /* Both p and P contribute */
+        invres = mulur(p - s, divru(invres, p));
+      else if (s<0)
+        /* Only p contributes */
+        invres = mulur(p, divru(invres, p - 1));
+    }
+  }
+  return gerepileuptoleaf(av, invres);
+}
+
 /* p | conductor of order of disc D ? */
 static int
 is_bad(GEN D, ulong p)
@@ -301,85 +361,108 @@ is_bad(GEN D, ulong p)
   avma = av; return r;
 }
 
-/* create B->FB, B->numFB; set B->badprim. Return L(kro_D, 1) */
-static GEN
-FBquad(struct buch_quad *B, long C2, long C1, GRHcheck_t *S)
+/* returns the n-th suitable ideal for the factorbase */
+static long
+nthidealquad(GEN D, long n)
 {
-  GEN Res = real_1(DEFAULTPREC), D = B->QFR->D;
-  double L = log((double)C2), SA = 0, SB = 0;
-  long i, p, s, LIM;
-  pari_sp av;
-  byteptr d = diffptr;
+  pari_sp av = avma;
+  forprime_t S;
+  ulong p;
+  (void)u_forprime_init(&S, 2, ULONG_MAX);
+  while (n > 0)
+  {
+    p = u_forprime_next(&S);
+    if (!is_bad(D, p) && kroiu(D, p) >= 0) n--;
+  }
+  avma = av; return p;
+}
 
+static int
+quadGRHchk(GEN D, GRHcheck_t *S, ulong LIMC)
+{
+  double logC = log((double)LIMC), SA = 0, SB = 0;
+  long i;
+
+  cache_prime_quad(S, LIMC, D);
+  for (i = 0;; i++)
+  {
+    GRHprime_t *pr = S->primes+i;
+    ulong p = pr->p;
+    long M;
+    double logNP, q, A, B;
+    if (p > LIMC) break;
+    if ((long)pr->dec < 0)
+    {
+      logNP = 2 * pr->logp;
+      q = 1/(double)p;
+    }
+    else
+    {
+      logNP = pr->logp;
+      q = 1/sqrt((double)p);
+    }
+    A = logNP * q; B = logNP * A; M = (long)(logC/logNP);
+    if (M > 1)
+    {
+      double inv1_q = 1 / (1-q);
+      A *= (1 - pow(q, M)) * inv1_q;
+      B *= (1 - pow(q, M)*(M+1 - M*q)) * inv1_q * inv1_q;
+    }
+    if ((long)pr->dec>0) { SA += 2*A;SB += 2*B; } else { SA += A; SB += B; }
+  }
+  return GRHok(S, logC, SA, SB);
+}
+
+/* create B->FB, B->numFB; set B->badprim. Return L(kro_D, 1) */
+static void
+FBquad(struct buch_quad *B, ulong C2, ulong C1, GRHcheck_t *S)
+{
+  GEN D = B->QFR->D;
+  long i;
+  pari_sp av;
+  GRHprime_t *pr;
+
+  cache_prime_quad(S, C2, D);
+  pr = S->primes;
   B->numFB = cgetg(C2+1, t_VECSMALL);
   B->FB    = cgetg(C2+1, t_VECSMALL);
   av = avma;
   B->KC = 0; i = 0;
-  maxprime_check((ulong)C2);
   B->badprim = gen_1;
-  for (p = 0;;) /* p <= C2 */
+  for (;; pr++) /* p <= C2 */
   {
-    NEXT_PRIME_VIADIFF(p, d);
+    ulong p = pr->p;
     if (!B->KC && p > C1) B->KC = i;
     if (p > C2) break;
-    s = krois(D,p);
-    if (s) Res = mulur(p, divru(Res, p - s));
-    switch (s)
+    switch ((long)pr->dec)
     {
       case -1: break; /* inert */
       case  0: /* ramified */
-        if (is_bad(D, (ulong)p)) { B->badprim = muliu(B->badprim, p); break; }
+        if (is_bad(D, p)) { B->badprim = muliu(B->badprim, p); break; }
         /* fall through */
       default:  /* split */
         i++; B->numFB[p] = i; B->FB[i] = p; break;
     }
-    if (!S->checkok)
-    {
-      double logp = log((double)p);
-      double logNP = s < 0? 2*logp: logp;
-      double q = s < 0? 1/(double)p: 1/sqrt((double)p);
-      double A = logNP * q, B = logNP * A;
-      long M = (long)(L/logNP);
-      if (M > 1)
-      {
-        double inv1_q = 1 / (1-q);
-        A *= (1 - pow(q, M)) * inv1_q;
-        B *= (1 - pow(q, M)*(M+1 - M*q)) * inv1_q * inv1_q;
-      }
-      if (s > 0) { SA += 2 * A; SB += 2 * B; } else { SA += A; SB += B; }
-    }
   }
-  if (!B->KC) return NULL;
-  if (!GRHok(S, L, SA, SB)) return NULL;
   B->KC2 = i;
   setlg(B->FB, B->KC2+1);
-  LIM = (expi(D) < 16)? 100: 1000;
-  while (p < LIM)
-  {
-    s = krois(D,p);
-    Res = mulur(p, divru(Res, p - s));
-    NEXT_PRIME_VIADIFF(p, d);
-  }
   if (B->badprim != gen_1)
-    gerepileall(av, 2, &Res, &B->badprim);
+    B->badprim = gerepileuptoint(av, B->badprim);
   else
   {
-    B->badprim = NULL;
-    Res = gerepileuptoleaf(av, Res);
+    B->badprim = NULL; avma = av;
   }
-  S->checkok = 1; return Res;
 }
 
 /* create B->vperm, return B->subFB */
 static GEN
-subFBquad(struct buch_quad *B, GEN D, double PROD)
+subFBquad(struct buch_quad *B, GEN D, double PROD, long minSFB)
 {
-  long i, j, minSFB, lgsub = 1, ino = 1, lv = B->KC+1;
+  long i, j, lgsub = 1, ino = 1, lv = B->KC+1;
   double prod = 1.;
   pari_sp av;
   GEN no;
 
-  minSFB = (expi(D) > 15)? 3: 2;
   B->vperm = cgetg(lv, t_VECSMALL);
   av = avma;
   no    = cgetg(lv, t_VECSMALL);
@@ -394,7 +477,7 @@ subFBquad(struct buch_quad *B, GEN D, double PROD)
       if (lgsub > minSFB && prod > PROD) break;
     }
   }
-  if (j == lv) return NULL;
+  /* lgsub >= 1 otherwise quadGRHchk is false */
   i = lgsub;
   for (j = 1; j < ino;i++,j++) B->vperm[i] = no[j];
   for (     ; i < lv; i++)     B->vperm[i] = i;
@@ -505,7 +588,7 @@ trivial_relations(struct buch_quad *B, GEN mat, GEN C)
   for (i = 1; i <= B->KC; i++)
   { /* ramified prime ==> trivial relation */
     if (umodiu(D, B->FB[i])) continue;
-    col = const_vecsmall(B->KC, 0);
+    col = zero_zv(B->KC);
     col[i] = 2; j++;
     gel(mat,j) = col;
     gel(C,j) = gen_0;
@@ -576,12 +659,14 @@ imag_relations(struct buch_quad *B, long need, long *pc, ulong LIMC, GEN mat)
         for (i=1; i<lgsub; i++) col[B->subFB[i]] += -fpd[i]-ex[i];
         add_fact(B, col, form2); col[fpd[-2]]--;
       }
+      if (DEBUGLEVEL) err_printf(" %ldP",s);
     }
     else
     {
       col = gel(mat,++s);
       for (i=1; i<lgsub; i++) col[B->subFB[i]] = -ex[i];
       add_fact(B, col, form);
+      if (DEBUGLEVEL) err_printf(" %ld",s);
     }
     col[current]--;
     if (++current > B->KC) current = 1;
@@ -867,11 +952,11 @@ Buchquad(GEN D, double cbach, double cbach2, long prec)
   pari_timer T;
   pari_sp av0 = avma, av, av2;
   const long RELSUP = 5;
-  long i, s, current, triv, sfb_trials, nrelsup, nreldep, need, nsubFB;
-  ulong LIMC, LIMC2, cp;
-  GEN W, cyc, res, gen, dep, mat, C, extraC, B, R, resc, Res, z, h = NULL; /*-Wall*/
-  double drc, lim, LOGD, LOGD2;
-  GRHcheck_t G, *GRHcheck = &G;
+  long i, s, current, triv, sfb_trials, nrelsup, nreldep, need, nsubFB, minSFB;
+  ulong low, high, LIMC0, LIMC, LIMC2, LIMCMAX, cp;
+  GEN W, cyc, res, gen, dep, mat, C, extraC, B, R, invhr, h = NULL; /*-Wall*/
+  double drc, sdrc, lim, LOGD, LOGD2;
+  GRHcheck_t GRHcheck;
   struct qfr_data QFR;
   struct buch_quad BQ;
   int FIRST = 1;
@@ -890,7 +975,7 @@ Buchquad(GEN D, double cbach, double cbach2, long prec)
     }
     BQ.PRECREG = 0;
   } else {
-    BQ.PRECREG = maxss(prec+1, MEDDEFAULTPREC + 2*(expi(QFR.D)/BITS_IN_LONG));
+    BQ.PRECREG = maxss(prec+EXTRAPRECWORD, nbits2prec(2*expi(QFR.D) + 128));
   }
   if (DEBUGLEVEL) timer_start(&T);
   BQ.primfact   = new_chunk(100);
@@ -902,10 +987,7 @@ Buchquad(GEN D, double cbach, double cbach2, long prec)
   LOGD = log(drc);
   LOGD2 = LOGD * LOGD;
 
-  lim = sqrt(drc);
-  /* resc = sqrt(D) w / 2^r1 (2pi)^r2 ~ hR / L(chi,1) */
-  if (BQ.PRECREG) resc = dbltor(lim / 2.);
-  else         resc = dbltor(lim / PI);
+  sdrc = lim = sqrt(drc);
   if (!BQ.PRECREG) lim /= sqrt(3.);
   cp = (ulong)exp(sqrt(LOGD*log(LOGD)/8.0));
   if (cp < 20) cp = 20;
@@ -913,33 +995,63 @@ Buchquad(GEN D, double cbach, double cbach2, long prec)
     if (cbach2 < cbach) cbach2 = cbach;
     cbach = 6.;
   }
-  if (cbach <= 0.) pari_err(talker,"Bach constant <= 0 in Buchquad");
+  if (cbach < 0.)
+    pari_err_DOMAIN("Buchquad","Bach constant","<",gen_0,dbltor(cbach));
   av = avma;
   BQ.powsubFB = BQ.subFB = NULL;
-  init_GRHcheck(GRHcheck, 2, BQ.PRECREG? 2: 0, LOGD);
+  minSFB = (expi(D) > 15)? 3: 2;
+  init_GRHcheck(&GRHcheck, 2, BQ.PRECREG? 2: 0, LOGD);
+  high = low = LIMC0 = maxss((long)(cbach2*LOGD2), 1);
+  LIMCMAX = (long)(6.*LOGD2);
+  /* 97/1223 below to ensure a good enough approximation of residue */
+  cache_prime_quad(&GRHcheck, expi(D) < 16 ? 97: 1223, D);
+  while (!quadGRHchk(D, &GRHcheck, high))
+  {
+    low = high;
+    high *= 2;
+  }
+  while (high - low > 1)
+  {
+    long test = (low+high)/2;
+    if (quadGRHchk(D, &GRHcheck, test))
+      high = test;
+    else
+      low = test;
+  }
+  if (high == LIMC0+1 && quadGRHchk(D, &GRHcheck, LIMC0))
+    LIMC2 = LIMC0;
+  else
+    LIMC2 = high;
+  if (LIMC2 > LIMCMAX) LIMC2 = LIMCMAX;
+  LIMC0 = (long)(cbach*LOGD2);
+  LIMC = cbach ? LIMC0 : LIMC2;
+  LIMC = maxss(LIMC, nthidealquad(D, 2));
 
 /* LIMC = Max(cbach*(log D)^2, exp(sqrt(log D loglog D) / 8)) */
 START:
   do
   {
-    if (!FIRST) cbach = check_bach(cbach,6.);
+    if (!FIRST) LIMC = check_LIMC(LIMC,LIMCMAX);
+    if (DEBUGLEVEL && LIMC > LIMC0)
+      err_printf("%s*** Bach constant: %f\n", FIRST?"":"\n", LIMC/LOGD2);
     FIRST = 0; avma = av;
     if (BQ.subFB) gunclone(BQ.subFB);
     if (BQ.powsubFB) gunclone(BQ.powsubFB);
     clearhash(BQ.hashtab);
-    LIMC = (ulong)(cbach*LOGD2);
-    if (LIMC < cp) { LIMC = cp; cbach = (double)LIMC / LOGD2; }
-    LIMC2 = (ulong)(maxdd(cbach,cbach2)*LOGD2);
+    if (LIMC < cp) LIMC = cp;
     if (LIMC2 < LIMC) LIMC2 = LIMC;
     if (BQ.PRECREG) qfr_data_init(QFR.D, BQ.PRECREG, &QFR);
 
-    Res = FBquad(&BQ, LIMC2, LIMC, GRHcheck);
+    FBquad(&BQ, LIMC2, LIMC, &GRHcheck);
     if (DEBUGLEVEL) timer_printf(&T, "factor base");
+    BQ.subFB = subFBquad(&BQ, QFR.D, lim + 0.5, minSFB);
+    if (DEBUGLEVEL) timer_printf(&T, "subFBquad = %Ps",
+                                 vecpermute(BQ.FB, BQ.subFB));
+    nsubFB = lg(BQ.subFB) - 1;
   }
-  while (!Res || !(BQ.subFB = subFBquad(&BQ, QFR.D, lim + 0.5)));
-  if (DEBUGLEVEL) timer_printf(&T, "subFBquad = %Ps",
-                               vecpermute(BQ.FB, BQ.subFB));
-  nsubFB = lg(BQ.subFB) - 1;
+  while (nsubFB < (expi(D) > 15 ? 3 : 2));
+  /* invhr = 2^r1 (2pi)^r2 / sqrt(D) w ~ L(chi,1) / hR */
+  invhr = gmul(dbltor((BQ.PRECREG?2.:PI)/sdrc), compute_invresquad(&GRHcheck));
   BQ.powsubFB = powsubFBquad(&BQ,CBUCH+1);
   if (DEBUGLEVEL) timer_printf(&T, "powsubFBquad");
   BQ.limhash = (LIMC & HIGHMASK)? (HIGHBIT>>1): LIMC*LIMC;
@@ -975,13 +1087,13 @@ START:
     }
     if (BQ.PRECREG) {
       for (i = triv+1; i<=need; i++) {
-        gel(mat,i) = const_vecsmall(BQ.KC, 0);
+        gel(mat,i) = zero_zv(BQ.KC);
         gel(extraC,i) = cgetr(BQ.PRECREG);
       }
       real_relations(&BQ, need - triv, &current, s,LIMC,mat + triv,extraC + triv);
     } else {
       for (i = triv+1; i<=need; i++) {
-        gel(mat,i) = const_vecsmall(BQ.KC, 0);
+        gel(mat,i) = zero_zv(BQ.KC);
         gel(extraC,i) = gen_0;
       }
       imag_relations(&BQ, need - triv, &current, LIMC,mat + triv);
@@ -992,7 +1104,7 @@ START:
     else
       W = hnfadd_i(W,BQ.vperm,&dep,&B,&C, mat,extraC);
     gerepileall(av2, 4, &W,&C,&B,&dep);
-    need = lg(dep)>1? lg(dep[1])-1: lg(B[1])-1;
+    need = BQ.KC - (lg(W)-1) - (lg(B)-1);
     if (need)
     {
       if (++nreldep > 15 && cbach < 1) goto START;
@@ -1002,11 +1114,10 @@ START:
     h = ZM_det_triangular(W);
     if (DEBUGLEVEL) err_printf("\n#### Tentative class number: %Ps\n", h);
 
-    z = mulrr(Res, resc); /* ~ hR if enough relations, a multiple otherwise */
-    switch(get_R(&BQ, C, (lg(C)-1) - (lg(B)-1) - (lg(W)-1), divir(h,z), &R))
+    switch(get_R(&BQ, C, (lg(C)-1) - (lg(B)-1) - (lg(W)-1), mulir(h,invhr), &R))
     {
       case fupb_PRECI:
-        BQ.PRECREG = (BQ.PRECREG<<1)-2;
+        BQ.PRECREG = precdbl(BQ.PRECREG);
         FIRST = 1; goto START;
 
       case fupb_RELAT:
@@ -1023,6 +1134,7 @@ START:
   if (!quad_be_honest(&BQ)) goto START;
   if (DEBUGLEVEL) timer_printf(&T, "be honest");
   clearhash(BQ.hashtab);
+  free_GRHcheck(&GRHcheck);
 
   gen = get_clgp(&BQ,W,&cyc,BQ.PRECREG);
   gunclone(BQ.subFB);
@@ -1040,7 +1152,7 @@ buchimag(GEN D, GEN c, GEN c2, GEN REL)
 
 GEN
 buchreal(GEN D, GEN flag, GEN c, GEN c2, GEN REL, long prec) {
-  if (signe(flag)) pari_err(impl,"narrow class group");
+  if (signe(flag)) pari_err_IMPL("narrow class group");
   (void)REL; return Buchquad(D,gtodouble(c),gtodouble(c2),prec);
 }
 
@@ -1048,14 +1160,14 @@ GEN
 quadclassunit0(GEN x, long flag, GEN data, long prec)
 {
   long lx;
-  double c1 = 0.2, c2 = 0.2;
+  double c1 = 0.0, c2 = 0.0;
 
   if (!data) lx=1;
   else
   {
     lx = lg(data);
-    if (typ(data)!=t_VEC || lx > 7)
-      pari_err(talker,"incorrect parameters in quadclassunit");
+    if (typ(data)!=t_VEC) pari_err_TYPE("quadclassunit", data);
+    if (lx > 7) pari_err_DIM("quadclassunit [tech vector]");
     if (lx > 3) lx = 3;
   }
   switch(lx)
@@ -1063,6 +1175,6 @@ quadclassunit0(GEN x, long flag, GEN data, long prec)
     case 3: c2 = gtodouble(gel(data,2));
     case 2: c1 = gtodouble(gel(data,1));
   }
-  if (flag) pari_err(impl,"narrow class group");
+  if (flag) pari_err_IMPL("narrow class group");
   return Buchquad(x,c1,c2,prec);
 }
