@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -102,7 +100,8 @@ hyperell_locally_soluble(GEN T,GEN p)
 {
   pari_sp av = avma;
   long res;
-  if (typ(T)!=t_POL || typ(p)!=t_INT) pari_err(typeer,"zpsoluble");
+  if (typ(T)!=t_POL) pari_err_TYPE("zpsoluble",T);
+  if (typ(p)!=t_INT) pari_err_TYPE("zpsoluble",p);
   RgX_check_ZX(T, "zpsoluble");
   res = zpsol(T,p,0,gen_1,gen_0) || zpsol(RgX_recip_shallow(T), p, 1, p, gen_0);
   avma = av; return res;
@@ -121,7 +120,7 @@ quad_char(GEN nf, GEN t, GEN pr)
     t = Fq_pow(t, diviiexact(ord, ordp), T,p); /* in F_p^* */
     if (typ(t) == t_POL)
     {
-      if (degpol(t)) pari_err(bugparier,"nfhilbertp");
+      if (degpol(t)) pari_err_BUG("nfhilbertp");
       t = gel(t,2);
     }
   }
@@ -152,7 +151,7 @@ psquarenf(GEN nf,GEN x,GEN pr,GEN modpr)
     if (v&1) return 0;
     v = (Z_quad_char(x, pr) == 1);
   } else {
-    v = int_elt_val(nf, x, p, pr_get_tau(pr), &x);
+    v = ZC_nfvalrem(nf, x, pr, &x);
     if (v&1) return 0;
     v = (quad_char(nf, x, modpr) == 1);
   }
@@ -183,7 +182,7 @@ psquare2nf(GEN nf,GEN x,GEN pr,GEN zinit)
     if (!signe(x)) return 1;
     v = Z_lvalrem(x, 2, &x) * pr_get_e(pr);
   } else
-    v = int_elt_val(nf, x, pr_get_p(pr), pr_get_tau(pr), &x);
+    v = ZC_nfvalrem(nf, x, pr, &x);
   if (v&1) return 0;
   /* now (x,pr) = 1 */
   v = check2(nf,x,zinit); avma = av; return v;
@@ -218,7 +217,7 @@ lemma7nf(GEN nf, GEN T, GEN pr, long nu, GEN x, GEN zinit)
 
   gpx = nfpoleval(nf, RgX_deriv(T), x);
   /* gx /= pi^la, pi a pr-uniformizer */
-  la = int_elt_val(nf, gx, pr_get_p(pr), pr_get_tau(pr), &gx);
+  la = ZC_nfvalrem(nf, gx, pr, &gx);
   mu = gequal0(gpx)? la+nu+1: nfval(nf,gpx,pr);
 
   if (la > (mu<<1)) return 1;
@@ -308,12 +307,9 @@ nf_hyperell_locally_soluble(GEN nf,GEN T,GEN pr)
   GEN repr, zinit, p1;
   pari_sp av = avma;
 
-  if (typ(T)!=t_POL) pari_err(notpoler,"nf_hyperell_locally_soluble");
+  if (typ(T)!=t_POL) pari_err_TYPE("nf_hyperell_locally_soluble",T);
   if (gequal0(T)) return 1;
   checkprid(pr); nf = checknf(nf);
-  pr = shallowcopy(pr);
-  gel(pr,5) = zk_scalar_or_multable(nf, pr_get_tau(pr));
-
   if (equaliu(pr_get_p(pr), 2))
   { /* tough case */
     zinit = Idealstar(nf, idealpows(nf,pr,1+2*pr_get_e(pr)), nf_INIT);
@@ -349,7 +345,7 @@ den_remove(GEN nf, GEN a)
       if (da) a = ZC_Z_mul(a, da);
       a = coltoliftalg(nf, a);
       return a;
-    default: pari_err(typeer,"nfhilbert");
+    default: pari_err_TYPE("nfhilbert",a);
       return NULL;/*not reached*/
   }
 }
@@ -457,7 +453,8 @@ nfhilbert0(GEN nf,GEN a,GEN b,GEN p)
   nf = checknf(nf);
   if (p) {
     checkprid(p);
-    if (gequal0(a) || gequal0(b)) pari_err (talker,"0 argument in nfhilbertp");
+    if (gequal0(a)) pari_err_DOMAIN("nfhilbert", "a", "=", gen_0, a);
+    if (gequal0(b)) pari_err_DOMAIN("nfhilbert", "b", "=", gen_0, b);
     return nfhilbertp(nf,a,b,p);
   }
   return nfhilbert(nf,a,b);
@@ -478,7 +475,7 @@ bnfsunit(GEN bnf,GEN S,long prec)
   GEN p1,nf,gen,M,U,H;
   GEN sunit,card,sreg,res,pow;
 
-  if (typ(S) != t_VEC) pari_err(typeer,"bnfsunit");
+  if (!is_vec_t(typ(S))) pari_err_TYPE("bnfsunit",S);
   bnf = checkbnf(bnf);
   nf = bnf_get_nf(bnf);
   gen = bnf_get_gen(bnf);
@@ -528,7 +525,7 @@ bnfsunit(GEN bnf,GEN S,long prec)
     H = mathnfspec(U1,&perm,&dep,&B,&p1);
     lH = lg(H);
     lB = lg(B);
-    if (lg(dep) > 1 && lg(dep[1]) > 1) pari_err(bugparier,"bnfsunit");
+    if (lg(dep) > 1 && lgcols(dep) > 1) pari_err_BUG("bnfsunit");
    /*                   [ H B  ]            [ H^-1   - H^-1 B ]
     * perm o HNF(U1) =  [ 0 Id ], inverse = [  0         Id   ]
     * (permute the rows)
@@ -569,7 +566,7 @@ static GEN
 make_unit(GEN nf, GEN bnfS, GEN *px)
 {
   long lB, cH, i, ls;
-  GEN den, gen, S, v, p1, xp, xb, N, HB, perm;
+  GEN den, gen, S, v, p1, xp, xb, N, N0, HB, perm;
 
   if (gequal0(*px)) return NULL;
   S = gel(bnfS,6); ls = lg(S);
@@ -588,14 +585,20 @@ make_unit(GEN nf, GEN bnfS, GEN *px)
   perm = gel(p1,1);
   HB   = gel(p1,2);
   den  = gel(p1,3);
-  cH = lg(HB[1]) - 1;
+  cH = nbrows(HB);
   lB = lg(HB) - cH;
-  v = cgetg(ls, t_VECSMALL);
+  v = zero_zv(ls-1);
+  N0 = N;
   for (i=1; i<ls; i++)
   {
-    GEN P = gel(S,i);
-    v[i] = (remii(N, pr_get_p(P)) == gen_0)? nfval(nf,xb,P): 0;
+    GEN P = gel(S,i), p = pr_get_p(P);
+    if ( remii(N, p) == gen_0 )
+    {
+      v[i] = nfval(nf,xb,P);
+      (void)Z_pvalrem(N0, p, &N0);
+    }
   }
+  if (!is_pm1(N0)) return NULL;
   /* here, x = S v */
   p1 = vecpermute(v, perm);
   v = ZM_zc_mul(HB, p1);
@@ -633,7 +636,7 @@ bnfissunit(GEN bnf,GEN bnfS,GEN x)
 
   bnf = checkbnf(bnf);
   nf = bnf_get_nf(bnf);
-  if (typ(bnfS)!=t_VEC || lg(bnfS)!=7) pari_err(typeer,"bnfissunit");
+  if (typ(bnfS)!=t_VEC || lg(bnfS)!=7) pari_err_TYPE("bnfissunit",bnfS);
   x = nf_to_scalar_or_alg(nf,x);
   v = NULL;
   if ( (w = make_unit(nf, bnfS, &x)) ) v = bnfisunit(bnf, x);
@@ -674,60 +677,72 @@ fa_pr_append(GEN nf,GEN rel,GEN N,GEN *prod,GEN *S1,GEN *S2)
     Zfa_pr_append(nf,rel,N,prod,S1,S2);
 }
 
+/* apply lift(rnfeltup) to all coeffs, without rnf structure */
 static GEN
-pol_up(GEN rnfeq, GEN x, long v)
+nfX_eltup(GEN nf, GEN rnfeq, GEN x)
 {
-  long i, l = lg(x);
-  GEN y = cgetg(l, t_POL); y[1] = x[1];
+  long i, l;
+  GEN zknf, czknf, y = cgetg_copy(x, &l);
+  y[1] = x[1]; nf_nfzk(nf, rnfeq, &zknf, &czknf);
+  for (i=2; i<l; i++) gel(y,i) = nfeltup(nf, gel(x,i), zknf, czknf);
+  return y;
+}
+/* FIXME: remove this */
+static void
+nfX_fix_var(GEN P, long v)
+{
+  long i, l = lg(P);
   for (i=2; i<l; i++)
   {
-    GEN t = eltreltoabs(rnfeq, gel(x,i));
-    if (typ(t) == t_POL) setvarn(t, v);
-    gel(y,i) = t;
+    GEN c = gel(P,i);
+    if (typ(c) == t_POL) setvarn(c, v);
   }
-  return y;
 }
 
 GEN
 rnfisnorminit(GEN T, GEN relpol, int galois)
 {
   pari_sp av = avma;
-  long i, l, drel, vbas;
+  long i, l, drel;
   GEN prod, S1, S2, gen, cyc, bnf, nf, nfabs, rnfeq, bnfabs, k, polabs;
   GEN y = cgetg(9, t_VEC);
 
-  if (galois < 0 || galois > 2) pari_err(flagerr, "rnfisnorminit");
-  T = get_bnfpol(T, &bnf, &nf); vbas = varn(T);
+  if (galois < 0 || galois > 2) pari_err_FLAG("rnfisnorminit");
+  T = get_bnfpol(T, &bnf, &nf);
   if (!bnf) bnf = Buchall(nf? nf: T, nf_FORCE, DEFAULTPREC);
   if (!nf) nf = bnf_get_nf(bnf);
 
   relpol = get_bnfpol(relpol, &bnfabs, &nfabs);
-  if (!gequal1(leading_term(relpol))) pari_err(impl,"non monic relative equation");
+  if (!gequal1(leading_term(relpol))) pari_err_IMPL("non monic relative equation");
   drel = degpol(relpol);
-  if (varncmp(varn(relpol), vbas) >= 0)
-    pari_err(talker,"main variable must be of higher priority in rnfisnorminit");
   if (drel <= 2) galois = 1;
 
-  rnfeq = NULL; /* no reltoabs needed */
-  if (nf_get_degree(nf) == 1) { /* over Q */
-    polabs = simplify_shallow(lift(relpol)); k = gen_0;
-  } else if (galois == 2) { /* needs reltoabs */
-    rnfeq = rnfequation2(bnf, relpol);
-    polabs = gel(rnfeq,1);
-    gel(rnfeq,2) = lift_intern(gel(rnfeq,2));
-    k = gel(rnfeq,3);
-  } else {
-    long sk;
-    polabs = rnfequationall(bnf, relpol, &sk, NULL);
-    k = stoi(sk);
-  }
-  if (!bnfabs || !gequal0(k)) bnfabs = Buchall(polabs, nf_FORCE, nf_get_prec(nf));
+  relpol = RgX_nffix("rnfisnorminit", T, relpol, 1);
+  if (nf_get_degree(nf) == 1) /* over Q */
+    rnfeq = mkvec5(relpol,gen_0,gen_0,T,relpol);
+  else if (galois == 2) /* needs eltup+abstorel */
+    rnfeq = nf_rnfeq(nf, relpol);
+  else /* needs abstorel */
+    rnfeq = nf_rnfeqsimple(nf, relpol);
+  polabs = gel(rnfeq,1);
+  k = gel(rnfeq,3);
+  if (!bnfabs || !gequal0(k))
+    bnfabs = Buchall(polabs, nf_FORCE, nf_get_prec(nf));
   if (!nfabs) nfabs = bnf_get_nf(bnfabs);
 
   if (galois == 2)
   {
-    GEN P = rnfeq? pol_up(rnfeq, relpol, vbas): relpol;
-    galois = nfissplit(gsubst(nfabs, nf_get_varn(nfabs), pol_x(vbas)), P);
+    GEN P;
+    long v = varn(T);
+    if (polabs == relpol)
+      P = relpol;
+    else
+    { /* FIXME: don't mess with variables, use proper priorities in nfabs. */
+      P = nfX_eltup(nf, rnfeq, relpol);
+      nfX_fix_var(P, v);
+    }
+    /* FIXME */
+    galois = nfissplit(gsubst(nfabs, nf_get_varn(nfabs), pol_x(v)), P);
   }
 
   prod = gen_1; S1 = S2 = cgetg(1, t_VEC);
@@ -737,18 +752,18 @@ rnfisnorminit(GEN T, GEN relpol, int galois)
   {
     GEN g = gel(gen,i);
     if (ugcd(umodiu(gel(cyc,i), drel), drel) == 1) break;
-    fa_pr_append(nf,bnfabs,gcoeff(g,1,1),&prod,&S1,&S2);
+    Zfa_pr_append(nf,bnfabs,gcoeff(g,1,1),&prod,&S1,&S2);
   }
   if (!galois)
   {
     GEN Ndiscrel = diviiexact(nf_get_disc(nfabs), powiu(nf_get_disc(nf), drel));
-    fa_pr_append(nf,bnfabs,absi(Ndiscrel), &prod,&S1,&S2);
+    Zfa_pr_append(nf,bnfabs,absi(Ndiscrel), &prod,&S1,&S2);
   }
 
   gel(y,1) = bnf;
   gel(y,2) = bnfabs;
   gel(y,3) = relpol;
-  gel(y,4) = get_theta_abstorel(T, relpol, k);
+  gel(y,4) = rnfeq;
   gel(y,5) = prod;
   gel(y,6) = S1;
   gel(y,7) = S2;
@@ -766,13 +781,16 @@ GEN
 rnfisnorm(GEN T, GEN x, long flag)
 {
   pari_sp av = avma;
-  GEN bnf, rel, relpol, theta, nfpol;
+  GEN bnf, rel, relpol, rnfeq, nfpol;
   GEN nf, aux, H, U, Y, M, A, bnfS, sunitrel, futu, prod, S1, S2;
   long L, i, drel, itu;
 
   if (typ(T) != t_VEC || lg(T) != 9)
-    pari_err(talker,"please apply rnfisnorminit first");
-  bnf = gel(T,1); rel = gel(T,2); relpol = gel(T,3); theta = gel(T,4);
+    pari_err_TYPE("rnfisnorm [please apply rnfisnorminit()]", T);
+  bnf = gel(T,1);
+  rel = gel(T,2);
+  relpol = gel(T,3);
+  rnfeq = gel(T,4);
   drel = degpol(relpol);
   bnf = checkbnf(bnf);
   rel = checkbnf(rel);
@@ -791,37 +809,32 @@ rnfisnorm(GEN T, GEN x, long flag)
     pari_warn(warner,"useless flag in rnfisnorm: the extension is Galois");
   if (flag > 0)
   {
-    byteptr d = diffptr;
-    long p = 0;
-    maxprime_check((ulong)flag);
-    for(;;)
-    {
-      NEXT_PRIME_VIADIFF(p, d);
-      if (p > flag) break;
+    forprime_t T;
+    ulong p;
+    u_forprime_init(&T, 2, flag);
+    while ( (p = u_forprime_next(&T)) )
       pr_append(nf,rel, utoipos(p),&prod,&S1,&S2);
-    }
   }
   else if (flag < 0)
-    fa_pr_append(nf,rel,stoi(-flag),&prod,&S1,&S2);
+    Zfa_pr_append(nf,rel, utoipos(-flag),&prod,&S1,&S2);
   /* overkill: prime ideals dividing x would be enough */
   fa_pr_append(nf,rel,idealnorm(nf,x), &prod,&S1,&S2);
 
   /* computation on T-units */
   futu = shallowconcat(bnf_get_fu(rel), bnf_get_tuU(rel));
-  bnfS = bnfsunit(bnf,S1,3);
-  sunitrel = shallowconcat(futu, gel(bnfsunit(rel,S2,3), 1));
+  bnfS = bnfsunit(bnf,S1,LOWDEFAULTPREC);
+  sunitrel = shallowconcat(futu, gel(bnfsunit(rel,S2,LOWDEFAULTPREC), 1));
 
   A = lift_intern(bnfissunit(bnf,bnfS,x));
   L = lg(sunitrel);
-  itu = lg(nf[6])-1; /* index of torsion unit in bnfsunit(nf) output */
+  itu = lg(nf_get_roots(nf))-1; /* index of torsion unit in bnfsunit(nf) output */
   M = cgetg(L+1,t_MAT);
   for (i=1; i<L; i++)
   {
-    GEN u = poleval(gel(sunitrel,i), theta); /* abstorel */
-    if (typ(u) != t_POLMOD) u = mkpolmod(u, gel(theta,1));
+    GEN u = eltabstorel(rnfeq, gel(sunitrel,i));
     gel(sunitrel,i) = u;
     u = bnfissunit(bnf,bnfS, gnorm(u));
-    if (lg(u) == 1) pari_err(bugparier,"rnfisnorm");
+    if (lg(u) == 1) pari_err_BUG("rnfisnorm");
     gel(u,itu) = lift_intern(gel(u,itu)); /* lift root of 1 part */
     gel(M,i) = u;
   }
@@ -832,14 +845,11 @@ rnfisnorm(GEN T, GEN x, long flag)
   /* Y: sols of MY = A over Q */
   setlg(Y, L);
   aux = factorback2(sunitrel, gfloor(Y));
-  x = gdiv(mkpolmod(x,nfpol), RgXQ_norm(lift_intern(aux), relpol));
-  if (typ(x) == t_POLMOD && (typ(x[2]) != t_POL || !degpol(gel(x,2))))
-  {
-    x = gel(x,2); /* rational number */
-    if (typ(x) == t_POL) x = gel(x,2);
-  }
+  x = mkpolmod(x,nfpol);
+  if (!gequal1(aux)) x = gdiv(x, gnorm(aux));
+  x = lift_if_rational(x);
   if (typ(aux) == t_POLMOD && degpol(nfpol) == 1)
-    gel(aux,2) = simplify_shallow( lift_intern(gel(aux,2)) );
+    gel(aux,2) = lift_if_rational(gel(aux,2));
   return gerepilecopy(av, mkvec2(aux, x));
 }
 
