@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2004  The PARI group.
+/* Copyright (C) 2004  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -21,10 +19,21 @@ void hashstr_dbg(hashtable *h);
 /* for qsort */
 typedef int (*QSCOMP)(const void *, const void *);
 
-/* for (ulong*) GENs */
-typedef ulong *uGEN;
-#define ucoeff(a,i,j)  (((uGEN*)(a))[j][i])
-#define ugel(a,i) ((uGEN*)(a))[i]
+#define ucoeff(a,i,j)  (((ulong**)(a))[j][i])
+#define umael(a,i,j)   (((ulong**)(a))[i][j])
+#define uel(a,i)       (((ulong*)(a))[i])
+
+/* to manipulate 'blocs' */
+#define BL_HEAD 4
+#define bl_base(x) (void*)((x) - BL_HEAD)
+#define bl_refc(x) (((GEN)x)[-4])
+#define bl_next(x) (((GEN*)x)[-3])
+#define bl_prev(x) (((GEN*)x)[-2])
+#define bl_num(x)  (((GEN)x)[-1])
+INLINE void
+clone_lock(GEN C) { if (isclone(C)) ++bl_refc(C); }
+INLINE void
+clone_unlock(GEN C) { if (isclone(C)) gunclone(C); }
 
 /* swap */
 #define lswap(x,y) {long _z=x; x=y; y=_z;}
@@ -38,9 +47,14 @@ typedef ulong *uGEN;
 GEN ellheightoo(GEN e, GEN z, long prec);
 void ellprint(GEN e);
 
+/* binary splitting */
+struct abpq { GEN *a, *b, *p, *q; };
+struct abpq_res { GEN P, Q, B, T; };
+void abpq_init(struct abpq *A, long n);
+void abpq_sum(struct abpq_res *r, long n1, long n2, struct abpq *A);
+
 /* generic */
-GEN trans_fix_arg(long *prec, GEN *s0, GEN *sig, pari_sp *av, GEN *res);
-GEN transc(GEN (*f) (GEN, long), GEN x, long prec);
+GEN trans_fix_arg(long *prec, GEN *s0, GEN *sig, GEN *tau, pari_sp *av, GEN *res);
 GEN sort_factor_pol(GEN y, int (*cmp)(GEN,GEN));
 
 /* loops */
@@ -49,6 +63,7 @@ GEN resetloop(GEN a, GEN b);
 GEN setloop(GEN a);
 
 /* parser */
+GEN  iferrpari(GEN a, GEN b, GEN c);
 void forpari(GEN a, GEN b, GEN node);
 void untilpari(GEN a, GEN b);
 void whilepari(GEN a, GEN b);
@@ -56,6 +71,7 @@ GEN  ifpari(GEN g, GEN a, GEN b);
 GEN  andpari(GEN a, GEN b);
 GEN  orpari(GEN a, GEN b);
 void ifpari_void(GEN g, GEN a, GEN b);
+GEN  ifpari_multi(GEN g, GEN a);
 GEN  geval_gp(GEN x, GEN t);
 
 GEN  gadde(GEN *x, GEN y);
@@ -67,7 +83,6 @@ GEN  gmode(GEN *x, GEN y);
 GEN  gmule(GEN *x, GEN y);
 GEN  gshiftle(GEN *x, long n);
 GEN  gshiftre(GEN *x, long n);
-GEN  gstore(GEN *x, GEN y);
 GEN  gsube(GEN *x, GEN y);
 GEN  gsub1e(GEN *x);
 GEN  gshift_right(GEN x, long n);
@@ -75,13 +90,14 @@ GEN  gshift_right(GEN x, long n);
 GEN  derivnum0(GEN a, GEN code, long prec);
 GEN  derivfun0(GEN code, GEN args, long prec);
 GEN  direuler0(GEN a, GEN b, GEN code, GEN c);
-GEN  divsum(GEN num, GEN code);
+void forcomposite(GEN a, GEN b, GEN code);
 void fordiv(GEN a, GEN code);
 void forell0(long a, long b, GEN code);
 void forprime(GEN a, GEN b, GEN code);
 void forstep(GEN a, GEN b, GEN s, GEN code);
 void forsubgroup0(GEN cyc, GEN bound, GEN code);
 void forvec(GEN x, GEN code, long flag);
+void forpart0(GEN k, GEN code , GEN nbound, GEN abound);
 GEN  intcirc0(GEN a, GEN R, GEN code, GEN tab, long prec);
 GEN  intfourcos0(GEN a, GEN b, GEN x, GEN code, GEN tab, long prec);
 GEN  intfourexp0(GEN a, GEN b, GEN x, GEN code, GEN tab, long prec);
@@ -99,16 +115,32 @@ GEN  prodinf0(GEN a, GEN code, long flag, long prec);
 GEN  produit(GEN a, GEN b, GEN code, GEN x);
 GEN  somme(GEN a, GEN b, GEN code, GEN x);
 GEN  sumalt0(GEN a, GEN code,long flag, long prec);
+GEN  sumdivexpr(GEN num, GEN code);
+GEN  sumdivmultexpr(GEN num, GEN code);
 GEN  suminf0(GEN a, GEN code, long prec);
 GEN  sumnum0(GEN a, GEN sig, GEN code, GEN tab, long flag, long prec);
 GEN  sumnumalt0(GEN a, GEN sig, GEN code, GEN tab, long flag, long prec);
 GEN  sumnuminit0(GEN a, GEN tab, long sgn, long prec);
 GEN  sumpos0(GEN a, GEN code, long flag,long prec);
+GEN  vecexpr0(GEN nmax, GEN code, GEN pred);
+GEN  vecexpr1(GEN nmax, GEN code, GEN pred);
 GEN  vecteursmall(GEN nmax, GEN code);
 GEN  vecteur(GEN nmax, GEN n);
 GEN  vvecteur(GEN nmax, GEN n);
 GEN  zbrent0(GEN a, GEN b, GEN code, long prec);
 
+/* mt */
+void mt_sigint(void);
+void mt_err_recover(long er);
+void mt_init_stack(size_t s);
+int  mt_is_thread(void);
+GEN  parapply_worker(GEN d, GEN code);
+GEN  pareval_worker(GEN code);
+void parfor(GEN a, GEN b, GEN code, GEN code2);
+GEN  parfor_worker(GEN i, GEN C);
+void parforprime(GEN a, GEN b, GEN code, GEN code2);
+GEN  parvector_worker(GEN i, GEN C);
+
 /* multiprecision */
 GEN   addrex01(GEN x);
 GEN   adduispec_offset(ulong s, GEN x, long offset, long nx);
@@ -118,7 +150,6 @@ ulong xgcduu(ulong d, ulong d1, int f, ulong* v, ulong* v1, long *s);
 ulong xxgcduu(ulong d, ulong d1, int f, ulong* u, ulong* u1, ulong* v, ulong* v1, long *s);
 GEN   divgunu(GEN x, ulong i);
 GEN   divrunu(GEN x, ulong i);
-GEN   logagmcx(GEN q, long prec);
 GEN   muliispec(GEN x, GEN y, long nx, long ny);
 GEN   red_montgomery(GEN T, GEN N, ulong inv);
 GEN   sqrispec(GEN x, long nx);
@@ -128,7 +159,6 @@ ulong *convi(GEN x, long *l);
 
 int approx_0(GEN x, GEN y);
 GEN bernfrac_using_zeta(long n);
-int OK_bern(long nb, long prec);
 
 /* powers */
 GEN    rpowuu(ulong a, ulong n, long prec);
@@ -139,7 +169,6 @@ double dabs(double s, double t);
 void   dcxlog(double s, double t, double *a, double *b);
 double dnorm(double s, double t);
 double dbllog2(GEN z);
-ulong  usqrtsafe(ulong a);
 
 /* hnf */
 GEN hnfadd(GEN m,GEN p,GEN* ptdep,GEN* ptA,GEN* ptC,GEN extramat,GEN extraC);
@@ -147,14 +176,14 @@ GEN hnfadd_i(GEN m,GEN p,GEN* ptdep,GEN* ptA,GEN* ptC,GEN extramat,GEN extraC);
 GEN hnfspec_i(GEN m,GEN p,GEN* ptdep,GEN* ptA,GEN* ptC,long k0);
 GEN hnfspec(GEN m,GEN p,GEN* ptdep,GEN* ptA,GEN* ptC,long k0);
 GEN mathnfspec(GEN x, GEN *ptperm, GEN *ptdep, GEN *ptB, GEN *ptC);
+GEN ZM_hnfmodall_i(GEN x, GEN dm, long flag);
 
 GEN LLL_check_progress(GEN Bnorm, long n0, GEN m, int final, long *ti_LLL);
 GEN extendedgcd(GEN A);
 
 /* miscellaneous linear algebra */
 GEN  imagecomplspec(GEN x, long *nlze);
-INLINE void rowselect_p(GEN A, GEN B, GEN p, long init);
-
+GEN  ZM_imagecomplspec(GEN x, long *nlze);
 GEN  dim1proj(GEN prh);
 GEN  detcyc(GEN cyc, long *L);
 
@@ -163,11 +192,9 @@ GEN merge_factor_i(GEN f, GEN g);
 /* integer factorization / discrete log */
 GEN   coprime_part(GEN x, GEN f);
 ulong ucoprime_part(ulong x, ulong f);
-ulong is_kth_power(GEN x, ulong p, GEN *pt, byteptr d);
+ulong is_kth_power(GEN x, ulong p, GEN *pt);
 GEN   mpqs(GEN N);
 ulong gcduodd(ulong x, ulong y);
-long  Z_lvalrem_stop(GEN n, ulong p, int *stop);
-long  u_lvalrem_stop(ulong *n, ulong p, int *stop);
 
 /* Polynomials */
 /* a) Arithmetic/conversions */
@@ -185,12 +212,14 @@ GEN  RgX_recipspec_shallow(GEN x, long l, long n);
 
 /* b) Modular */
 GEN  bezout_lift_fact(GEN T, GEN Tmod, GEN p, long e);
-GEN  Kronecker_to_FpXQX(GEN z, GEN pol, GEN p);
-GEN  FpX_quad_root(GEN x, GEN p, int unknown);
+long F2x_split_Berlekamp(GEN *t);
+long Flx_split_Berlekamp(GEN *t, ulong p);
 long FpX_split_Berlekamp(GEN *t, GEN pp);
+long FqX_split_Berlekamp(GEN *t, GEN T, GEN p);
+GEN  FpX_quad_root(GEN x, GEN p, int unknown);
 GEN  FqX_split_all(GEN z, GEN T, GEN p);
-long FqX_split_by_degree(GEN *pz, GEN u, GEN q, GEN T, GEN p);
-long FqX_split_deg1(GEN *pz, GEN u, GEN q, GEN T, GEN p);
+long FqX_split_by_degree(GEN *pz, GEN u, GEN T, GEN p);
+long FqX_split_deg1(GEN *pz, GEN u, GEN T, GEN p);
 GEN  FqX_split_roots(GEN z, GEN T, GEN p, GEN pol);
 GEN  polsym_gen(GEN P, GEN y0, long n, GEN T, GEN N);
 GEN  ZXQ_charpoly_sqf(GEN A, GEN B, long *lambda, long v);
@@ -198,8 +227,7 @@ GEN  ZX_disc_all(GEN,ulong);
 GEN  ZX_resultant_all(GEN A, GEN B, GEN dB, ulong bound);
 GEN  ZX_ZXY_resultant_all(GEN A, GEN B, long *lambda, GEN *LPRS);
 GEN  RgXQ_minpoly_naive(GEN y, GEN P);
-GEN lift_intern0(GEN x,long v);
-#define lift_intern(x) (lift_intern0((x),-1))
+GEN lift_intern(GEN x);
 
 /* c) factorization */
 double cauchy_bound(GEN p);
@@ -215,19 +243,31 @@ GEN special_pivot(GEN x);
 GEN vandermondeinversemod(GEN L, GEN T, GEN den, GEN mod);
 GEN ZX_monic_factorpadic(GEN f, GEN p, long prec);
 
-/* Finite field */
+/* Finite fields */
 
 enum { t_FF_FpXQ = 0, t_FF_Flxq = 1, t_FF_F2xq = 2 };
+GEN FF_ellinit(GEN E, GEN fg);
+GEN FF_elldata(GEN E, GEN fg);
+
+/* Elliptic curves */
+/* common to Q and Rg */
+enum { R_PERIODS = 1, R_ETA, R_ROOTS, R_AB };
+
+enum { Qp_ROOT = 1, Qp_TATE };
+enum { Q_GROUPGEN = 5, Q_GLOBALRED, Q_ROOTNO, Q_MINIMALMODEL };
+
+/* common to Fp and Fq */
+enum { FF_CARD = 1, FF_GROUP, FF_GROUPGEN, FF_O };
 
 /* for Buchall_param */
 enum { fupb_NONE, fupb_RELAT, fupb_LARGE, fupb_PRECI };
 
 /* Allocation / gerepile */
+void   setdebugvar(long n);
 void   debug_stack(void);
 void   fill_stack(void);
 void   init_dalloc(void);
 double *dalloc(size_t n);
-INLINE void   gerepilecoeffs2(pari_sp av, GEN x, int n, GEN y, int o);
 void   minim_alloc(long n, double ***q, GEN *x, double **y,  double **z, double **v);
 int    pop_entree_block(entree *ep, long loc);
 int    pop_val_if_newer(entree *ep, long loc);
@@ -243,6 +283,8 @@ INLINE long
 is_keyword_char(char c) { return (isalnum((int)c) || c=='_'); }
 
 /* Interfaces (GP, etc.) */
+hashtable *hash_from_link(GEN e, GEN names, int use_stack);
+void gen_relink(GEN x, hashtable *table);
 entree* is_entry_intern(const char *s, entree **table, long *hash);
 entree* do_alias(entree *ep);
 char* get_sep(const char *t);
@@ -257,8 +299,10 @@ void* get_stack(double fraction, long min);
 void  init_graph(void);
 void  free_graph(void);
 void  initout(int initerr);
-void  init80col(void);
-int   pari_kernel_init(void);
+void  resetout(int initerr);
+void  init_linewrap(long w);
+void  pari_kernel_init(void);
+void  pari_kernel_close(void);
 void  print_functions_hash(const char *s);
 void  print_all_user_fun(int member);
 GEN   readbin(const char *name, FILE *f, int *vector);
@@ -274,7 +318,11 @@ extern ulong precreal;
 
 /* history */
 typedef struct {
-  GEN *res;    /* array of previous results, FIFO */
+  GEN z; /* result */
+  time_t t; /* time to obtain result */
+} gp_hist_cell;
+typedef struct {
+  gp_hist_cell *v; /* array of previous results, FIFO */
   size_t size; /* # res */
   ulong total; /* # of results computed since big bang */
 } gp_hist;
@@ -302,6 +350,7 @@ typedef struct {
 
 void lim_lines_output(char *s, long n, long max);
 void gen_output(GEN x, pariout_t *T);
+void fputGEN_pariout(GEN x, pariout_t *T, FILE *out);
 
 void parsestate_reset(void);
 void parsestate_save(struct pari_parsestate *state);
@@ -316,24 +365,37 @@ void evalstate_reset(void);
 void evalstate_restore(struct pari_evalstate *state);
 void evalstate_save(struct pari_evalstate *state);
 
+void mtstate_save(long *pending);
+void mtstate_reset(void);
+void mtstate_restore(long *pending);
+
 void debug_context(void);
 
 /* GP_DATA */
 typedef struct {
   gp_hist *hist;
   gp_pp *pp;
-  gp_path *path;
+  gp_path *path, *sopath;
   pariout_t *fmt;
-  ulong lim_lines, flags;
+  ulong lim_lines, flags, linewrap;
   int echo, breakloop, recover, use_readline; /* GP-specific */
-  int secure, simplify, strictmatch, chrono; /* libpari ? */
+  int secure, simplify, strictmatch, strictargs, chrono; /* libpari ? */
   pari_timer *T;
-  long last_time; /* duration of last interactive gp command */
+  ulong primelimit; /* deprecated */
+  ulong threadsize;
 } gp_data;
 extern gp_data *GP_DATA;
   /* GP_DATA->flags */
 enum { gpd_QUIET=1, gpd_TEST=2, gpd_EMACS=256, gpd_TEXMACS=512};
 
+typedef struct {
+  const char *s;
+  size_t ls;
+  char **dir;
+} forpath_t;
+void forpath_init(forpath_t *T, gp_path *path, const char *s);
+char *forpath_next(forpath_t *T);
+
 /* GP output && output format */
 void gpwritebin(const char *s, GEN x);
 extern char *current_logfile;
@@ -386,6 +448,7 @@ GEN  gp_history(gp_hist *H, long p, char *old, char *entry);
 void delete_dirs(gp_path *p);
 void gp_expand_path(gp_path *p);
 const char *pari_default_path(void);
+int path_is_absolute(char *s);
 
 typedef struct input_method {
 /* mandatory */
@@ -404,61 +467,38 @@ char *file_getline(Buffer *b, char **s0, input_method *IM);
 
 /* By files */
 
-/* Flx.c */
-
-GEN     FlxX_recipspec(GEN x, long l, long n, long vs);
-GEN     FlxX_sub(GEN x, GEN y, ulong p);
-GEN     FlxX_subspec(GEN x, GEN y, ulong p, long lx, long ly);
-GEN     FlxX_swap(GEN x, long n, long ws);
-GEN     zxX_to_Kronecker(GEN P, GEN Q);
-GEN     zxX_to_Kronecker_spec(GEN P, long lp, GEN Q);
-GEN     Flx_addshift(GEN x, GEN y, ulong p, long d);
-GEN     Flx_addspec(GEN x, GEN y, ulong p, long lx, long ly);
-GEN     Flx_even_odd_comb(GEN P, ulong u, ulong v, ulong p);
-GEN     Flx_mulspec(GEN a, GEN b, ulong p, long na, long nb);
-GEN     Flx_negspec(GEN x, ulong p, long l);
-GEN     Flx_recipspec(GEN x, long l, long n);
-GEN     Flx_sqrspec(GEN a, ulong p, long na);
-GEN     Flx_subspec(GEN x, GEN y, ulong p, long lx, long ly);
-GEN     Kronecker_to_FlxqX(GEN z, GEN T, ulong p);
-GEN     FlxqX_invMontgomery(GEN T, GEN Q, ulong p);
-GEN     FlxqX_mulspec(GEN x, GEN y, GEN T, ulong p, long lx, long ly);
-GEN     FlxqX_rem_Montgomery(GEN x, GEN mg, GEN T, GEN Q, ulong p);
+/* FpE.c */
+long Fl_elltrace_CM(int CM, ulong a4, ulong a6, ulong p);
 
 /* Qfb.c */
 
 GEN     redimagsl2(GEN q, GEN *U);
-GEN     redrealsl2(GEN V);
-GEN     redrealsl2step(GEN A);
+GEN     redrealsl2(GEN V, GEN d, GEN rd);
+GEN     redrealsl2step(GEN A, GEN d, GEN rd);
 GEN     redtausl2(GEN t, GEN *U);
 
 /* alglin1.c */
 typedef long (*pivot_fun)(GEN,GEN,long,GEN);
+GEN ZM_pivots(GEN x0, long *rr);
 GEN RgM_pivots(GEN x0, GEN data, long *rr, pivot_fun pivot);
-INLINE void vecselect_p(GEN A, GEN B, GEN p, long init, long lB);
 
 /* arith1.c */
 
-GEN     bestappr_mod(GEN x, GEN A, GEN B);
 int     is_gener_Fp(GEN x, GEN p, GEN p_1, GEN L);
 int     is_gener_Fl(ulong x, ulong p, ulong p_1, GEN L);
 
 /* arith2.c */
 
-byteptr initprimes0(ulong maxnum, long *lenp, ulong *lastp);
+int     divisors_init(GEN n, GEN *pP, GEN *pE);
 long    set_optimize(long what, GEN g);
 
-/* base1.c */
-
-GEN     nffromhnfbasis(GEN nf, GEN x);
-GEN     nftohnfbasis(GEN nf, GEN x);
-
 /* base2.c */
 
 GEN     gen_if_principal(GEN bnf, GEN x);
 int     nfissquarefree(GEN nf, GEN x);
-GEN     nfreducemodpr_i(GEN x, GEN prh);
 GEN     polsymmodp(GEN g, GEN p);
+GEN     nfbasis_gp(GEN T, GEN P, GEN junk);
+GEN     nfdisc_gp(GEN T, GEN P, GEN junk);
 
 /* base3.c */
 
@@ -477,13 +517,6 @@ long    val_norm(GEN x, GEN p, long *vz);
 
 GEN     check_and_build_nfabs(GEN rnf);
 GEN     check_and_build_norms(GEN rnf);
-GEN     checkrnfeq(GEN x);
-
-/* bibli1.c */
-
-GEN     plindep(GEN x);
-GEN     pslq(GEN x);
-GEN     pslqL2(GEN x);
 
 /* buch1.c */
 
@@ -492,8 +525,12 @@ GEN     qfbforms(GEN D);
 
 /* buch2.c */
 
-typedef struct GRHcheck_t { double cD, cN; long checkok; } GRHcheck_t;
+typedef struct GRHprime_t { ulong p; double logp; GEN dec; } GRHprime_t;
+typedef struct GRHcheck_t { double cD, cN; GRHprime_t *primes; long clone, nprimes, maxprimes; ulong limp; forprime_t P; } GRHcheck_t;
+void    free_GRHcheck(GRHcheck_t *S);
 void    init_GRHcheck(GRHcheck_t *S, long N, long R1, double LOGD);
+void    GRH_ensure(GRHcheck_t *S, long nb);
+ulong   GRH_last_prime(GRHcheck_t *S);
 int     GRHok(GRHcheck_t *S, double L, double SA, double SB);
 GEN     check_and_build_matal(GEN bnf);
 GEN     extract_full_lattice(GEN x);
@@ -507,10 +544,6 @@ GEN     minkowski_bound(GEN D, long N, long r2, long prec);
 int     subgroup_conductor_ok(GEN H, GEN L);
 GEN     subgrouplist_cond_sub(GEN bnr, GEN C, GEN bound);
 
-/* elliptic.c */
-
-GEN     weipell0(GEN e, long prec, long PREC);
-
 /* ellsea.c */
 
 void    pari_close_seadata(void);
@@ -523,19 +556,26 @@ char *  env_ok(const char *s);
 void    filestate_restore(pariFILE *F);
 void    killallfiles(void);
 pariFILE* newfile(FILE *f, const char *name, int type);
+void    pari_init_homedir(void);
+void    pari_close_homedir(void);
 void    pari_init_files(void);
 void    pari_close_files(void);
 int     popinfile(void);
-GEN     readobj(FILE *f, int *ptc);
 pariFILE* try_pipe(const char *cmd, int flag);
-void    writeGEN(GEN x, FILE *f);
-void    writenamedGEN(GEN x, const char *s, FILE *f);
+
+/* Flxq_log.c */
+
+GEN Flxq_log_index(GEN a0, GEN b0, GEN m, GEN T0, ulong p);
+
+/* FlxqE.c */
+
+GEN     ZpXQ_norm_pcyc(GEN x, GEN T, GEN q, GEN p);
+long    zx_is_pcyc(GEN T);
 
 /* galconj.c */
 
 GEN     galoiscosets(GEN O, GEN perm);
 long    intheadlong(GEN x, GEN mod);
-long    isomborne(GEN P, GEN den, GEN p);
 GEN     listznstarelts(long m, long p);
 GEN     matheadlong(GEN W, GEN mod);
 GEN     matrixnorm(GEN M, long prec);
@@ -573,11 +613,13 @@ GEN     squfof(GEN n);
 long    BPSW_psp_nosmalldiv(GEN N);
 int     Fl_MR_Jaeschke(ulong n, long k);
 int     MR_Jaeschke(GEN n, long k);
-int     uisprime_nosmalldiv(ulong n);
+long    isanypower_nosmalldiv(GEN N, GEN *px);
+void    prime_table_next_p(ulong a, byteptr *pd, ulong *pp, ulong *pn);
+int     uisprime_101(ulong n);
+int     uisprime_661(ulong n);
 
 /* init.c */
 
-void    err_recover(long numerr);
 void    pari_init_defaults(void);
 void    pari_init_stack(size_t size, size_t old);
 
@@ -599,14 +641,14 @@ GEN     perm_to_GAP(GEN p);
 GEN     F2x_Berlekamp_ker(GEN u);
 GEN     Flx_Berlekamp_ker(GEN u, ulong p);
 GEN     FpX_Berlekamp_ker(GEN u, GEN p);
+GEN     FlxqX_Berlekamp_ker(GEN u, GEN T, ulong p);
+GEN     FpXQX_Berlekamp_ker(GEN u, GEN T, GEN p);
+GEN     F2x_factcantor(GEN f, long flag);
+GEN     Flx_factcantor(GEN f, ulong p, long flag);
 GEN     FpX_factcantor(GEN f, GEN pp, long flag);
-GEN     FqX_Berlekamp_ker(GEN u, GEN T, GEN q, GEN p);
 GEN     FqX_rand(long d1, long v, GEN T, GEN p);
-long    FqX_split_Berlekamp(GEN *t, GEN q, GEN T, GEN p);
-GEN     Zp_appr(GEN f, GEN a);
 int     cmp_padic(GEN x, GEN y);
 GEN     factcantor0(GEN f, GEN pp, long flag);
-GEN     trivfact(void);
 
 /* polarit2.c */
 
@@ -614,9 +656,7 @@ GEN     sylvestermatrix_i(GEN x, GEN y);
 
 /* QX_factor */
 
-GEN     DDF_roots(GEN pol, GEN polp, GEN p);
 void    factor_quad(GEN x, GEN res, long *ptcnt);
-long    logint(GEN B, GEN y, GEN *ptq);
 
 /* FpX.c */
 
@@ -631,7 +671,7 @@ GEN     FpX_direct_compositum(GEN A, GEN B, GEN p);
 ulong   ZX_ZXY_ResBound(GEN A, GEN B, GEN dB);
 GEN     ffinit_Artin_Shreier(GEN ip, long l);
 GEN     ffinit_rand(GEN p, long n);
-byteptr init_modular(ulong *p);
+void    init_modular(forprime_t *S);
 GEN     polint_triv(GEN xa, GEN ya);
 
 /* random.c */
@@ -660,10 +700,12 @@ GEN     znstar_small(GEN zn);
 
 /* trans1.c */
 
+GEN     logagmcx(GEN q, long prec);
 void    pari_init_floats(void);
 void    pari_close_floats(void);
 GEN     rootsof1complex(GEN n, long prec);
 GEN     rootsof1padic(GEN n, GEN y);
+GEN     zellagmcx(GEN a0, GEN b0, GEN r, GEN t, long prec);
 
 /* trans2.c */
 
@@ -677,6 +719,7 @@ GEN     czeta(GEN s0, long prec);
 GEN     double_eta_quotient(GEN a, GEN w, GEN D, long p, long q, GEN pq, GEN sqrtD);
 GEN     inv_szeta_euler(long n, double lba, long prec);
 GEN     polylogd0(long m, GEN x, long flag, long prec);
+GEN     trueE2(GEN tau, long prec);
 GEN     twistpartialzeta(GEN q, long f, long c, GEN va, GEN cff);
 
 ENDEXTERN
