@@ -1,7 +1,5 @@
 #line 2 "../src/kernel/gmp/mp.c"
-/* $Id$
-
-Copyright (C) 2002-2003  The PARI group.
+/* Copyright (C) 2002-2003  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -45,19 +43,36 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /*We need PARI invmod renamed to invmod_pari*/
 #define INVMOD_PARI
 
-static void *gmp_realloc(void *ptr, size_t old_size, size_t new_size) {
+static void *pari_gmp_realloc(void *ptr, size_t old_size, size_t new_size) {
   (void)old_size; return (void *) pari_realloc(ptr,new_size);
 }
 
-static void gmp_free(void *ptr, size_t old_size){
+static void pari_gmp_free(void *ptr, size_t old_size){
   (void)old_size; pari_free(ptr);
 }
 
-int pari_kernel_init(void)
+static void *(*old_gmp_malloc)(size_t new_size);
+static void *(*old_gmp_realloc)(void *ptr, size_t old_size, size_t new_size);
+static void (*old_gmp_free)(void *ptr, size_t old_size);
+
+void
+pari_kernel_init(void)
 {
-  /* Use pari_malloc instead of malloc */
-  mp_set_memory_functions((void *(*)(size_t)) pari_malloc, gmp_realloc, gmp_free);
-  return 0;
+  mp_get_memory_functions (&old_gmp_malloc, &old_gmp_realloc, &old_gmp_free);
+  mp_set_memory_functions((void *(*)(size_t)) pari_malloc, pari_gmp_realloc, pari_gmp_free);
+}
+
+void
+pari_kernel_close(void)
+{
+  void *(*new_gmp_malloc)(size_t new_size);
+  void *(*new_gmp_realloc)(void *ptr, size_t old_size, size_t new_size);
+  void (*new_gmp_free)(void *ptr, size_t old_size);
+  mp_get_memory_functions (&new_gmp_malloc, &new_gmp_realloc, &new_gmp_free);
+  if (new_gmp_malloc==pari_malloc) new_gmp_malloc = old_gmp_malloc;
+  if (new_gmp_realloc==pari_gmp_realloc) new_gmp_realloc = old_gmp_realloc;
+  if (new_gmp_free==pari_gmp_free) new_gmp_free = old_gmp_free;
+  mp_set_memory_functions(new_gmp_malloc, new_gmp_realloc, new_gmp_free);
 }
 
 #define LIMBS(x)  ((mp_limb_t *)((x)+2))
@@ -153,21 +168,23 @@ incpos(GEN a)
 static GEN
 incneg(GEN a)
 {
-  long i, l = lgefint(a);
+  long i, l = lgefint(a)-1;
   if (a[2]--)
   {
-    if (l == 3 && !a[2])
+    if (!a[l]) /* implies l = 2 */
     {
       a[0] = evaltyp(t_INT) | _evallg(2);
       a[1] = evalsigne(0) | evallgefint(2);
     }
     return a;
   }
-  for (i=3; i<l; i++)
+  for (i=3; i<=l; i++)
     if (a[i]--) break;
-  l -= i - 2;
-  a[0] = evaltyp(t_INT) | _evallg(l);
-  a[1] = evalsigne(-1) | evallgefint(l);
+  if (!a[l])
+  {
+    a[0] = evaltyp(t_INT) | _evallg(l);
+    a[1] = evalsigne(-1) | evallgefint(l);
+  }
   return a;
 }
 
@@ -269,8 +286,8 @@ roundr_up_ip(GEN x, long l)
   long i = l;
   for(;;)
   {
-    if (++x[--i]) break;
-    if (i == 2) { x[2] = HIGHBIT; setexpo(x, expo(x)+1); break; }
+    if (++((ulong*)x)[--i]) break;
+    if (i == 2) { x[2] = HIGHBIT; shiftr_inplace(x, 1); break; }
   }
 }
 
@@ -315,60 +332,58 @@ affir(GEN x, GEN y)
   }
 }
 
-GEN
-shifti(GEN x, long n)
+INLINE GEN
+shiftispec(GEN x, long nx, long n)
 {
-  const long s=signe(x);
-  long lz,lx,m;
-  GEN z;
+  long ny,m;
+  GEN yd, y;
+
+  if (!n) return icopyspec(x, nx);
 
-  if (!s) return gen_0;
-  if (!n) return icopy(x);
-  lx = lgefint(x);
   if (n > 0)
   {
     long d = dvmdsBIL(n, &m);
     long i;
 
-    lz = lx + d + (m!=0);
-    z = cgeti(lz);
-    for (i=0; i<d; i++) LIMBS(z)[i] = 0;
+    ny = nx + d + (m!=0);
+    y = cgeti(ny + 2); yd = y + 2;
+    for (i=0; i<d; i++) yd[i] = 0;
 
-    if (!m) xmpn_copy(LIMBS(z)+d, LIMBS(x), NLIMBS(x));
+    if (!m) xmpn_copy((mp_limb_t *) (yd + d), (mp_limb_t *) x, nx);
     else
     {
-      ulong carry = mpn_lshift(LIMBS(z)+d, LIMBS(x), NLIMBS(x), m);
-      if (carry) z[lz - 1] = carry;
-      else lz--;
+      ulong carryd = mpn_lshift((mp_limb_t *) (yd + d), (mp_limb_t *) x, nx, m);
+      if (carryd) yd[ny - 1] = carryd;
+      else ny--;
     }
   }
   else
   {
     long d = dvmdsBIL(-n, &m);
 
-    lz = lx - d;
-    if (lz<3) return gen_0;
-    z = cgeti(lz);
+    ny = nx - d;
+    if (ny < 1) return gen_0;
+    y = cgeti(ny + 2); yd = y + 2;
 
-    if (!m) xmpn_copy(LIMBS(z), LIMBS(x) + d, NLIMBS(x) - d);
+    if (!m) xmpn_copy((mp_limb_t *) yd, (mp_limb_t *) (x + d), nx - d);
     else
     {
-      mpn_rshift(LIMBS(z), LIMBS(x) + d, NLIMBS(x) - d, m);
-      if (z[lz - 1] == 0)
+      mpn_rshift((mp_limb_t *) yd, (mp_limb_t *) (x + d), nx - d, m);
+      if (yd[ny - 1] == 0)
       {
-        if (lz == 3) { avma = (pari_sp)(z+3); return gen_0; }
-        lz--;
+        if (ny == 1) { avma = (pari_sp)(yd + 1); return gen_0; }
+        ny--;
       }
     }
   }
-  z[1] = evalsigne(s)|evallgefint(lz);
-  return z;
+  y[1] = evalsigne(1)|evallgefint(ny + 2);
+  return y;
 }
 
 GEN
-trunc2nr_lg(GEN x, long lx, long n)
+mantissa2nr(GEN x, long n)
 {
-  long ly, i, m, s = signe(x);
+  long ly, i, m, s = signe(x), lx = lg(x);
   GEN y;
   if (!s) return gen_0;
   if (!n)
@@ -423,7 +438,7 @@ truncr(GEN x)
   GEN y;
   if ((s=signe(x)) == 0 || (e=expo(x)) < 0) return gen_0;
   d = nbits2prec(e+1); m = remsBIL(e);
-  if (d > lg(x)) pari_err(precer, "truncr (precision loss in truncation)");
+  if (d > lg(x)) pari_err_PREC( "truncr (precision loss in truncation)");
 
   y=cgeti(d); y[1] = evalsigne(s) | evallgefint(d);
   if (++m == BITS_IN_LONG)
@@ -447,7 +462,7 @@ floorr(GEN x)
   if (signe(x) >= 0) return truncr(x);
   if ((e=expo(x)) < 0) return gen_m1;
   d = nbits2prec(e+1); m = remsBIL(e);
-  lx=lg(x); if (d>lx) pari_err(precer, "floorr (precision loss in truncation)");
+  lx=lg(x); if (d>lx) pari_err_PREC( "floorr (precision loss in truncation)");
   y = cgeti(d+1);
   if (++m == BITS_IN_LONG)
   {
@@ -474,15 +489,18 @@ END:
 }
 
 INLINE int
-absi_cmp_lg(GEN x, GEN y, long l)
+cmpiispec(GEN x, GEN y, long lx, long ly)
 {
-  return mpn_cmp(LIMBS(x),LIMBS(y),l-2);
+  if (lx < ly) return -1;
+  if (lx > ly) return  1;
+  return mpn_cmp((mp_limb_t*)x,(mp_limb_t*)y, lx);
 }
 
 INLINE int
-absi_equal_lg(GEN x, GEN y, long l)
+equaliispec(GEN x, GEN y, long lx, long ly)
 {
-  return !mpn_cmp(LIMBS(x),LIMBS(y),l-2);
+  if (lx != ly) return 0;
+  return !mpn_cmp((mp_limb_t*)x,(mp_limb_t*)y, lx);
 }
 
 /***********************************************************************/
@@ -525,30 +543,6 @@ addumului(ulong a, ulong b, GEN y)
   avma=(pari_sp)z; return z;
 }
 
-GEN muliispec(GEN x, GEN y, long nx, long ny);
-
-/* We must have nx>=ny. This lets garbage on the stack.
-   This handle squares correctly (mpn_mul is optimized
-   for squares).
-*/
-
-INLINE GEN
-muliispec_mirror(GEN x, GEN y, long nx, long ny)
-{
-  GEN cx=new_chunk(nx),cy;
-  GEN z;
-  xmpn_mirrorcopy((mp_limb_t *)cx,(mp_limb_t *)x,nx);
-  if (x==y) cy=cx; /*If nx<ny cy will be too short*/
-  else
-  {
-    cy=new_chunk(ny);
-    xmpn_mirrorcopy((mp_limb_t *)cy,(mp_limb_t *)y,ny);
-  }
-  z=muliispec(cx, cy, nx, ny);
-  xmpn_mirror(LIMBS(z), NLIMBS(z));
-  return z;
-}
-
 /***********************************************************************/
 /**                                                                   **/
 /**                          DIVISION                                 **/
@@ -560,7 +554,7 @@ umodiu(GEN y, ulong x)
 {
   long sy=signe(y);
   ulong hi;
-  if (!x) pari_err(gdiver);
+  if (!x) pari_err_INV("umodiu",gen_0);
   if (!sy) return 0;
   hi = mpn_mod_1(LIMBS(y),NLIMBS(y),x);
   if (!hi) return 0;
@@ -574,7 +568,7 @@ diviu_rem(GEN y, ulong x, ulong *rem)
   long ly;
   GEN z;
 
-  if (!x) pari_err(gdiver);
+  if (!x) pari_err_INV("diviu_rem",gen_0);
   if (!signe(y)) { *rem = 0; return gen_0; }
 
   ly = lgefint(y);
@@ -593,7 +587,7 @@ divis_rem(GEN y, long x, long *rem)
   long sy=signe(y),ly,s;
   GEN z;
 
-  if (!x) pari_err(gdiver);
+  if (!x) pari_err_INV("divis_rem",gen_0);
   if (!sy) { *rem = 0; return gen_0; }
   if (x<0) { s = -sy; x = -x; } else s = sy;
 
@@ -614,7 +608,7 @@ divis(GEN y, long x)
   long sy=signe(y),ly,s;
   GEN z;
 
-  if (!x) pari_err(gdiver);
+  if (!x) pari_err_INV("divis",gen_0);
   if (!sy) return gen_0;
   if (x<0) { s = -sy; x = -x; } else s=sy;
 
@@ -709,7 +703,7 @@ divri(GEN x, GEN y)
 {
   long  s = signe(y);
 
-  if (!s) pari_err(gdiver);
+  if (!s) pari_err_INV("divri",gen_0);
   if (!signe(x)) return real_0_bit(expo(x) - expi(y));
   if (!is_bigint(y)) {
     GEN z = divru(x, y[2]);
@@ -726,7 +720,7 @@ divrr(GEN x, GEN y)
   ulong y0,y1;
   GEN r, r1;
 
-  if (!sy) pari_err(gdiver);
+  if (!sy) pari_err_INV("divrr",y);
   e = expo(x) - expo(y);
   if (!sx) return real_0_bit(e);
   if (sy<0) sx = -sx;
@@ -771,7 +765,7 @@ divrr(GEN x, GEN y)
         GEN y1 = y+1;
         j = lr-i; r1[j] = subll(r1[j],y1[j]);
         for (j--; j>0; j--) r1[j] = subllx(r1[j],y1[j]);
-        j=i; do r[--j]++; while (j && !r[j]);
+        j=i; do ((ulong*)r)[--j]++; while (j && !r[j]);
       }
       hiremainder = r1[1]; overflow = 0;
       qp = divll(r1[2],y0); k = hiremainder;
@@ -809,7 +803,7 @@ divrr(GEN x, GEN y)
         r1[1] -= hiremainder;
         while (r1[1])
         {
-          qp++; if (!qp) { j=i; do r[--j]++; while (j && !r[j]); }
+          qp++; if (!qp) { j=i; do ((ulong*)r)[--j]++; while (j && !r[j]); }
           j = lr-i-(lr-i>=ly); r1[j] = subll(r1[j],y[j]);
           for (j--; j>1; j--) r1[j] = subllx(r1[j],y[j]);
           r1[1] -= overflow;
@@ -851,7 +845,11 @@ dvmdii(GEN x, GEN y, GEN *z)
   pari_sp av;
   GEN r,q;
 
-  if (!sy) { if (z == ONLY_REM && !sx) return gen_0; pari_err(gdiver); }
+  if (!sy)
+  {
+    if (z == ONLY_REM && !sx) return gen_0;
+    pari_err_INV("dvmdii",y);
+  }
   if (!sx)
   {
     if (!z || z == ONLY_REM) return gen_0;
@@ -953,7 +951,7 @@ red_montgomery(GEN T, GEN N, ulong inv)
   d = NLIMBS(T); /* <= 2*k */
   if (d == 0) return gen_0;
 #ifdef DEBUG
-  if (d > 2*k) pari_err(bugparier,"red_montgomery");
+  if (d > 2*k) pari_err_BUG("red_montgomery");
 #endif
   if (k == 1)
   { /* as below, special cased for efficiency */
@@ -1028,7 +1026,7 @@ red_montgomery(GEN T, GEN N, ulong inv)
   GEN res = remii(mulii(T, Fp_inv(R, N)), N);
   if (k > lgefint(N)
     || !equalii(remii(Td,N),res)
-    || cmpii(Td, addii(shifti(T, -s), N)) >= 0) pari_err(bugparier,"red_montgomery");
+    || cmpii(Td, addii(shifti(T, -s), N)) >= 0) pari_err_BUG("red_montgomery");
 }
 #endif
   avma = (pari_sp)Td; return Td;
@@ -1074,7 +1072,7 @@ diviuexact(GEN x, ulong y)
 GEN
 diviiexact(GEN x, GEN y)
 {
-  if (!signe(y)) pari_err(gdiver);
+  if (!signe(y)) pari_err_INV("diviiexact",y);
   if (lgefint(y) == 3)
   {
     GEN z = diviuexact(x, y[2]);
@@ -1115,6 +1113,22 @@ diviiexact(GEN x, GEN y)
 }
 #endif
 
+/* assume yz != and yz | x */
+GEN
+diviuuexact(GEN x, ulong y, ulong z)
+{
+  long tmp[4];
+  ulong t;
+  LOCAL_HIREMAINDER;
+  t = mulll(y, z);
+  if (!hiremainder) return diviuexact(x, t);
+  tmp[0] = evaltyp(t_INT)|_evallg(4);
+  tmp[1] = evalsigne(1)|evallgefint(4);
+  tmp[2] = t;
+  tmp[3] = hiremainder;
+  return diviiexact(x, tmp);
+}
+
 
 /********************************************************************/
 /**                                                                **/
@@ -1143,6 +1157,26 @@ muliispec(GEN x, GEN y, long nx, long ny)
   zd[1] = evalsigne(1) | evallgefint(lz);
   return zd;
 }
+GEN
+muluui(ulong x, ulong y, GEN z)
+{
+  long t, s = signe(z);
+  GEN r;
+  LOCAL_HIREMAINDER;
+
+  if (!x || !y || !signe(z)) return gen_0;
+  t = mulll(x,y);
+  if (!hiremainder)
+    r = muluispec(t, z+2, lgefint(z)-2);
+  else
+  {
+    long tmp[2];
+    tmp[1] = hiremainder;
+    tmp[0] = t;
+    r = muliispec(z+2,tmp, lgefint(z)-2, 2);
+  }
+  setsigne(r,s); return r;
+}
 
 GEN
 sqrispec(GEN x, long nx)
@@ -1155,13 +1189,53 @@ sqrispec(GEN x, long nx)
 
   lz = (nx<<1)+2;
   zd = cgeti(lz);
+#ifdef mpn_sqr
+  mpn_sqr(LIMBS(zd), (mp_limb_t *)x, nx);
+#else
   mpn_mul_n(LIMBS(zd), (mp_limb_t *)x, (mp_limb_t *)x, nx);
+#endif
   if (zd[lz-1]==0) lz--;
 
   zd[1] = evalsigne(1) | evallgefint(lz);
   return zd;
 }
 
+INLINE GEN
+sqrispec_mirror(GEN x, long nx)
+{
+  GEN cx=new_chunk(nx);
+  GEN z;
+  xmpn_mirrorcopy((mp_limb_t *)cx,(mp_limb_t *)x,nx);
+  z=sqrispec(cx, nx);
+  xmpn_mirror(LIMBS(z), NLIMBS(z));
+  return z;
+}
+
+/* leaves garbage on the stack. */
+INLINE GEN
+muliispec_mirror(GEN x, GEN y, long nx, long ny)
+{
+  GEN cx, cy, z;
+  long s = 0;
+  while (nx && x[nx-1]==0) { nx--; s++; }
+  while (ny && y[ny-1]==0) { ny--; s++; }
+  cx=new_chunk(nx); cy=new_chunk(ny);
+  xmpn_mirrorcopy((mp_limb_t *)cx,(mp_limb_t *)x,nx);
+  xmpn_mirrorcopy((mp_limb_t *)cy,(mp_limb_t *)y,ny);
+  z =  nx>=ny ? muliispec(cx, cy, nx, ny): muliispec(cy, cx, ny, nx);
+  if (s)
+  {
+    long i, lz = lgefint(z) + s;
+    (void)new_chunk(s);
+    z -= s;
+    for (i=0; i<s; i++) z[2+i]=0;
+    z[1] = evalsigne(1) | evallgefint(lz);
+    z[0] = evaltyp(t_INT) | evallg(lz);
+  }
+  xmpn_mirror(LIMBS(z), NLIMBS(z));
+  return z;
+}
+
 /* x % (2^n), assuming n >= 0 */
 GEN
 remi2n(GEN x, long n)
@@ -1251,7 +1325,7 @@ sqrtr_abs(GEN a)
   else
   {
     ulong u;
-    b = (mp_limb_t *) trunc2nr_lg(a,l+2,-1);
+    b = (mp_limb_t *) mantissa2nr(a,-1);
     b[1] = ((ulong)a[l+1])<<(BITS_IN_LONG-1);
     b = (mp_limb_t *) new_chunk(l);
     xmpn_zero(b,l+1); /* overwrites the former b[0] */
