@@ -1,7 +1,5 @@
 #line 2 "../src/kernel/none/gcdll.c"
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -12,7 +10,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /***********************************************************************/
 /**                                                                   **/
@@ -1067,29 +1065,20 @@ lgcdii(ulong* d, ulong* d1,
 
 /* 1 / Mod(x,p). Assume x < p */
 ulong
-Fl_inv(ulong x, ulong p)
+Fl_invsafe(ulong x, ulong p)
 {
   long s;
   ulong xv, xv1, g = xgcduu(p, x, 1, &xv, &xv1, &s);
-  if (g != 1UL) pari_err(invmoder, mkintmod(utoi(x), utoi(p)));
+  if (g != 1UL) return 0UL;
   xv = xv1 % p; if (s < 0) xv = p - xv;
   return xv;
 }
 
-#if 0
-/* assume m > 0 */
-long
-Fl_inv_signed(long a, long m)
+/* 1 / Mod(x,p). Assume x < p */
+ulong
+Fl_inv(ulong x, ulong p)
 {
-  if (a >= 0)
-  {
-    if (a > m) a %= m;
-  }
-  else
-  {
-    if (-a > m) a %= m;
-    a += m;
-  }
-  return (long)Fl_inv((ulong)a, (ulong)m);
+  ulong xv  = Fl_invsafe(x, p);
+  if (!xv && p!=1UL) pari_err_INV("Fl_inv", mkintmod(utoi(x), utoi(p)));
+  return xv;
 }
-#endif
