@@ -1,5 +1,4 @@
-/*
-Copyright (C) 2006  The PARI group.
+/* Copyright (C) 2006  The PARI group.
 
 This file is part of the PARI package.
 
@@ -69,6 +68,7 @@ typedef enum {Llocal, Lmy} Ltype;
 struct vars_s
 {
   Ltype type; /*Only Llocal and Lmy are allowed */
+  int inl;
   entree *ep;
 };
 
@@ -91,21 +91,21 @@ static THREAD struct frame_s *frames;
 void
 pari_init_compiler(void)
 {
-  stack_init(&s_opcode,sizeof(*opcode),(void **)&opcode);
-  stack_init(&s_operand,sizeof(*operand),(void **)&operand);
-  stack_init(&s_data,sizeof(*data),(void **)&data);
-  stack_init(&s_lvar,sizeof(*localvars),(void **)&localvars);
-  stack_init(&s_dbginfo,sizeof(*dbginfo),(void **)&dbginfo);
-  stack_init(&s_frame,sizeof(*frames),(void **)&frames);
+  pari_stack_init(&s_opcode,sizeof(*opcode),(void **)&opcode);
+  pari_stack_init(&s_operand,sizeof(*operand),(void **)&operand);
+  pari_stack_init(&s_data,sizeof(*data),(void **)&data);
+  pari_stack_init(&s_lvar,sizeof(*localvars),(void **)&localvars);
+  pari_stack_init(&s_dbginfo,sizeof(*dbginfo),(void **)&dbginfo);
+  pari_stack_init(&s_frame,sizeof(*frames),(void **)&frames);
   offset=-1;
 }
 void
 pari_close_compiler(void)
 {
-  stack_delete(&s_opcode);
-  stack_delete(&s_operand);
-  stack_delete(&s_data);
-  stack_delete(&s_lvar);
+  pari_stack_delete(&s_opcode);
+  pari_stack_delete(&s_operand);
+  pari_stack_delete(&s_data);
+  pari_stack_delete(&s_lvar);
 }
 
 struct codepos
@@ -167,7 +167,7 @@ compilestate_restore(struct pari_compilestate *comp)
 }
 
 static GEN
-getfunction(struct codepos *pos, long arity, long nbmvar, GEN text)
+getfunction(const struct codepos *pos, long arity, long nbmvar, GEN text, long gap)
 {
   long lop =s_opcode.n+1-pos->opcode;
   long ldat=s_data.n+1-pos->data;
@@ -192,6 +192,7 @@ getfunction(struct codepos *pos, long arity, long nbmvar, GEN text)
     s[i] = opcode[i+pos->opcode-1];
     mael(cl, 3, i) = operand[i+pos->opcode-1];
     dbg[i] = dbginfo[i+pos->opcode-1]-dbgstart;
+    if (dbg[i]<0) dbg[i]+=gap;
   }
   s[i]=0;
   s_opcode.n=pos->opcode;
@@ -204,7 +205,8 @@ getfunction(struct codepos *pos, long arity, long nbmvar, GEN text)
       gunclone(data[i+pos->data-1]);
     }
   s_data.n=pos->data;
-  s_lvar.n=pos->localvars;
+  while (s_lvar.n>pos->localvars && !localvars[s_lvar.n-1].inl)
+    s_lvar.n--;
   for(i=1;i<lfram;i++)
   {
     long j=i+pos->frames-1;
@@ -221,15 +223,15 @@ getfunction(struct codepos *pos, long arity, long nbmvar, GEN text)
 static GEN
 getclosure(struct codepos *pos)
 {
-  return getfunction(pos,0,0,NULL);
+  return getfunction(pos,0,0,NULL,0);
 }
 
 static void
 op_push_loc(op_code o, long x, const char *loc)
 {
-  long n=stack_new(&s_opcode);
-  long m=stack_new(&s_operand);
-  long d=stack_new(&s_dbginfo);
+  long n=pari_stack_new(&s_opcode);
+  long m=pari_stack_new(&s_operand);
+  long d=pari_stack_new(&s_dbginfo);
   opcode[n]=o;
   operand[m]=x;
   dbginfo[d]=loc;
@@ -245,9 +247,9 @@ static void
 op_insert_loc(long k, op_code o, long x, const char *loc)
 {
   long i;
-  long n=stack_new(&s_opcode);
-  (void) stack_new(&s_operand);
-  (void) stack_new(&s_dbginfo);
+  long n=pari_stack_new(&s_opcode);
+  (void) pari_stack_new(&s_operand);
+  (void) pari_stack_new(&s_dbginfo);
   for (i=n-1; i>=k; i--)
   {
     opcode[i+1] = opcode[i];
@@ -262,7 +264,7 @@ op_insert_loc(long k, op_code o, long x, const char *loc)
 static long
 data_push(GEN x)
 {
-  long n=stack_new(&s_data);
+  long n=pari_stack_new(&s_data);
   data[n] = x?gclone(x):x;
   return n-offset;
 }
@@ -270,15 +272,16 @@ data_push(GEN x)
 static void
 var_push(entree *ep, Ltype type)
 {
-  long n=stack_new(&s_lvar);
+  long n=pari_stack_new(&s_lvar);
   localvars[n].ep   = ep;
+  localvars[n].inl  = 0;
   localvars[n].type = type;
 }
 
 static void
 frame_push(GEN x)
 {
-  long n=stack_new(&s_frame);
+  long n=pari_stack_new(&s_frame);
   frames[n].pc = s_opcode.n-1;
   frames[n].frame = gclone(x);
 }
@@ -301,19 +304,20 @@ pack_localvars(void)
 }
 
 void
-push_frame(GEN C, long lpc)
+push_frame(GEN C, long lpc, long dummy)
 {
-  char *code=GSTR(gel(C,2))-1;
-  GEN oper=gel(C,3);
-  GEN frpc=gmael(C,5,2);
-  GEN fram=gmael(C,5,3);
+  const char *code=closure_codestr(C);
+  GEN oper=closure_get_oper(C);
+  GEN dbg=closure_get_dbg(C);
+  GEN frpc=gel(dbg,2);
+  GEN fram=gel(dbg,3);
   long pc, j=1, lfr = lg(frpc);
   if (lpc==-1)
   {
     long k;
     GEN e = gel(fram, 1);
     for(k=1; k<lg(e); k++)
-      var_push((entree*)e[k], Lmy);
+      var_push(dummy?NULL:(entree*)e[k], Lmy);
     return;
   }
   if (lg(C)<8) while (j<lfr && frpc[j]==0) j++;
@@ -326,7 +330,7 @@ push_frame(GEN C, long lpc)
       long k;
       GEN e = gel(fram,j);
       for(k=1; k<lg(e); k++)
-        var_push((entree*)e[k], Lmy);
+        var_push(dummy?NULL:(entree*)e[k], Lmy);
       j++;
     }
   }
@@ -382,11 +386,12 @@ localvars_find(GEN pack, entree *ep)
 
 /*
  Flags for copy optimisation:
- -- FLreturn: The result must survive the closure.
+ -- Freturn: The result will be returned.
+ -- FLsurvive: The result must survive the closure.
  -- FLnocopy: The result will never be updated nor part of a user variable.
  -- FLnocopylex: The result will never be updated nor part of dynamic variable.
 */
-enum FLflag {FLreturn=1, FLnocopy=2, FLnocopylex=4};
+enum FLflag {FLreturn=1, FLsurvive=2, FLnocopy=4, FLnocopylex=8};
 
 static void
 copyifclone(long n, long mode, long flag, long mask)
@@ -394,7 +399,7 @@ copyifclone(long n, long mode, long flag, long mask)
   if (mode==Ggen && !(flag&mask))
   {
     op_push(OCcopyifclone,0,n);
-    if (!(flag&FLreturn) && DEBUGLEVEL)
+    if (!(flag&FLsurvive) && DEBUGLEVEL)
       pari_warn(warner,"compiler generates copy for `%.*s'",
                        tree[n].len,tree[n].str);
   }
@@ -425,6 +430,7 @@ parseproto(char const **q, char *c, const char *str)
     case 'V':
     case 'I':
     case 'E':
+    case 'J':
     case 'n':
     case 'P':
     case 'r':
@@ -464,6 +470,7 @@ parseproto(char const **q, char *c, const char *str)
       *c=*p;
     *q=p+1;
     return PPstd;
+  case 'E':
   case 's':
     if (p[1]=='*')
     {
@@ -525,7 +532,7 @@ compilecast_loc(int type, int mode, const char *loc)
     else compile_varerr(loc);
      break;
   default:
-    pari_err(bugparier,"compilecast, type unknown %ld",mode);
+    pari_err_BUG("compilecast [unknown type]");
   }
 }
 
@@ -599,11 +606,9 @@ ctxmvar(void)
 }
 
 INLINE int
-is_func_named(long x, const char *s)
+is_func_named(entree *ep, const char *s)
 {
-  if (tree[x].x!=CSTentry) return 0;
-  if (strlen(s)!=tree[x].len) return 0;
-  return !strncmp(tree[x].str, s, tree[x].len);
+  return !strcmp(ep->name, s);
 }
 
 INLINE int
@@ -622,17 +627,25 @@ str_defproto(const char *p, const char *q, const char *loc)
   op_push_loc(OCpushgen,data_push(strntoGENexp(q+1,len)),loc);
 }
 
+static long
+countlisttogen(long n, Ffunc f)
+{
+  long x,i;
+  if (n==-1 || tree[n].f==Fnoarg) return 0;
+  for(x=n, i=0; tree[x].f==f ;x=tree[x].x, i++);
+  return i+1;
+}
+
 static GEN
 listtogen(long n, Ffunc f)
 {
-  GEN z;
-  long x,i,nb;
-  if (n==-1 || tree[n].f==Fnoarg) return cgetg(1,t_VECSMALL);
-  for(x=n, i=0; tree[x].f==f ;x=tree[x].x, i++);
-  nb=i+1;
-  z=cgetg(nb+1, t_VECSMALL);
-  for (x=n; i>0; z[i+1]=tree[x].y, x=tree[x].x, i--);
-  z[1]=x;
+  long x,i,nb = countlisttogen(n, f);
+  GEN z=cgetg(nb+1, t_VECSMALL);
+  if (nb)
+  {
+    for (x=n, i = nb-1; i>0; z[i+1]=tree[x].y, x=tree[x].x, i--);
+    z[1]=x;
+  }
   return z;
 }
 
@@ -651,14 +664,56 @@ checkdups(GEN arg, GEN vep)
   if (l!=0) compile_err("variable declared twice",tree[arg[l]].str);
 }
 
+enum {MAT_range,MAT_std,MAT_line,MAT_column,VEC_std};
+
+static int
+matindex_type(long n)
+{
+  long x = tree[n].x, y = tree[n].y;
+  long fxx = tree[tree[x].x].f, fxy = tree[tree[x].y].f;
+  if (y==-1)
+  {
+    if (fxy!=Fnoarg) return MAT_range;
+    if (fxx==Fnoarg) compile_err("missing index",tree[n].str);
+    return VEC_std;
+  }
+  else
+  {
+    long fyx = tree[tree[y].x].f, fyy = tree[tree[y].y].f;
+    if (fxy!=Fnoarg || fyy!=Fnoarg) return MAT_range;
+    if (fxx==Fnoarg && fyx==Fnoarg) compile_err("missing index",tree[n].str);
+    if (fxx==Fnoarg) return MAT_column;
+    if (fyx==Fnoarg) return MAT_line;
+    return MAT_std;
+  }
+}
+
 static entree *
 getlvalue(long n)
 {
-  while (tree[n].f==Ffacteurmat || tree[n].f==Ftag)
+  while ((tree[n].f==Fmatcoeff && matindex_type(tree[n].y)!=MAT_range) || tree[n].f==Ftag)
     n=tree[n].x;
   return getvar(n);
 }
 
+INLINE void
+compilestore(long vn, entree *ep, long n)
+{
+  if (vn)
+    op_push(OCstorelex,vn,n);
+  else
+    op_push(OCstoredyn,(long)ep,n);
+}
+
+INLINE void
+compilenewptr(long vn, entree *ep, long n)
+{
+  if (vn)
+    op_push(OCnewptrlex,vn,n);
+  else
+    op_push(OCnewptrdyn,(long)ep,n);
+}
+
 static void
 compilelvalue(long n)
 {
@@ -667,74 +722,90 @@ compilelvalue(long n)
     return;
   else
   {
-    long x=tree[n].x;
-    long y=tree[n].y;
-    long yx=tree[y].x;
-    long yy=tree[y].y;
-    long f=tree[y].f;
-    if (tree[x].f==Ffacteurmat && f==Fmatrix && yy==-1 &&
-        tree[tree[x].y].f==FmatrixL)
+    long x = tree[n].x, y = tree[n].y;
+    long yx = tree[y].x, yy = tree[y].y;
+    long m = matindex_type(y);
+    if (m == MAT_range)
+      compile_err("not an lvalue",tree[n].str);
+    if (m == VEC_std && tree[x].f==Fmatcoeff)
     {
-      compilelvalue(tree[x].x);
-      compilenode(tree[tree[x].y].x,Gsmall,0);
-      compilenode(yx,Gsmall,0);
-      op_push(OCcompo2ptr,0,y);
-      return;
+      int mx = matindex_type(tree[x].y);
+      if (mx==MAT_line)
+      {
+        int xy = tree[x].y, xyx = tree[xy].x;
+        compilelvalue(tree[x].x);
+        compilenode(tree[xyx].x,Gsmall,0);
+        compilenode(tree[yx].x,Gsmall,0);
+        op_push(OCcompo2ptr,0,y);
+        return;
+      }
     }
     compilelvalue(x);
-    compilenode(yx,Gsmall,0);
-    if (f==Fmatrix && yy==-1)
-      op_push(OCcompo1ptr,0,y);
-    else
+    switch(m)
     {
-      switch(f)
-      {
-      case Fmatrix:
-        compilenode(yy,Gsmall,0);
-        op_push(OCcompo2ptr,0,y);
-        break;
-      case FmatrixR:
-        op_push(OCcompoCptr,0,y);
-        break;
-      case FmatrixL:
-        op_push(OCcompoLptr,0,y);
-        break;
-      }
+    case VEC_std:
+      compilenode(tree[yx].x,Gsmall,0);
+      op_push(OCcompo1ptr,0,y);
+      break;
+    case MAT_std:
+      compilenode(tree[yx].x,Gsmall,0);
+      compilenode(tree[yy].x,Gsmall,0);
+      op_push(OCcompo2ptr,0,y);
+      break;
+    case MAT_line:
+      compilenode(tree[yx].x,Gsmall,0);
+      op_push(OCcompoLptr,0,y);
+      break;
+    case MAT_column:
+      compilenode(tree[yy].x,Gsmall,0);
+      op_push(OCcompoCptr,0,y);
+      break;
     }
   }
 }
 
 static void
-compilefacteurmat(long n, int mode)
+compilematcoeff(long n, int mode)
 {
-  long x=tree[n].x;
-  long y=tree[n].y;
-  long yx=tree[y].x;
-  long yy=tree[y].y;
-  long f=tree[y].f;
+  long x=tree[n].x, y=tree[n].y;
+  long yx=tree[y].x, yy=tree[y].y;
+  long m=matindex_type(y);
   compilenode(x,Ggen,FLnocopy);
-  compilenode(yx,Gsmall,0);
-  if (f==Fmatrix && yy==-1)
+  switch(m)
   {
+  case VEC_std:
+    compilenode(tree[yx].x,Gsmall,0);
     op_push(OCcompo1,mode,y);
     return;
-  }
-  switch(f)
-  {
-  case Fmatrix:
-    compilenode(yy,Gsmall,0);
+  case MAT_std:
+    compilenode(tree[yx].x,Gsmall,0);
+    compilenode(tree[yy].x,Gsmall,0);
     op_push(OCcompo2,mode,y);
     return;
-  case FmatrixR:
-    op_push(OCcompoC,0,y);
-    compilecast(n,Gvec,mode);
-    return;
-  case FmatrixL:
+  case MAT_line:
+    compilenode(tree[yx].x,Gsmall,0);
     op_push(OCcompoL,0,y);
     compilecast(n,Gvec,mode);
     return;
+  case MAT_column:
+    compilenode(tree[yy].x,Gsmall,0);
+    op_push(OCcompoC,0,y);
+    compilecast(n,Gvec,mode);
+    return;
+  case MAT_range:
+    compilenode(tree[yx].x,Gsmall,0);
+    compilenode(tree[yx].y,Gsmall,0);
+    if (yy==-1)
+      op_push(OCcallgen,(long)is_entry("_[_.._]"),n);
+    else
+    {
+      compilenode(tree[yy].x,Gsmall,0);
+      compilenode(tree[yy].y,Gsmall,0);
+      op_push(OCcallgen,(long)is_entry("_[_.._,_.._]"),n);
+    }
+    return;
   default:
-    pari_err(bugparier,"compilefacteurmat");
+    pari_err_BUG("compilematcoeff");
   }
 }
 
@@ -761,7 +832,7 @@ compilevec(long n, long mode, op_code op)
   op_push(op,l,n);
   for (i=1;i<l;i++)
   {
-    compilenode(arg[i],Ggen,FLreturn);
+    compilenode(arg[i],Ggen,FLsurvive);
     op_push(OCstackgen,i,n);
   }
   avma=ltop;
@@ -795,7 +866,7 @@ compilemat(long n, long mode)
     for(j=1;j<lgcol;j++)
     {
       k-=lglin;
-      compilenode(col[j], Ggen, FLreturn);
+      compilenode(col[j], Ggen, FLsurvive);
       op_push(OCstackgen,k,n);
     }
   }
@@ -836,6 +907,19 @@ cattovec(long n, long fnum)
   return stack;
 }
 
+static GEN
+compilelambda(long n, long y, GEN vep, struct codepos *pos)
+{
+  long nbmvar, lev = vep ? lg(vep)-1 : 0;
+  GEN text=cgetg(3,t_VEC);
+  gel(text,1)=strtoGENstr(lev? ((entree*) vep[1])->name: "");
+  gel(text,2)=strntoGENstr(tree[y].str,tree[y].len);
+  nbmvar=ctxmvar()-lev;
+  if (lev) op_push(OCgetargs,lev,n);
+  compilenode(y,Ggen,FLsurvive|FLreturn);
+  return getfunction(pos,lev,nbmvar,text,2);
+}
+
 static void
 compilecall(long n, int mode)
 {
@@ -873,6 +957,199 @@ compileuserfunc(entree *ep, long n, int mode)
   compilecall(n, mode);
 }
 
+static GEN
+compilefuncinline(long n, long c, long a, long flag, long isif, long lev, long *ev)
+{
+  struct codepos pos;
+  int type=c=='I'?Gvoid:Ggen;
+  long rflag=c=='I'?0:FLsurvive;
+  GEN vep = NULL;
+  if (isif && (flag&FLreturn)) rflag|=FLreturn;
+  getcodepos(&pos);
+  if (lev)
+  {
+    long i;
+    GEN varg=cgetg(lev+1,t_VECSMALL);
+    vep=cgetg(lev+1,t_VECSMALL);
+    for(i=0;i<lev;i++)
+    {
+      entree *ve;
+      if (ev[i]<0)
+        compile_err("missing variable name", tree[a].str-1);
+      ve = getvar(ev[i]);
+      vep[i+1]=(long)ve;
+      varg[i+1]=ev[i];
+      var_push(ve,Lmy);
+    }
+    checkdups(varg,vep);
+    frame_push(vep);
+  }
+  if (c=='J')
+    return compilelambda(n,a,vep,&pos);
+  else if (tree[a].f==Fnoarg)
+    compilecast(a,Gvoid,type);
+  else
+    compilenode(a,type,rflag);
+  return getclosure(&pos);
+}
+
+static long
+countvar(GEN arg)
+{
+  long i, l = lg(arg);
+  long n = l-1;
+  for(i=1; i<l; i++)
+  {
+    long a=arg[i];
+    if (tree[a].f==Fassign)
+    {
+      long x = detag(tree[a].x);
+      if (tree[x].f==Fvec && tree[x].x>=0)
+        n += countlisttogen(tree[x].x,Fmatrixelts)-1;
+    }
+  }
+  return n;
+}
+
+static void
+compileuninline(GEN arg)
+{
+  long j;
+  if (lg(arg) > 1)
+    compile_err("too many arguments",tree[arg[1]].str);
+  for(j=0; j<s_lvar.n; j++)
+    if(!localvars[j].inl)
+      pari_err(e_MISC,"uninline is only valid at top level");
+  s_lvar.n = 0;
+}
+
+static void
+compilemy(GEN arg, const char *str, int inl)
+{
+  long i, j, k, l = lg(arg);
+  long n = countvar(arg);
+  GEN vep = cgetg(n+1,t_VECSMALL);
+  GEN ver = cgetg(n+1,t_VECSMALL);
+  if (inl)
+  {
+    for(j=0; j<s_lvar.n; j++)
+      if(!localvars[j].inl)
+        pari_err(e_MISC,"inline is only valid at top level");
+  }
+  for(k=0, i=1; i<l; i++)
+  {
+    long a=arg[i];
+    if (tree[a].f==Fassign)
+    {
+      long x = detag(tree[a].x);
+      if (tree[x].f==Fvec && tree[x].x>=0)
+      {
+        GEN vars = listtogen(tree[x].x,Fmatrixelts);
+        long nv = lg(vars)-1;
+        for (j=1; j<=nv; j++)
+        {
+          ver[++k] = vars[j];
+          vep[k] = (long)getvar(ver[k]);
+        }
+        continue;
+      } else ver[++k] = x;
+    } else ver[++k] = a;
+    vep[k] = (long)getvar(ver[k]);
+  }
+  checkdups(ver,vep);
+  for(i=1; i<=n; i++) var_push(NULL,Lmy);
+  op_push_loc(OCnewframe,inl?-n:n,str);
+  frame_push(vep);
+  for (k=0, i=1; i<l; i++)
+  {
+    long a=arg[i];
+    if (tree[a].f==Fassign)
+    {
+      long x = detag(tree[a].x);
+      if (tree[x].f==Fvec && tree[x].x>=0)
+      {
+        GEN vars = listtogen(tree[x].x,Fmatrixelts);
+        long nv = lg(vars)-1;
+        compilenode(tree[a].y,Ggen,FLnocopy);
+        op_push(OCdup,nv-1,x);
+        for (j=1; j<=nv; j++)
+        {
+          long v = detag(vars[j]);
+          op_push(OCpushlong,j,v);
+          op_push(OCcompo1,Ggen,v);
+          k++;
+          op_push(OCstorelex,-n+k-1,a);
+          localvars[s_lvar.n-n+k-1].ep=(entree*)vep[k];
+          localvars[s_lvar.n-n+k-1].inl=inl;
+        }
+        continue;
+      }
+      else if (!is_node_zero(tree[a].y))
+      {
+        compilenode(tree[a].y,Ggen,FLnocopy);
+        op_push(OCstorelex,-n+k,a);
+      }
+    }
+    k++;
+    localvars[s_lvar.n-n+k-1].ep=(entree*)vep[k];
+    localvars[s_lvar.n-n+k-1].inl=inl;
+  }
+}
+
+static long
+localpush(op_code op, long a)
+{
+  entree *ep = getvar(a);
+  long vep  = (long) ep;
+  op_push(op,vep,a);
+  var_push(ep,Llocal);
+  return vep;
+}
+
+static void
+compilelocal(GEN arg)
+{
+  long i, j, k, l = lg(arg);
+  long n = countvar(arg);
+  GEN vep = cgetg(n+1,t_VECSMALL);
+  GEN ver = cgetg(n+1,t_VECSMALL);
+  for(k=0, i=1; i<l; i++)
+  {
+    long a=arg[i];
+    if (tree[a].f==Fassign)
+    {
+      long x = detag(tree[a].x);
+      if (tree[x].f==Fvec && tree[x].x>=0)
+      {
+        GEN vars = listtogen(tree[x].x,Fmatrixelts);
+        long nv = lg(vars)-1;
+        compilenode(tree[a].y,Ggen,FLnocopy);
+        op_push(OCdup,nv-1,x);
+        for (j=1; j<=nv; j++)
+        {
+          long v = detag(vars[j]);
+          op_push(OCpushlong,j,v);
+          op_push(OCcompo1,Ggen,v);
+          vep[++k] = localpush(OClocalvar, v);
+          ver[k] = v;
+        }
+        continue;
+      } else if (!is_node_zero(tree[a].y))
+      {
+        compilenode(tree[a].y,Ggen,FLnocopy);
+        ver[++k] = x;
+        vep[k] = localpush(OClocalvar, ver[k]);
+        continue;
+      }
+      else
+        ver[++k] = x;
+    } else
+      ver[++k] = a;
+    vep[k] = localpush(OClocalvar0, ver[k]);
+  }
+  checkdups(ver,vep);
+}
+
 static void
 compilefunc(entree *ep, long n, int mode, long flag)
 {
@@ -905,81 +1182,61 @@ compilefunc(entree *ep, long n, int mode, long flag)
       str=tree[n].str;
     while(*str==')') str++;
   }
-  if (tree[n].f==Faffect)
+  if (tree[n].f==Fassign)
   {
     nb=2; lnc=2; lnl=2; arg=mkvecsmall2(x,y);
   }
-  else if (is_func_named(x,"if") && mode==Gvoid)
-    ep=is_entry("_void_if");
-  else if (is_func_named(x,"my"))
+  else if (is_func_named(ep,"if"))
   {
-    long lgarg;
-    GEN vep = cgetg_copy(arg, &lgarg);
-
-    if (nb)
-    {
-      long i;
-      for(i=1;i<=nb;i++)
-      {
-        long a=arg[i];
-        vep[i]=(long)getvar(tree[a].f==Faffect?tree[a].x:a);
-        var_push(NULL,Lmy);
-      }
-      checkdups(arg,vep);
-      op_push_loc(OCnewframe,nb,str);
-      frame_push(vep);
-      for (i=1;i<=nb;i++)
-      {
-        long a=arg[i];
-        if (tree[a].f==Faffect && !is_node_zero(tree[a].y))
-        {
-          compilenode(tree[a].y,Ggen,FLnocopy);
-          op_push(OCstorelex,-nb+i-1,a);
-        }
-        localvars[s_lvar.n-nb+i-1].ep=(entree*)vep[i];
-      }
-    }
+    if (nb>=4)
+      ep=is_entry("_multi_if");
+    else if (mode==Gvoid)
+      ep=is_entry("_void_if");
+  }
+  else if (is_func_named(ep,"return") && (flag&FLreturn) && nb<=1)
+  {
+    if (nb==0) op_push(OCpushgnil,0,n);
+    else compilenode(arg[1],Ggen,FLsurvive|FLreturn);
+    avma=ltop;
+    return;
+  }
+  else if (is_func_named(ep,"inline"))
+  {
+    compilemy(arg, str, 1);
     compilecast(n,Gvoid,mode);
     avma=ltop;
     return;
   }
-  else if (is_func_named(x,"local"))
+  else if (is_func_named(ep,"uninline"))
   {
-    long lgarg;
-    long i;
-    GEN vep = cgetg_copy(arg, &lgarg);
-    for (i=1;i<=nb;i++)
-    {
-      entree *en;
-      long a=arg[i];
-      op_code op=OClocalvar0;
-      if (tree[a].f==Faffect)
-      {
-        if (!is_node_zero(tree[a].y))
-        {
-          compilenode(tree[a].y,Ggen,FLnocopy);
-          op=OClocalvar;
-        }
-        a=tree[a].x;
-      }
-      vep[i] = (long) (en = getvar(a));
-      op_push(op,vep[i],a);
-      var_push(en,Llocal);
-    }
-    checkdups(arg,vep);
+    compileuninline(arg);
+    compilecast(n,Gvoid,mode);
+    avma=ltop;
+    return;
+  }
+  else if (is_func_named(ep,"my"))
+  {
+    compilemy(arg, str, 0);
+    compilecast(n,Gvoid,mode);
+    avma=ltop;
+    return;
+  }
+  else if (is_func_named(ep,"local"))
+  {
+    compilelocal(arg);
     compilecast(n,Gvoid,mode);
     avma=ltop;
     return;
   }
   /*We generate dummy code for global() for compatibility with gp2c*/
-  else if (is_func_named(x,"global"))
+  else if (is_func_named(ep,"global"))
   {
     long i;
     for (i=1;i<=nb;i++)
     {
       long a=arg[i];
       long en;
-      if (tree[a].f==Faffect)
+      if (tree[a].f==Fassign)
       {
         compilenode(tree[a].y,Ggen,0);
         a=tree[a].x;
@@ -997,7 +1254,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
     avma=ltop;
     return;
   }
-  else if (is_func_named(x,"O") || (compatible==OLDALL && is_func_named(x,"o")))
+  else if (is_func_named(ep,"O") || (compatible==OLDALL && is_func_named(ep,"o")))
   {
     if (nb!=1)
       compile_err("wrong number of arguments", tree[n].str+tree[n].len-1);
@@ -1045,7 +1302,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
       {
       case PPstd:
         if (j>nb) compile_err("too few arguments", tree[n].str+tree[n].len-1);
-        if (c!='I' && c!='E')
+        if (c!='I' && c!='E' && c!='J')
         {
           long x = tree[arg[j]].x, f = tree[arg[j]].f;
           if (f==Fnoarg)
@@ -1118,10 +1375,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
             }
             else
             {
-              if (vn)
-                op_push(OCnewptrlex, vn,n);
-              else
-                op_push(OCnewptrdyn, (long)ep,n);
+              compilenewptr(vn,ep,n);
               compilelvalue(a);
               op_push(OCpushptr, 0,n);
             }
@@ -1130,35 +1384,12 @@ compilefunc(entree *ep, long n, int mode, long flag)
           }
         case 'I':
         case 'E':
+        case 'J':
           {
-            struct codepos pos;
-            long a=arg[j++];
-            int type=c=='I'?Gvoid:Ggen;
-            long flag=c=='I'?0:FLreturn;
-            getcodepos(&pos);
-            if (lev)
-            {
-              long i;
-              GEN vep=cgetg(lev+1,t_VECSMALL);
-              GEN varg=cgetg(lev+1,t_VECSMALL);
-              for(i=0;i<lev;i++)
-              {
-                entree *ve;
-                if (ev[i]<0)
-                  compile_err("missing variable name", tree[a].str-1);
-                ve = getvar(ev[i]);
-                vep[i+1]=(long)ve;
-                varg[i+1]=ev[i];
-                var_push(ve,Lmy);
-              }
-              checkdups(varg,vep);
-              frame_push(vep);
-            }
-            if (tree[a].f==Fnoarg)
-              compilecast(a,Gvoid,type);
-            else
-              compilenode(a,type,flag);
-            op_push(OCpushgen, data_push(getclosure(&pos)),a);
+            long a = arg[j++];
+            GEN  d = compilefuncinline(n, c, a, flag, is_func_named(ep,"if"), lev, ev);
+            op_push(OCpushgen, data_push(d), a);
+            if (lg(d)==8) op_push(OCsaveframe,FLsurvive,n);
             break;
           }
         case 'V':
@@ -1205,7 +1436,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
               op_push(OCvec, nb+1, a);
               for(l=1; l<=nb; l++)
               {
-                compilenode(g[l], Ggen, FLreturn);
+                compilenode(g[l], Ggen, FLsurvive);
                 op_push(OCstackgen,l, a);
               }
               op_push(OCpop, 1, a);
@@ -1215,7 +1446,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
             break;
           }
         default:
-          pari_err(talker,"Unknown prototype code `%c' for `%.*s'",c,
+          pari_err(e_MISC,"Unknown prototype code `%c' for `%.*s'",c,
               tree[x].len, tree[x].str);
         }
         break;
@@ -1261,7 +1492,7 @@ compilefunc(entree *ep, long n, int mode, long flag)
           op_push(OCprecdl,0,n);
           break;
         default:
-          pari_err(talker,"Unknown prototype code `%c' for `%.*s'",c,
+          pari_err(e_MISC,"Unknown prototype code `%c' for `%.*s'",c,
               tree[x].len, tree[x].str);
         }
         break;
@@ -1282,13 +1513,23 @@ compilefunc(entree *ep, long n, int mode, long flag)
           op_push(OCtostr, -1, n);
           break;
         default:
-          pari_err(talker,"Unknown prototype code `%c' for `%.*s'",c,
+          pari_err(e_MISC,"Unknown prototype code `%c' for `%.*s'",c,
               tree[x].len, tree[x].str);
         }
         break;
       case PPstar:
         switch(c)
         {
+        case 'E':
+          {
+            long k, n=nb+1-j;
+            GEN g=cgetg(n+1,t_VEC);
+            for(k=1; k<=n; k++)
+              gel(g, k) = compilefuncinline(n, c, arg[j+k-1], flag, 1, lev, ev);
+            op_push(OCpushgen, data_push(g), arg[j]);
+            j=nb+1;
+            break;
+          }
         case 's':
           {
             long n=nb+1-j;
@@ -1297,13 +1538,13 @@ compilefunc(entree *ep, long n, int mode, long flag)
             for(l1=0,k=1;k<=n;k++)
             {
               gel(g,k)=cattovec(arg[j+k-1],OPcat);
-              l1+=lg(g[k])-1;
+              l1+=lg(gel(g,k))-1;
             }
             op_push_loc(OCvec, l1+1, str);
             for(m=1,k=1;k<=n;k++)
-              for(l=1;l<lg(g[k]);l++,m++)
+              for(l=1;l<lg(gel(g,k));l++,m++)
               {
-                compilenode(mael(g,k,l),Ggen,FLreturn);
+                compilenode(mael(g,k,l),Ggen,FLsurvive);
                 op_push(OCstackgen,m,mael(g,k,l));
               }
             op_push_loc(OCpop, 1, str);
@@ -1311,12 +1552,12 @@ compilefunc(entree *ep, long n, int mode, long flag)
             break;
           }
         default:
-          pari_err(talker,"Unknown prototype code `%c*' for `%.*s'",c,
+          pari_err(e_MISC,"Unknown prototype code `%c*' for `%.*s'",c,
               tree[x].len, tree[x].str);
         }
         break;
       default:
-        pari_err(bugparier,"PPproto %d in compilefunc",mod);
+        pari_err_BUG("compilefunc [unknown PPproto]");
       }
       q=p;
     }
@@ -1384,7 +1625,7 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
   getcodepos(&pos);
   dbgstart = loc;
   if (nbdata > arity)
-    pari_err(talker,"too many parameters for closure `%s'", ep->name);
+    pari_err(e_MISC,"too many parameters for closure `%s'", ep->name);
   for(i=1; i<= nbdata; i++)
     op_push_loc(OCpushgen,data_push(NULL),loc);
   arity -= nbdata;
@@ -1468,7 +1709,7 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
         op_push_loc(OCdefaultlong,-index,loc);
         break;
       default:
-        pari_err(talker,"Unknown prototype code `D%c' for `%s'",c,ep->name);
+        pari_err(e_MISC,"Unknown prototype code `D%c' for `%s'",c,ep->name);
       }
       break;
     case PPdefaultmulti:
@@ -1490,17 +1731,18 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
         op_push_loc(OCtostr,-index,loc);
         break;
       default:
-        pari_err(talker,
+        pari_err(e_MISC,
             "Unknown prototype code `D...,%c,' for `%s'",c,ep->name);
       }
       break;
     case PPstar:
       switch(c)
       {
+      case 'E':
       case 's':
         return NULL;
       default:
-        pari_err(talker,"Unknown prototype code `%c*' for `%s'",c,ep->name);
+        pari_err(e_MISC,"Unknown prototype code `%c*' for `%s'",c,ep->name);
       }
       break;
     default:
@@ -1512,7 +1754,7 @@ genclosure(entree *ep, const char *loc, long  nbdata, int check)
   op_push_loc(ret_op, (long) ep, loc);
   if (ret_flag==FLnocopy) op_push_loc(OCcopy,0,loc);
   compilecast_loc(ret_typ, Ggen, loc);
-  return getfunction(&pos,nb+arity,0,strtoGENstr(ep->name));
+  return getfunction(&pos,nb+arity,0,strtoGENstr(ep->name),0);
 }
 
 GEN
@@ -1532,12 +1774,12 @@ strtoclosure(const char *s, long n,  ...)
   pari_sp av = avma;
   entree *ep = is_entry(s);
   GEN C;
-  if (!ep) pari_err(talker,"no function named \"%s\"",s);
+  if (!ep) pari_err(e_MISC,"no function named \"%s\"",s);
   ep = do_alias(ep);
   if ((!EpSTATIC(ep) && EpVALENCE(ep)!=EpINSTALL) || !ep->value)
-    pari_err(talker,"not a built-in/install'ed function: \"%s\"",s);
+    pari_err(e_MISC,"not a built-in/install'ed function: \"%s\"",s);
   C = genclosure(ep,ep->name,n,0);
-  if (!C) pari_err(talker,"function prototype unsupported: \"%s\"",s);
+  if (!C) pari_err(e_MISC,"function prototype unsupported: \"%s\"",s);
   else
   {
     va_list ap;
@@ -1574,55 +1816,86 @@ closurefunc(entree *ep, long n, long mode)
   avma=ltop;
 }
 
+static void
+compileseq(long n, int mode, long flag)
+{
+  pari_sp av = avma;
+  GEN L = listtogen(n, Fseq);
+  long i, l = lg(L)-1;
+  for(i = 1; i < l; i++)
+    compilenode(L[i],Gvoid,0);
+  compilenode(L[l],mode,flag&(FLreturn|FLsurvive));
+  avma = av;
+}
+
 static void
 compilenode(long n, int mode, long flag)
 {
   long x,y;
-  if (n<0)
-    pari_err(bugparier,"compilenode");
+#ifdef STACK_CHECK
+  if (PARI_stack_limit && (void*) &x <= PARI_stack_limit)
+    pari_err(e_MISC, "expression nested too deeply");
+#endif
+  if (n<0) pari_err_BUG("compilenode");
   x=tree[n].x;
   y=tree[n].y;
 
   switch(tree[n].f)
   {
   case Fseq:
-    if (tree[x].f!=Fnoarg)
-      compilenode(x,Gvoid,0);
-    compilenode(y,mode,flag&FLreturn);
+    compileseq(n, mode, flag);
     return;
-  case Ffacteurmat:
-    compilefacteurmat(n,mode);
+  case Fmatcoeff:
+    compilematcoeff(n,mode);
     if (mode==Ggen && !(flag&FLnocopy))
       op_push(OCcopy,0,n);
     return;
-  case Faffect:
+  case Fassign:
     x = detag(x);
-    if (tree[x].f==Fentry)
+    if (tree[x].f==Fvec && tree[x].x>=0)
     {
-      entree *ep=getvar(x);
-      long vn=getmvar(ep);
-      compilenode(y,Ggen,FLnocopy);
-      if (vn)
-        op_push(OCstorelex,vn,n);
-      else
-        op_push(OCstoredyn,(long)ep,n);
-      if (mode!=Gvoid)
+      GEN vars = listtogen(tree[x].x,Fmatrixelts);
+      long i, l = lg(vars)-1;
+      compilenode(y,Ggen,mode==Gvoid?FLnocopy:flag&FLsurvive);
+      op_push(OCdup,mode==Gvoid?l-1:l,x);
+      for(i=1; i<=l; i++)
       {
-        if (vn)
-        {
-          op_push(OCpushlex,vn,n);
-          copyifclone(n,mode,flag,FLnocopy|FLnocopylex);
-        }
+        long a = detag(vars[i]);
+        entree *ep=getlvalue(a);
+        long vn=getmvar(ep);
+        op_push(OCpushlong,i,a);
+        op_push(OCcompo1,Ggen,a);
+        if (tree[a].f==Fentry)
+          compilestore(vn,ep,n);
         else
         {
-          op_push(OCpushdyn,(long)ep,n);
-          copyifclone(n,mode,flag,FLnocopy);
+          compilenewptr(vn,ep,n);
+          compilelvalue(a);
+          op_push(OCstoreptr,0,a);
         }
-        compilecast(n,Ggen,mode);
       }
+      if (mode!=Gvoid)
+        compilecast(n,Ggen,mode);
     }
     else
-      compilefunc(is_entry("_=_"),n,mode,0);
+    {
+      entree *ep=getlvalue(x);
+      long vn=getmvar(ep);
+      if (tree[x].f!=Fentry)
+      {
+        compilenewptr(vn,ep,n);
+        compilelvalue(x);
+      }
+      compilenode(y,Ggen,mode==Gvoid?FLnocopy:flag&FLsurvive);
+      if (mode!=Gvoid)
+        op_push(OCdup,1,n);
+      if (tree[x].f==Fentry)
+        compilestore(vn,ep,n);
+      else
+        op_push(OCstoreptr,0,x);
+      if (mode!=Gvoid)
+        compilecast(n,Ggen,mode);
+    }
     return;
   case Fconst:
     {
@@ -1655,7 +1928,7 @@ compilenode(long n, int mode, long flag)
           break;
         }
       default:
-        pari_err(bugparier,"compilenode, unsupported constant");
+        pari_err_BUG("compilenode, unsupported constant");
       }
       avma=ltop;
       return;
@@ -1712,13 +1985,15 @@ compilenode(long n, int mode, long flag)
       pari_sp ltop=avma;
       struct codepos pos;
       GEN arg=listtogen(x,Flistarg);
-      long nb, lgarg, nbmvar;
+      long nb, lgarg, nbmvar, gap;
+      long strict = GP_DATA->strictargs;
       GEN vep = cgetg_copy(arg, &lgarg);
       GEN text=cgetg(3,t_VEC);
       gel(text,1)=strntoGENstr(tree[x].str,tree[x].len);
       gel(text,2)=strntoGENstr(tree[y].str,tree[y].len);
       getcodepos(&pos);
       dbgstart=tree[x].str+tree[x].len;
+      gap = tree[y].str-dbgstart;
       nbmvar=ctxmvar();
       nb = lgarg-1;
       if (nb)
@@ -1727,33 +2002,41 @@ compilenode(long n, int mode, long flag)
         for(i=1;i<=nb;i++)
         {
           long a=arg[i];
-          vep[i]=(long)getvar(tree[a].f==Faffect?tree[a].x:a);
+          vep[i]=(long)getvar(tree[a].f==Fassign?tree[a].x:a);
           var_push(NULL,Lmy);
         }
         checkdups(arg,vep);
-        op_push(OCgetargs,nb,y);
+        op_push(OCgetargs,nb,x);
         frame_push(vep);
         for (i=1;i<=nb;i++)
         {
           long a=arg[i];
-          if (tree[a].f==Faffect)
+          long y = tree[a].y;
+          if (tree[a].f==Fassign && (strict || !is_node_zero(y)))
           {
-            struct codepos lpos;
-            getcodepos(&lpos);
-            compilenode(tree[a].y,Ggen,0);
-            op_push(OCpushgen, data_push(getclosure(&lpos)),a);
+            if (tree[y].f==Fsmall)
+              compilenode(y,Ggen,a);
+            else
+            {
+              struct codepos lpos;
+              getcodepos(&lpos);
+              compilenode(y,Ggen,a);
+              op_push(OCpushgen, data_push(getclosure(&lpos)),a);
+            }
             op_push(OCdefaultarg,-nb+i-1,a);
           }
           localvars[s_lvar.n-nb+i-1].ep=(entree*)vep[i];
         }
       }
+      if (strict)
+        op_push(OCcheckuserargs,nb,x);
       dbgstart=tree[y].str;
       if (y>=0 && tree[y].f!=Fnoarg)
-        compilenode(y,Ggen,FLreturn);
+        compilenode(y,Ggen,FLsurvive|FLreturn);
       else
         compilecast(n,Gvoid,Ggen);
-      op_push(OCpushgen, data_push(getfunction(&pos,nb,nbmvar,text)),n);
-      if (nbmvar) op_push(OCsaveframe,!!(flag&FLreturn),n);
+      op_push(OCpushgen, data_push(getfunction(&pos,nb,nbmvar,text,gap)),n);
+      if (nbmvar) op_push(OCsaveframe,!!(flag&FLsurvive),n);
       compilecast(n, Gclosure, mode);
       avma=ltop;
       return;
@@ -1765,7 +2048,7 @@ compilenode(long n, int mode, long flag)
     compilecast(n,Gvoid,mode);
     return;
   default:
-    pari_err(bugparier,"compilenode");
+    pari_err_BUG("compilenode");
   }
 }
 
@@ -1775,8 +2058,8 @@ gp_closure(long n)
   struct codepos pos;
   getcodepos(&pos);
   dbgstart=tree[n].str;
-  compilenode(n,Ggen,FLreturn);
-  return getfunction(&pos,0,0,strntoGENstr(tree[n].str,tree[n].len));
+  compilenode(n,Ggen,FLsurvive|FLreturn);
+  return getfunction(&pos,0,0,strntoGENstr(tree[n].str,tree[n].len),0);
 }
 
 GEN
@@ -1813,7 +2096,7 @@ closure_deriv(GEN G)
   op_push_loc(OCpop,1,code);
   op_push_loc(OCprecreal,0,code);
   op_push_loc(OCcallgen,(long)is_entry("_derivfun"),code);
-  return gerepilecopy(ltop, getfunction(&pos,arity,0,text));
+  return gerepilecopy(ltop, getfunction(&pos,arity,0,text,0));
 }
 
 static long
@@ -1856,18 +2139,17 @@ optimizemat(long n)
 }
 
 static void
-optimizefacteurmat(long n)
+optimizematcoeff(long n)
 {
   long x=tree[n].x;
   long y=tree[n].y;
   long yx=tree[y].x;
   long yy=tree[y].y;
-  long f=tree[y].f;
   long fl;
   optimizenode(x);
   optimizenode(yx);
   fl=tree[x].flags&tree[yx].flags;
-  if (f==Fmatrix && yy>=0)
+  if (yy>=0)
   {
     optimizenode(yy);
     fl&=tree[yy].flags;
@@ -1920,6 +2202,7 @@ optimizefunc(entree *ep, long n)
           break;
         case 'I':
         case 'E':
+        case 'J':
           optimizenode(arg[j]);
           fl&=tree[arg[j]].flags;
           tree[arg[j++]].flags=COsafelex|COsafedyn;
@@ -1949,7 +2232,7 @@ optimizefunc(entree *ep, long n)
         case '=':
           {
             long a=arg[j++], y=tree[a].y;
-            if (tree[a].f!=Faffect)
+            if (tree[a].f!=Fassign)
               compile_err("expected character: '=' instead of",
                   tree[a].str+tree[a].len);
             optimizenode(y);
@@ -1960,7 +2243,7 @@ optimizefunc(entree *ep, long n)
           fl &= vec_optimize(cattovec(arg[j++], OPcat));
           break;
         default:
-          pari_err(talker,"Unknown prototype code `%c' for `%.*s'",c,
+          pari_err(e_MISC,"Unknown prototype code `%c' for `%.*s'",c,
               tree[x].len, tree[x].str);
         }
         break;
@@ -1973,6 +2256,18 @@ optimizefunc(entree *ep, long n)
       case PPstar:
         switch(c)
         {
+        case 'E':
+          {
+            long n=nb+1-j;
+            long k;
+            for(k=1;k<=n;k++)
+            {
+              optimizenode(arg[j+k-1]);
+              fl &= tree[arg[j+k-1]].flags;
+            }
+            j=nb+1;
+            break;
+          }
         case 's':
           {
             long n=nb+1-j;
@@ -1983,12 +2278,12 @@ optimizefunc(entree *ep, long n)
             break;
           }
         default:
-          pari_err(talker,"Unknown prototype code `%c*' for `%.*s'",c,
+          pari_err(e_MISC,"Unknown prototype code `%c*' for `%.*s'",c,
               tree[x].len, tree[x].str);
         }
         break;
       default:
-        pari_err(bugparier,"PPproto %d in compilefunc",mod);
+        pari_err_BUG("optimizefun [unknown PPproto]");
       }
     }
   }
@@ -2008,27 +2303,48 @@ optimizecall(long n)
   avma=av;
 }
 
+static void
+optimizeseq(long n)
+{
+  pari_sp av = avma;
+  GEN L = listtogen(n, Fseq);
+  long i, l = lg(L)-1, flags=-1L;
+  for(i = 1; i <= l; i++)
+  {
+    optimizenode(L[i]);
+    flags &= tree[L[i]].flags;
+  }
+  avma = av;
+  tree[n].flags = flags;
+}
+
 void
 optimizenode(long n)
 {
   long x,y;
+#ifdef STACK_CHECK
+  if (PARI_stack_limit && (void*) &x <= PARI_stack_limit)
+    pari_err(e_MISC, "expression nested too deeply");
+#endif
   if (n<0)
-    pari_err(bugparier,"optimizenode");
+    pari_err_BUG("optimizenode");
   x=tree[n].x;
   y=tree[n].y;
 
   switch(tree[n].f)
   {
   case Fseq:
-    if (tree[x].f!=Fnoarg)
-      optimizenode(x);
+    optimizeseq(n);
+    return;
+  case Frange:
+    optimizenode(x);
     optimizenode(y);
     tree[n].flags=tree[x].flags&tree[y].flags;
-    return;
-  case Ffacteurmat:
-    optimizefacteurmat(n);
     break;
-  case Faffect:
+  case Fmatcoeff:
+    optimizematcoeff(n);
+    break;
+  case Fassign:
     optimizenode(x);
     optimizenode(y);
     tree[n].flags=0;
@@ -2069,6 +2385,6 @@ optimizenode(long n)
     tree[n].flags=tree[x].flags;
     return;
   default:
-    pari_err(bugparier,"optimizenode");
+    pari_err_BUG("optimizenode");
   }
 }
