@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -49,7 +47,7 @@ get_sep(const char *t)
         if (! (*s++ = *t++) ) return buf;
     }
     if (s == lim)
-      pari_err(talker,"get_sep: argument too long (< %ld chars)", GET_SEP_SIZE);
+      pari_err(e_MISC,"get_sep: argument too long (< %ld chars)", GET_SEP_SIZE);
   }
 }
 
@@ -71,10 +69,10 @@ my_int(char *s)
 
   while (isdigit((int)*p)) {
     ulong m;
-    if (n > (~0UL / 10)) pari_err(syntaxer,"integer too large",s,s);
+    if (n > (~0UL / 10)) pari_err(e_SYNTAX,"integer too large",s,s);
     n *= 10; m = n;
     n += *p++ - '0';
-    if (n < m) pari_err(syntaxer,"integer too large",s,s);
+    if (n < m) pari_err(e_SYNTAX,"integer too large",s,s);
   }
   if (n)
   {
@@ -84,9 +82,9 @@ my_int(char *s)
       case 'm': case 'M': n = safe_mul(n,1000000UL);    p++; break;
       case 'g': case 'G': n = safe_mul(n,1000000000UL); p++; break;
     }
-    if (!n) pari_err(syntaxer,"integer too large",s,s);
+    if (!n) pari_err(e_SYNTAX,"integer too large",s,s);
   }
-  if (*p) pari_err(syntaxer,"I was expecting an integer here", s, s);
+  if (*p) pari_err(e_SYNTAX,"I was expecting an integer here", s, s);
   return n;
 }
 
@@ -101,7 +99,7 @@ get_int(const char *s, long dflt)
   if (!isdigit((int)*p)) return dflt;
 
   n = (long)my_int(p);
-  if (n < 0) pari_err(syntaxer,"integer too large",s,s);
+  if (n < 0) pari_err(e_SYNTAX,"integer too large",s,s);
   return minus? -n: n;
 }
 
@@ -109,7 +107,7 @@ ulong
 get_uint(const char *s)
 {
   char *p = get_sep(s);
-  if (*p == '-') pari_err(syntaxer,"arguments must be positive integers",s,s);
+  if (*p == '-') pari_err(e_SYNTAX,"arguments must be positive integers",s,s);
   return my_int(p);
 }
 
@@ -143,9 +141,9 @@ sd_toggle(const char *v, long flag, const char *s, int *ptn)
     if (n == state) return gnil;
     if (n != !state)
     {
-      char *t = stackmalloc(64 + strlen(s));
+      char *t = stack_malloc(64 + strlen(s));
       (void)sprintf(t, "default: incorrect value for %s [0:off / 1:on]", s);
-      pari_err(syntaxer, t, v,v);
+      pari_err(e_SYNTAX, t, v,v);
     }
     state = *ptn = n;
   }
@@ -168,10 +166,10 @@ sd_ulong_init(const char *v, const char *s, ulong *ptn, ulong Min, ulong Max)
     ulong n = get_uint(v);
     if (n > Max || n < Min)
     {
-      char *buf = stackmalloc(strlen(s) + 2 * 20 + 40);
+      char *buf = stack_malloc(strlen(s) + 2 * 20 + 40);
       (void)sprintf(buf, "default: incorrect value for %s [%lu-%lu]",
                     s, Min, Max);
-      pari_err(syntaxer, buf, v,v);
+      pari_err(e_SYNTAX, buf, v,v);
     }
     *ptn = n;
   }
@@ -216,7 +214,7 @@ sd_realprecision(const char *v, long flag)
     if (prec == precreal) return gnil;
     precreal = prec;
   }
-  if (flag == d_RETURN) return stoi(fmt->sigd);
+  if (flag == d_RETURN) return stoi(prec2ndec(precreal));
   if (flag == d_ACKNOWLEDGE)
   {
     long n = prec2ndec(precreal);
@@ -253,7 +251,7 @@ gp_get_color(char **st)
       long i = 0;
       for (a[0] = s = ++v; *s && *s != ']'; s++)
         if (*s == ',') { *s = 0; a[++i] = s+1; }
-      if (*s != ']') pari_err(syntaxer,"expected character: ']'",s, *st);
+      if (*s != ']') pari_err(e_SYNTAX,"expected character: ']'",s, *st);
       *s = 0; for (i++; i<3; i++) a[i] = "";
       /*    properties    |   color    | background */
       c = (atoi(a[2])<<8) | atoi(a[0]) | (atoi(a[1])<<4);
@@ -284,6 +282,8 @@ sd_colors(const char *v, long flag)
       v = "1, 5, 3, 7, 6, 2, 3"; /* Assume recent ReadLine. */
     if (l <= 7 && strncmp(v, "lightbg", l) == 0)
       v = "1, 6, 3, 4, 5, 2, 3"; /* Assume recent ReadLine. */
+    if (l <= 8 && strncmp(v, "brightfg", l) == 0)      /* Good for windows consoles */
+      v = "9, 13, 11, 15, 14, 10, 11";
     if (l <= 6 && strncmp(v, "boldfg", l) == 0)        /* Good for darkbg consoles */
       v = "[1,,1], [5,,1], [3,,1], [7,,1], [6,,1], , [2,,1]";
     v0 = s = filtre(v, 0);
@@ -330,7 +330,7 @@ sd_format(const char *v, long flag)
   {
     char c = *v;
     if (c!='e' && c!='f' && c!='g')
-      pari_err(syntaxer,"default: inexistent format",v,v);
+      pari_err(e_SYNTAX,"default: inexistent format",v,v);
     fmt->format = c; v++;
 
     if (isdigit((int)*v))
@@ -344,7 +344,7 @@ sd_format(const char *v, long flag)
   }
   if (flag == d_RETURN)
   {
-    char *s = stackmalloc(64);
+    char *s = stack_malloc(64);
     (void)sprintf(s, "%c.%ld", fmt->format, fmt->sigd);
     return strtoGENstr(s);
   }
@@ -378,12 +378,6 @@ sd_secure(const char *v, long flag)
   return sd_toggle(v,flag,"secure", &(GP_DATA->secure));
 }
 
-static THREAD long dbg = -1;
-void
-dbg_block(void) { if (DEBUGLEVEL) { dbg = DEBUGLEVEL; DEBUGLEVEL = 0; } }
-void
-dbg_release(void) { if (dbg >= 0) { DEBUGLEVEL = dbg; dbg = -1; } }
-
 GEN
 sd_debug(const char *v, long flag)
 { return sd_ulong(v,flag,"debug",&DEBUGLEVEL, 0,20,NULL); }
@@ -403,45 +397,45 @@ init_hist(gp_data *D, size_t s, ulong t)
   gp_hist *H = D->hist;
   H->total = t;
   H->size = s;
-  H->res = (GEN *) pari_calloc(s * sizeof(GEN));
+  H->v = (gp_hist_cell*)pari_calloc(s * sizeof(gp_hist_cell));
 }
 GEN
-sd_histsize(const char *v, long flag)
+sd_histsize(const char *s, long flag)
 {
   gp_hist *H = GP_DATA->hist;
   ulong n = H->size;
-  GEN r = sd_ulong(v,flag,"histsize",&n, 1,
+  GEN r = sd_ulong(s,flag,"histsize",&n, 1,
                      (LONG_MAX / sizeof(long)) - 1,NULL);
   if (n != H->size)
   {
     const ulong total = H->total;
     long g, h, k, kmin;
-    GEN *resG = H->res, *resH; /* G = old data, H = new one */
-    size_t sG = H->size, sH;
+    gp_hist_cell *v = H->v, *w; /* v = old data, w = new one */
+    size_t sv = H->size, sw;
 
     init_hist(GP_DATA, n, total);
     if (!total) return r;
 
-    resH = H->res;
-    sH   = H->size;
+    w = H->v;
+    sw= H->size;
     /* copy relevant history entries */
-    g     = (total-1) % sG;
-    h = k = (total-1) % sH;
-    kmin = k - minss(sH, sG);
+    g     = (total-1) % sv;
+    h = k = (total-1) % sw;
+    kmin = k - minss(sw, sv);
     for ( ; k > kmin; k--, g--, h--)
     {
-      resH[h] = resG[g];
-      resG[g] = NULL;
-      if (!g) g = sG;
-      if (!h) h = sH;
+      w[h]   = v[g];
+      v[g].z = NULL;
+      if (!g) g = sv;
+      if (!h) h = sw;
     }
     /* clean up */
-    for ( ; resG[g]; g--)
+    for ( ; v[g].z; g--)
     {
-      gunclone(resG[g]);
-      if (!g) g = sG;
+      gunclone(v[g].z);
+      if (!g) g = sv;
     }
-    pari_free((void*)resG);
+    pari_free((void*)v);
   }
   return r;
 }
@@ -458,7 +452,7 @@ TeX_define2(const char *s, const char *def) {
 static FILE *
 open_logfile(const char *s) {
   FILE *log = fopen(s, "a");
-  if (!log) pari_err(openfiler,"logfile",s);
+  if (!log) pari_err_FILE("logfile",s);
 #ifndef WINCE
   setbuf(log,(char *)NULL);
 #endif
@@ -511,6 +505,10 @@ sd_TeXstyle(const char *v, long flag)
   GP_DATA->fmt->TeXstyle = n; return z;
 }
 
+GEN
+sd_nbthreads(const char *v, long flag)
+{ return sd_ulong(v,flag,"nbthreads",&pari_mt_nbthreads, 1,LONG_MAX,NULL); }
+
 GEN
 sd_output(const char *v, long flag)
 {
@@ -538,18 +536,20 @@ sd_parisize(const char *v, long flag)
 }
 
 GEN
-sd_primelimit(const char *v, long flag)
+sd_threadsize(const char *v, long flag)
 {
-  ulong max = maxprime(), n = max;
-  GEN r = sd_ulong(v,flag,"primelimit",&n, 0,2*(ulong)(LONG_MAX-1024) + 1,NULL);
-  if (n != max)
-  {
-    byteptr ptr = initprimes(n);
-    pari_free(diffptr); diffptr = ptr;
-  }
+  ulong size = GP_DATA->threadsize, n = size;
+  GEN r = sd_ulong(v,flag,"threadsize",&n, 0,LONG_MAX,NULL);
+  if (n != size)
+    GP_DATA->threadsize = n;
   return r;
 }
 
+GEN
+sd_primelimit(const char *v, long flag)
+{ return sd_ulong(v,flag,"primelimit",&(GP_DATA->primelimit),
+                  0,2*(ulong)(LONG_MAX-1024) + 1,NULL); }
+
 GEN
 sd_simplify(const char *v, long flag)
 { return sd_toggle(v,flag,"simplify", &(GP_DATA->simplify)); }
@@ -558,6 +558,10 @@ GEN
 sd_strictmatch(const char *v, long flag)
 { return sd_toggle(v,flag,"strictmatch", &(GP_DATA->strictmatch)); }
 
+GEN
+sd_strictargs(const char *v, long flag)
+{ return sd_toggle(v,flag,"strictargs", &(GP_DATA->strictargs)); }
+
 GEN
 sd_string(const char *v, long flag, const char *s, char **pstr)
 {
@@ -566,7 +570,7 @@ sd_string(const char *v, long flag, const char *s, char **pstr)
   {
     char *str, *ev = path_expand(v);
     long l = strlen(ev) + 256;
-    str = (char *) malloc(l);
+    str = (char *) pari_malloc(l);
     strftime_expand(ev,str, l-1); pari_free(ev);
     if (GP_DATA->secure)
     {
@@ -624,10 +628,9 @@ sd_datadir(const char *v, long flag)
   return gnil;
 }
 
-GEN
-sd_path(const char *v, long flag)
+static GEN
+sd_PATH(const char *v, long flag, const char* s, gp_path *p)
 {
-  gp_path *p = GP_DATA->path;
   if (v)
   {
     pari_free((void*)p->PATH);
@@ -637,9 +640,15 @@ sd_path(const char *v, long flag)
   }
   if (flag == d_RETURN) return strtoGENstr(p->PATH);
   if (flag == d_ACKNOWLEDGE)
-    pari_printf("   path = \"%s\"\n",p->PATH);
+    pari_printf("   %s = \"%s\"\n", s, p->PATH);
   return gnil;
 }
+GEN
+sd_path(const char *v, long flag)
+{ return sd_PATH(v, flag, "path", GP_DATA->path); }
+GEN
+sd_sopath(char *v, int flag)
+{ return sd_PATH(v, flag, "sopath", GP_DATA->sopath); }
 
 static const char *DFT_PRETTYPRINTER = "tex2mail -TeX -noindent -ragged -by_par";
 GEN
@@ -652,7 +661,7 @@ sd_prettyprinter(const char *v, long flag)
     int cancel = (!strcmp(v,"no"));
 
     if (GP_DATA->secure)
-      pari_err(talker,"[secure mode]: can't modify 'prettyprinter' default (to %s)",v);
+      pari_err(e_MISC,"[secure mode]: can't modify 'prettyprinter' default (to %s)",v);
     if (!strcmp(v,"yes")) v = DFT_PRETTYPRINTER;
     if (old && strcmp(old,v) && pp->file)
     {
@@ -695,7 +704,7 @@ defaults_list(pari_stack *s)
   long i;
   for (i = 0; i < functions_tblsz; i++)
     for (ep = defaults_hash[i]; ep; ep = ep->next)
-      if (ep->menu == 16) stack_pushp(s, ep);
+      if (ep->menu == 16) pari_stack_pushp(s, ep);
 }
 /* ep associated to function f of arity 2. Call f(v,flag) */
 static GEN
@@ -710,17 +719,17 @@ setdefault(const char *s, const char *v, long flag)
     pari_stack st;
     entree **L;
     long i;
-    stack_init(&st, sizeof(*L), (void**)&L);
+    pari_stack_init(&st, sizeof(*L), (void**)&L);
     defaults_list(&st);
     qsort (L, st.n, sizeof(*L), compare_name);
     for (i = 0; i < st.n; i++) (void)call_f2(L[i], NULL, d_ACKNOWLEDGE);
-    stack_delete(&st);
+    pari_stack_delete(&st);
     return gnil;
   }
   ep = is_entry_intern(s, defaults_hash, NULL);
   if (!ep)
   {
-    pari_err(talker,"unknown default: %s",s);
+    pari_err(e_MISC,"unknown default: %s",s);
     return NULL; /* not reached */
   }
   return call_f2(ep, v, flag);
@@ -737,12 +746,11 @@ default0(const char *a, const char *b) { return setdefault(a,b, b? d_SILENT: d_R
 /*                     INITIALIZE GP_DATA                           */
 /*                                                                  */
 /********************************************************************/
-/* initialize D->path */
+/* initialize path */
 static void
-init_path(gp_data *D)
+init_path(gp_path *path, const char *v)
 {
-  gp_path *path = D->path;
-  path->PATH = pari_strdup(pari_default_path());
+  path->PATH = pari_strdup(v);
   path->dirs = NULL;
 }
 
@@ -773,30 +781,34 @@ default_gp_data(void)
   static gp_data __GPDATA, *D = &__GPDATA;
   static gp_hist __HIST;
   static gp_pp   __PP;
-  static gp_path __PATH;
+  static gp_path __PATH, __SOPATH;
   static pari_timer __T;
 
   D->flags       = 0;
+  D->primelimit  = 500000;
 
   /* GP-specific */
   D->breakloop   = 1;
   D->echo        = 0;
   D->lim_lines   = 0;
+  D->linewrap    = 0;
   D->recover     = 1;
   D->chrono      = 0;
 
+  D->strictargs  = 0;
   D->strictmatch = 1;
   D->simplify    = 1;
   D->secure      = 0;
   D->use_readline= 0;
-  D->last_time   = 0;
   D->T    = &__T;
   D->hist = &__HIST;
   D->pp   = &__PP;
   D->path = &__PATH;
+  D->sopath=&__SOPATH;
   init_fmt(D);
   init_hist(D, 5000, 0);
-  init_path(D);
+  init_path(D->path, pari_default_path());
+  init_path(D->sopath, "");
   init_pp(D);
   return D;
 }
