@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -71,8 +69,8 @@ red_cyclop(GEN T, long p)
   if (d <= -2) return T;
 
   /* reduce mod (x^p - 1) */
-  y = leafcopy(T); z = (GEN*)(y+2);
-  for (i = 0; i<=d; i++) z[i] = addii(z[i], z[i+p]);
+  y = ZX_mod_Xnm1(T, p);
+  z = (GEN*)(y+2);
 
   /* reduce mod x^(p-1) + ... + 1 */
   d = p-1;
@@ -102,7 +100,7 @@ red_cyclo2n_ip(GEN x, long n)
 {
   long i, pow2 = 1L<<(n-1);
   for (i = lg(x)-1; i>pow2+1; i--)
-    if (signe(x[i])) gel(x,i-pow2) = subii(gel(x,i-pow2), gel(x,i));
+    if (signe(gel(x,i))) gel(x,i-pow2) = subii(gel(x,i-pow2), gel(x,i));
   return normalizepol_lg(x, i+1);
 }
 static GEN
@@ -310,117 +308,161 @@ powpolmod(Cache *C, Red *R, long p, long k, GEN jac)
   return _powpolmod(C, jac, R, _sqr);
 }
 
-/* globfa contains the odd prime divisors of e(t) */
+/* Return e(t) = \prod_{p-1 | t} p^{1+v_p(t)}}
+ * faet contains the odd prime divisors of e(t) */
 static GEN
-e(ulong t, GEN *globfa)
+compute_e(ulong t, GEN *faet)
 {
-  GEN fa, P, E, s, Primes;
-  ulong nbd, m, k, d;
-  long lfa, i, j;
+  GEN L, P, D = divisorsu(t);
+  long l = lg(D);
+  ulong k;
 
-  fa = factoru(t);
-  P = gel(fa,1);
-  E = gel(fa,2); lfa = lg(P);
-  nbd = 1;
-  for (i=1; i<lfa; i++) { E[i]++; nbd *= E[i]; }
-  Primes = vecsmalltrunc_init(nbd + 1);
-  s = gen_2; /* nbd = number of divisors */
-  for (k=0; k<nbd; k++)
+  P = vecsmalltrunc_init(l);
+  L = vecsmalltrunc_init(l);
+  for (k=l-1; k>1; k--) /* k != 1: avoid d = 1 */
   {
-    m = k; d = 1;
-    for (j=1; m; j++)
-    {
-      d *= upowuu(P[j], m % E[j]);
-      m /= E[j];
-    }
-    /* d runs through the divisors of t */
+    ulong d = D[k];
     if (uisprime(++d))
-    {
-      if (d > 5000000) return gen_0;
-      if (d != 2) vecsmalltrunc_append(Primes, d);
-      s = muliu(s, upowuu(d, 1 + u_lval(t,d)));
+    { /* we want q = 1 (mod p) prime, not too large */
+      if (d > 50000000) return gen_0;
+      vecsmalltrunc_append(P, d);
+      vecsmalltrunc_append(L, upowuu(d, 1 + u_lval(t,d)));
     }
   }
-  if (globfa) { vecsmall_sort(Primes); *globfa = Primes; }
-  return s;
+  if (faet) *faet = P;
+  return shifti(zv_prod_Z(L), 2 + u_lval(t,2));
 }
 
-static ulong
-compt(GEN N)
-{
-  pari_sp av0 = avma;
-  double Cd = 100 * rtodbl(logr_abs(itor(N,DEFAULTPREC))) / log(10.);
-  ulong t, C = (ulong)ceil(Cd);
-  GEN B;
-  avma = av0;
-  /* C < [200*log_10 e(t)] ==> return t. For e(t) < 10^529, N < 10^1058 */
-  if (C <    540) return        6;
-  if (C <    963) return       12;
-  if (C <   1023) return       24;
-  if (C <   1330) return       48;
-  if (C <   1628) return       36;
-  if (C <   1967) return       60;
-  if (C <   2349) return      120;
-  if (C <   3083) return      180;
-  if (C <   3132) return      240;
-  if (C <   3270) return      504;
-  if (C <   3838) return      360;
-  if (C <   4115) return      420;
-  if (C <   4621) return      720;
-  if (C <   4987) return      840;
-  if (C <   5079) return     1440;
-  if (C <   6212) return     1260;
-  if (C <   6686) return     1680;
-  if (C <   8137) return     2520;
-  if (C <   8415) return     3360;
-  if (C <  10437) return     5040;
-  if (C <  11643) return    13860;
-  if (C <  12826) return    10080;
-  if (C <  13369) return    16380;
-  if (C <  13540) return    21840;
-  if (C <  15060) return    18480;
-  if (C <  15934) return    27720;
-  if (C <  17695) return    32760;
-  if (C <  18816) return    36960;
-  if (C <  21338) return    55440;
-  if (C <  23179) return    65520;
-  if (C <  23484) return    98280;
-  if (C <  27465) return   110880;
-  if (C <  30380) return   131040;
-  if (C <  31369) return   166320;
-  if (C <  33866) return   196560;
-  if (C <  34530) return   262080;
-  if (C <  36195) return   277200;
-  if (C <  37095) return   360360;
-  if (C <  38179) return   480480;
-  if (C <  41396) return   332640;
-  if (C <  43301) return   554400;
-  if (C <  47483) return   720720;
-  if (C <  47742) return   665280;
-  if (C <  50202) return   831600;
-  if (C <  52502) return  1113840;
-  if (C <  60245) return  1441440;
-  if (C <  63112) return  1663200;
-  if (C <  65395) return  2227680;
-  if (C <  69895) return  2162160;
-  if (C <  71567) return  2827440;
-  if (C <  75708) return  3326400;
-  if (C <  79377) return  3603600;
-  if (C <  82703) return  6126120;
-  if (C <  91180) return  4324320;
-  if (C <  93978) return  6683040;
-  if (C <  98840) return  7207200;
-  if (C <  99282) return 11138400;
-  if (C < 105811) return  8648640;
+/* Table obtained by the following script:
 
+install(compute_e, "LD&"); \\ remove 'static' first
+
+table(first = 6, step = 6, MAXT = 8648640)=
+{
+  emax = 0;
+  forstep(t = first, MAXT, step,
+    e = compute_e(t);
+    if (e > 1.9*emax, emax = e;
+      printf("  if (C < %5.5g) return %8d;\n", 2*log(e)/log(2)*0.9999, t)
+    );
+  );
+}
+
+Previous values can be recovered using the following table:
+
+T=[6,12,24,48,36,60,120,180,240,504,360,420,720,840,1440,1260,1680,2520,3360,5040,13860,10080,16380,21840,18480,27720,32760,36960,55440,65520,98280,110880,131040,166320,196560,262080,277200,360360,480480,332640,554400,720720,665280,831600,1113840,1441440,1663200,2227680,2162160,2827440,3326400,3603600,6126120,4324320,6683040,7207200,11138400,8648640];
+f(t) = 2*log(compute_e(t))/log(2);
+for (i=1,#T, t=T[i]; printf("  if (C < %5.5g) return %8d;\n", f(t),t));
+
+*/
+
+/* assume C < 3514.6 */
+static ulong
+compute_t_small(double C)
+{
+  if (C < 17.953) return        6;
+  if (C < 31.996) return       12;
+  if (C < 33.996) return       24;
+  if (C < 54.079) return       36;
+  if (C < 65.325) return       60;
+  if (C < 68.457) return       72;
+  if (C < 70.783) return      108;
+  if (C < 78.039) return      120;
+  if (C < 102.41) return      180;
+  if (C < 127.50) return      360;
+  if (C < 136.68) return      420;
+  if (C < 153.43) return      540;
+  if (C < 165.66) return      840;
+  if (C < 169.17) return     1008;
+  if (C < 178.52) return     1080;
+  if (C < 192.68) return     1200;
+  if (C < 206.34) return     1260;
+  if (C < 211.94) return     1620;
+  if (C < 222.09) return     1680;
+  if (C < 225.11) return     2016;
+  if (C < 244.19) return     2160;
+  if (C < 270.29) return     2520;
+  if (C < 279.50) return     3360;
+  if (C < 293.62) return     3780;
+  if (C < 346.68) return     5040;
+  if (C < 348.70) return     6480;
+  if (C < 383.34) return     7560;
+  if (C < 396.68) return     8400;
+  if (C < 426.04) return    10080;
+  if (C < 458.34) return    12600;
+  if (C < 527.16) return    15120;
+  if (C < 595.38) return    25200;
+  if (C < 636.29) return    30240;
+  if (C < 672.53) return    42840;
+  if (C < 684.90) return    45360;
+  if (C < 708.78) return    55440;
+  if (C < 771.30) return    60480;
+  if (C < 775.86) return    75600;
+  if (C < 859.62) return    85680;
+  if (C < 893.16) return   100800;
+  if (C < 912.27) return   110880;
+  if (C < 966.13) return   128520;
+  if (C < 1009.1) return   131040;
+  if (C < 1041.9) return   166320;
+  if (C < 1124.9) return   196560;
+  if (C < 1251.0) return   257040;
+  if (C < 1375.0) return   332640;
+  if (C < 1431.0) return   393120;
+  if (C < 1483.3) return   514080;
+  if (C < 1546.3) return   655200;
+  if (C < 1585.8) return   665280;
+  if (C < 1661.3) return   786240;
+  if (C < 1667.5) return   831600;
+  if (C < 1676.9) return   917280;
+  if (C < 1728.0) return   982800;
+  if (C < 1747.4) return  1081080;
+  if (C < 1773.6) return  1179360;
+  if (C < 1810.6) return  1285200;
+  if (C < 1924.5) return  1310400;
+  if (C < 2001.1) return  1441440;
+  if (C < 2096.3) return  1663200;
+  if (C < 2165.8) return  1965600;
+  if (C < 2321.6) return  2162160;
+  if (C < 2368.2) return  2751840;
+  if (C < 2377.2) return  2827440;
+  if (C < 2514.7) return  3326400;
+  if (C < 2588.5) return  3341520;
+  if (C < 2636.6) return  3603600;
+  if (C < 2667.2) return  3931200;
+  if (C < 3028.6) return  4324320;
+  if (C < 3045.5) return  5654880;
+  if (C < 3080.5) return  6652800;
+  if (C < 3121.6) return  6683040;
+  if (C < 3283.1) return  7207200;
+  return  8648640;
+}
+
+/* return t such that e(t) > sqrt(N), set *faet = odd prime divisors of e(t) */
+static ulong
+compute_t(GEN N, GEN *e, GEN *faet)
+{
+  /* 2^e b <= N < 2^e (b+1), where b >= 2^52. Approximating log_2 N by
+   * log2(gtodouble(N)) ~ e+log2(b), the error is less than log(1+1/b) < 1e-15*/
+  double C = dbllog2(N) + 1e-6; /* > log_2 N */
+  ulong t;
+  GEN B;
+  /* Return "smallest" t such that f(t) >= C, which implies e(t) > sqrt(N) */
+  /* For N < 2^3515 ~ 10^1058 */
+  if (C < 3514.6)
+  {
+    t = compute_t_small(C);
+    *e = compute_e(t, faet);
+    return t;
+  }
   B = sqrti(N);
   for (t = 8648640+840;; t+=840)
   {
     pari_sp av = avma;
-    if (cmpii(e(t, NULL), B) > 0) break;
+    *e = compute_e(t, faet);
+    if (cmpii(*e, B) > 0) break;
     avma = av;
   }
-  avma = av0; return t;
+  return t;
 }
 
 /* T[i] = discrete log of i in (Z/q)^*, q odd prime
@@ -463,7 +505,7 @@ static GEN
 get_jac(Cache *C, ulong q, long pk, GEN tabg)
 {
   ulong x, qs2 = q>>1; /* (q-1)/2 */
-  GEN vpk = const_vecsmall(pk, 0);
+  GEN vpk = zero_zv(pk);
 
   for (x=2; x<=qs2; x++) vpk[ tabg[x]%pk + 1 ] += 2;
   vpk[ tabg[x]%pk + 1 ]++; /* x = (q+1)/2 */
@@ -474,16 +516,13 @@ get_jac(Cache *C, ulong q, long pk, GEN tabg)
 static GEN
 get_jac2(GEN N, ulong q, long k, GEN *j2q, GEN *j3q)
 {
-  GEN jpq, vpk, T;
+  GEN jpq, vpk, T = computetabdl(q);
   ulong x, pk, i, qs2;
 
-  if (k == 1) return NULL;
-
-  T = computetabdl(q);
   /* could store T[x+1] + T[x] + qs2 (cf compute_g).
    * Recompute instead, saving half the memory. */
   pk = 1UL << k;;
-  vpk = const_vecsmall(pk, 0);
+  vpk = zero_zv(pk);
 
   qs2 = q>>1; /* (q-1)/2 */
 
@@ -510,37 +549,6 @@ get_jac2(GEN N, ulong q, long k, GEN *j2q, GEN *j3q)
   return jpq;
 }
 
-static void
-calcjac(Cache **pC, GEN globfa, GEN *ptabfaq, GEN *ptabj)
-{
-  GEN J, tabg, faq, tabfaq, tabj, P, PE;
-  long lfaq, j;
-  ulong i, q, l;
-  pari_sp av;
-
-  l = lg(globfa);
-  *ptabfaq = tabfaq= cgetg(l,t_VEC);
-  *ptabj   = tabj  = cgetg(l,t_VEC);
-  for (i=1; i<l; i++)
-  {
-    q = globfa[i]; /* odd prime */
-    faq = factoru_pow(q-1); gel(tabfaq,i) = faq;
-    P = gel(faq,1); lfaq = lg(P);
-    PE= gel(faq,3);
-    av = avma;
-    tabg = compute_g(q);
-
-    J = cgetg(lfaq,t_VEC);
-    gel(J,1) = cgetg(1,t_STR); /* dummy */
-    for (j=2; j<lfaq; j++) /* skip p = P[1] = 2 */
-    {
-      long pe = PE[j];
-      gel(J,j) = get_jac(pC[pe], q, pe, tabg);
-    }
-    gel(tabj,i) = gerepilecopy(av, J);
-  }
-}
-
 /* N = 1 mod p^k, return an elt of order p^k in (Z/N)^* */
 static GEN
 finda(Cache *Cp, GEN N, long pk, long p)
@@ -654,6 +662,7 @@ filltabs(Cache *C, Cache *Cp, Red *R, long p, long pk, long ltab)
   av = avma; m = divis(R->N, pk);
   for (e=1; e<=ltab && signe(m); e++)
   {
+//    long s = Z_lvalrem(m, 2, &m);
     long s = vali(m); m = shifti(m,-s);
     tabt[e] = e==1? s: s + R->k;
     taba[e] = signe(m)? ((mod2BIL(m) & R->mask)+1)>>1: 0;
@@ -667,56 +676,45 @@ filltabs(Cache *C, Cache *Cp, Red *R, long p, long pk, long ltab)
 static Cache *
 alloc_cache(void)
 {
-  Cache *C = (Cache*)new_chunk(sizeof(Cache) / sizeof(long));
+  Cache *C = (Cache*)stack_malloc(sizeof(Cache));
   C->matvite = NULL;
   C->avite   = NULL;
   C->ctsgt = 0; return C;
 }
 
 static Cache **
-calcglobs(Red *R, ulong t, long *pltab, GEN *pP)
+calcglobs(Red *R, ulong t, long *plpC, long *pltab, GEN *pP)
 {
   GEN fat, P, E, PE;
-  long lv, lfa, pk, p, e, i, k;
-  long ltab, b;
+  long lv, i, k, b;
   Cache **pC;
 
-  b = bit_accuracy(lgefint(R->N)) - 1;
-  while ( !bittest(R->N,b) ) b--;
-  b++;
-
+  b = expi(R->N)+1;
   k = 3; while (((k+1)*(k+2) << (k-1)) < b) k++;
-  *pltab = ltab = (b/k) + 2;
+  *pltab = (b/k)+2;
   R->k  = k;
   R->lv = 1L << (k-1);
   R->mask = (1UL << k) - 1;
 
   fat = factoru_pow(t);
-  P = gel(fat,1); lfa = lg(P);
+  P = gel(fat,1);
   E = gel(fat,2);
   PE= gel(fat,3);
-  lv = 1;
-  for (i=1; i<lfa; i++)
-  {
-    long pe = PE[i];
-    if (pe > lv) lv = pe;
-  }
-  pC = (Cache**)cgetg(lv + 1, t_VEC);
+  *plpC = lv = vecsmall_max(PE); /* max(p^e, p^e | t) */
+  pC = (Cache**)stack_malloc((lv+1)*sizeof(Cache*));
   pC[1] = alloc_cache(); /* to be used as temp in step5() */
   for (i = 2; i <= lv; i++) pC[i] = NULL;
-  for (i=1; i<lfa; i++)
+  for (i=1; i<lg(P); i++)
   {
-    pk = p = P[i];
-    e = E[i];
+    long pk, p = P[i], e = E[i];
+    pk = p;
     for (k=1; k<=e; k++, pk*=p)
     {
       pC[pk] = alloc_cache();
-      if (!filltabs(pC[pk], pC[p], R, p,pk, ltab)) return NULL;
+      if (!filltabs(pC[pk], pC[p], R, p,pk, *pltab)) return NULL;
     }
   }
-  if (DEBUGLEVEL) err_printf("\n");
-  *pP = P;
-  return pC;
+  *pP = P; return pC;
 }
 
 /* sig_a^{-1}(z) for z in Q(zeta_pk) and sig_a: zeta -> zeta^a. Assume
@@ -794,13 +792,14 @@ look_eta2(long k, GEN z)
 }
 
 static long
-step4a(Cache *C, Red *R, ulong q, long p, long k, GEN jpq)
+step4a(Cache *C, Red *R, ulong q, long p, long k, GEN tabg)
 {
   const long pk = upowuu(p,k);
   long ind;
-  GEN s1, s2, s3;
+  GEN jpq, s1, s2, s3;
 
-  if (!jpq) jpq = get_jac(C, q, pk, compute_g(q));
+  if (!tabg) tabg = compute_g(q);
+  jpq = get_jac(C, q, pk, tabg);
   s1 = autvec_TH(pk, jpq, C->E, C->cyc);
   s2 = powpolmod(C,R, p,k, s1);
   s3 = autvec_AL(pk, jpq, C->E, R);
@@ -873,44 +872,48 @@ step4d(Cache *C, Red *R, ulong q)
   return -1;
 }
 
+static GEN
+_res(long a, long b) { return b? mkvec2s(a, b): mkvecs(a); }
+
 /* return 1 [OK so far] or <= 0 [not a prime] */
-static long
-step5(Cache **pC, Red *R, long p, GEN et, ulong ltab)
+static GEN
+step5(Cache **pC, Red *R, long p, GEN et, ulong ltab, long lpC)
 {
   pari_sp av;
-  ulong ct = 1, q;
+  ulong q;
   long pk, k, fl = -1;
-  byteptr d = diffptr+2;
   Cache *C, *Cp;
+  forprime_t T;
 
-  for (q = 3; *d; )
-  {
-    if (q%p != 1 || umodiu(et,q) == 0) goto repeat;
-
-    if (umodiu(R->N,q) == 0) return -1;
+  (void)u_forprime_arith_init(&T, 3, ULONG_MAX, 1,p);
+  while( (q = u_forprime_next(&T)) )
+  { /* q = 1 (mod p) */
+    if (umodiu(et,q) == 0) continue;
+    if (umodiu(R->N,q) == 0) return _res(1,p);
     k = u_lval(q-1, p);
     pk = upowuu(p,k);
-    if (pk < lg(pC) && pC[pk]) { C = pC[pk]; Cp = pC[p]; }
-    else {
-      C = pC[1]; C->matvite = NULL; /* re-init */
+    if (pk <= lpC && pC[pk]) {
+      C = pC[pk];
+      Cp = pC[p];
+    } else {
+      C = pC[1];
       Cp = NULL;
+      C->matvite = NULL; /* re-init */
     }
 
-    if (!filltabs(C, Cp, R, p, pk, ltab)) return 0;
-    av = avma; R->C = C->cyc;
+    av = avma;
+    if (!filltabs(C, Cp, R, p, pk, ltab)) return _res(1,0);
+    R->C = C->cyc;
     if (p >= 3)      fl = step4a(C,R, q,p,k, NULL);
     else if (k >= 3) fl = step4b(C,R, q,k);
     else if (k == 2) fl = step4c(C,R, q);
     else             fl = step4d(C,R, q);
-    if (fl == -1) return (long)(-q);
-    if (fl == 1) return ct;
+    if (fl == -1) return _res(q,p);
+    if (fl == 1) return NULL; /*OK*/
     avma = av;
-    ct++;
-   repeat:
-    NEXT_PRIME_VIADIFF(q,d);
   }
-  pari_err(bugparier,"aprcl test fails! this is highly improbable");
-  return 0;
+  pari_err_BUG("aprcl test fails! This is highly improbable");
+  return NULL;
 }
 
 static GEN
@@ -924,27 +927,23 @@ step6(GEN N, ulong t, GEN et)
   for (i=1; i<t; i++)
   {
     r = remii(mulii(r,N1), et);
-    if (gequal1(r)) break;
+    if (equali1(r)) break;
     if (!signe(remii(N,r)) && !equalii(r,N)) return mkvec2(r, gen_0);
     if ((i & 0x1f) == 0) r = gerepileuptoint(av, r);
   }
   return gen_1;
 }
 
-static GEN
-_res(long a, long b) { return b? mkvec2s(a, b): mkvecs(a); }
-
 static GEN
 aprcl(GEN N)
 {
-  GEN et, fat, flaglp, tabfaq, tabj, res, globfa;
-  long i, j, l, ltab, lfat, fl, ctglob = 0;
-  ulong p, q, t;
-  pari_sp av;
+  GEN et, fat, flaglp, faet = NULL; /*-Wall*/
+  long i, j, l, ltab, lfat, lpC;
+  ulong t;
   Red R;
   Cache **pC;
 
-  if (typ(N) != t_INT) pari_err(arither1);
+  if (typ(N) != t_INT) pari_err_TYPE("aprcl",N);
   if (cmpis(N,12) <= 0)
     switch(itos(N))
     {
@@ -952,78 +951,73 @@ aprcl(GEN N)
       default: return _res(0,0);
     }
   if (Z_issquare(N)) return _res(0,0);
-  t = compt(N);
-  if (DEBUGLEVEL) err_printf("Starting APRCL: Choosing t = %ld\n",t);
-  et = e(t, &globfa);
-  if (cmpii(sqri(et),N) < 0) pari_err(bugparier,"aprcl: e(t) too small");
-  if (!gequal1(gcdii(N,mului(t,et)))) return _res(1,0);
+  t = compute_t(N, &et, &faet);
+  if (DEBUGLEVEL) err_printf("Starting APRCL with t = %ld\n",t);
+  if (cmpii(sqri(et),N) < 0) pari_err_BUG("aprcl: e(t) too small");
+  if (!equali1(gcdii(N,mului(t,et)))) return _res(1,0);
 
   R.N = N;
   R.N2= shifti(N, -1);
-  pC = calcglobs(&R, t, &ltab, &fat);
+  pC = calcglobs(&R, t, &lpC, &ltab, &fat);
   if (!pC) return _res(1,0);
-  lfat = lg(fat); p = fat[lfat-1]; /* largest p | t */
-  flaglp = cgetg(p+1, t_VECSMALL);
-  flaglp[2] = 0;
+  lfat = lg(fat);
+  flaglp = cgetg(lfat, t_VECSMALL);
+  flaglp[1] = 0;
   for (i=2; i<lfat; i++)
   {
-    p = fat[i]; q = p*p;
-    flaglp[p] = (Fl_powu(umodiu(N,q),p-1,q) != 1);
+    ulong p = fat[i];
+    flaglp[i] = equaliu(Fp_powu(N, p-1, sqru(p)), 1);
   }
-  calcjac(pC, globfa, &tabfaq, &tabj);
+  vecsmall_sort(faet);
 
-  av = avma; l = lg(globfa);
-  if (DEBUGLEVEL>2)
+  l = lg(faet);
+  if (DEBUGLEVEL>2) err_printf("Step4: %ld q-values\n", l-1);
+  for (i=l-1; i>0; i--) /* decreasing order: slowest first */
   {
-    err_printf("Jacobi sums and tables computed\n");
-    err_printf("Step4: q-values (# = %ld): ", l-1);
-  }
-  for (i=l-1; i>0; i--)
-  {
-    GEN faq = gel(tabfaq,i), P = gel(faq,1), E = gel(faq,2), PE = gel(faq,3);
+    pari_sp av1 = avma;
+    ulong q = faet[i];
+    GEN faq = factoru_pow(q-1), tabg = compute_g(q);
+    GEN P = gel(faq,1), E = gel(faq,2), PE = gel(faq,3);
     long lfaq = lg(P);
-    q = globfa[i]; if (DEBUGLEVEL>2) err_printf("%ld ",q);
-    for (j=1; j<lfaq; j++, avma = av)
+    pari_sp av2 = avma;
+    if (DEBUGLEVEL>2) err_printf("testing Jacobi sums for q = %ld...",q);
+    for (j=1; j<lfaq; j++, avma = av2)
     {
-      Cache *C;
-      long pe, e;
-      p = P[j];
-      e = E[j];
-      pe= PE[j]; C = pC[pe];
+      long p = P[j], e = E[j], pe = PE[j], fl;
+      Cache *C = pC[pe];
       R.C = C->cyc;
-      if (p >= 3)      fl = step4a(C,&R, q,p,e, gmael(tabj,i,j));
+      if (p >= 3)      fl = step4a(C,&R, q,p,e, tabg);
       else if (e >= 3) fl = step4b(C,&R, q,e);
       else if (e == 2) fl = step4c(C,&R, q);
       else             fl = step4d(C,&R, q);
       if (fl == -1) return _res(q,p);
-      if (fl == 1) flaglp[p] = 1;
+      if (fl == 1) flaglp[ zv_search(fat, p) ] = 0;
     }
+    if (DEBUGLEVEL>2) err_printf("OK\n");
+    avma = av1;
   }
-  if (DEBUGLEVEL>2) err_printf("\nStep5: testing conditions lp\n");
-  for (i=1; i<lfat; i++)
+  if (DEBUGLEVEL>2) err_printf("Step5: testing conditions lp\n");
+  for (i=2; i<lfat; i++) /*skip fat[1] = 2*/
   {
-    p = fat[i]; if (flaglp[p]) continue;
-
-    fl = step5(pC, &R, p, et, ltab);
-    if (!fl) return _res(1,0);
-    if (fl < 0) return _res(fl,p);
-    if (fl > ctglob) ctglob = fl; /* DEBUG */
+    pari_sp av = avma;
+    long p = fat[i];
+    GEN r;
+    if (flaglp[i] && (r = step5(pC, &R, p, et, ltab, lpC))) return r;
+    avma = av;
   }
-  if (DEBUGLEVEL>2) err_printf("Step6: testing potential divisors\n");
-  res = step6(N, t, et);
   if (DEBUGLEVEL>2)
   {
     ulong sc = pC[1]->ctsgt;
     err_printf("Individual Fermat powerings:\n");
-    for (i=2; i<lg(pC); i++)
+    for (i=2; i<lpC; i++)
       if (pC[i]) {
         err_printf("  %-3ld: %3ld\n", i, pC[i]->ctsgt);
         sc += pC[i]->ctsgt;
       }
     err_printf("Number of Fermat powerings = %lu\n",sc);
-    err_printf("Maximal number of nondeterministic steps = %lu\n",ctglob);
+    err_printf("Step6: testing potential divisors\n");
   }
-  return res;
+  return step6(N, t, et);
 }
 
 long
